/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
 *  \file   ni_quadra_dec_api.c
 *
 *  \brief  Private definitions used by ni_quadra_dec_api.c for video processing
 *          tasks
 ******************************************************************************/

#include "ni_defs.h"
#include "ni_rsrc_api.h"
#include "ni_util.h"
#include "ni_device_api.h"
#include "ni_av_codec.h"
#include "ni_bitstream.h"
#include "ni_quadra_dec_api.h"

#define USER_DATA_UNREGISTERED_SEI_PAYLOAD_TYPE 5
#define NI_DEC_PKT_BUF_NUM_MAX 300
// copy maximum number of bytes of a string from src to dst, ensuring null byte
// terminated
static void ni_quadra_xcoder_strncpy(char *dst, const char *src, int max)
{
    if (dst && src && max)
    {
        *dst = '\0';
        strncpy(dst, src, max);
        *(dst + max - 1) = '\0';
    }
}

void ni_quadra_align_free(void *opaque, uint8_t *data)
{
    ni_buf_t *buf = (ni_buf_t *)opaque;
    if (buf)
    {
        ni_decoder_frame_buffer_pool_return_buf(buf,
                                                (ni_buf_pool_t *)buf->pool);
    }
}

void ni_quadra_frame_free(void *opaque, uint8_t *data)
{
    int ret;
    if (data)
    {
        niFrameSurface1_t *p_data3 =
            (niFrameSurface1_t *)((uint8_t *)data);
        if (p_data3->ui16FrameIdx != 0)
        {
            ni_log(NI_LOG_DEBUG,
                   "Recycle trace ui16FrameIdx = [%d] DevHandle %d\n",
                   p_data3->ui16FrameIdx, p_data3->device_handle);
            ret = ni_hwframe_buffer_recycle(p_data3, p_data3->device_handle);
            if (ret != NI_RETCODE_SUCCESS)
            {
                ni_log(NI_LOG_ERROR,
                       "ERROR Failed to recycle trace ui16frameidx = [%d] "
                       "DevHandle %d\n",
                       p_data3->ui16FrameIdx, p_data3->device_handle);
            }
        }
        ni_aligned_free(data);
        (void)data;   // suppress cppcheck
    }
}

void ni_quadra_dec_close(ni_quadra_dec_context_t *ni_dec_ctx)
{
    ni_retcode_t ret;
    ni_xcoder_params_t *p_param = ni_dec_ctx->api_param;
    ni_session_context_t *p_session_ctx = ni_dec_ctx->api_ctx;
    ni_session_data_io_t *p_ni_packet = NULL;
    ni_session_data_io_t *p_session_data = ni_dec_ctx->p_session_data;
    ni_session_data_io_t *p_session_data1 = ni_dec_ctx->p_session_data1;
    ni_session_data_io_t *p_session_data2 = ni_dec_ctx->p_session_data2;

    if (p_session_ctx)
    {
        ret = ni_device_session_close(p_session_ctx, ni_dec_ctx->eos,
                                      NI_DEVICE_TYPE_DECODER);
        if (NI_RETCODE_SUCCESS != ret)
        {
            ni_log(NI_LOG_ERROR,
                   "Failed to close Decode Session (status = %d)\n", ret);
        }
        if (ni_cmp_fw_api_ver(
            (char *) &p_session_ctx->fw_rev[NI_XCODER_REVISION_API_MAJOR_VER_IDX], "6rg") >= 0)
        {
            ni_dec_ctx->ui32FramesCorrupted = p_session_ctx->session_statistic.ui32FramesCorrupted;
        }

        ni_device_session_context_clear(p_session_ctx);

        if (ni_dec_ctx->suspended)
        {
            ni_log(NI_LOG_ERROR,
                   "File BLK handle %d close suspended to frames Uninit\n",
                   p_session_ctx->blk_io_handle);   // suspended_device_handle
#ifdef __linux__
            ni_device_close(p_session_ctx->device_handle);
#endif
        } else
        {
#ifdef _WIN32
            ni_device_close(p_session_ctx->device_handle);
#elif __linux__
            ni_device_close(p_session_ctx->device_handle);
            ni_device_close(p_session_ctx->blk_io_handle);
#endif
        }

        p_session_ctx->device_handle = NI_INVALID_DEVICE_HANDLE;
        p_session_ctx->blk_io_handle = NI_INVALID_DEVICE_HANDLE;
    }
    free(ni_dec_ctx->extradata);
    ni_dec_ctx->extradata = NULL;
    ni_dec_ctx->extradata_size = 0;
    ni_dec_ctx->got_first_key_frame = 0;
    if (ni_dec_ctx->input_data_fifo)
    {
        // push back
        do
        {
            p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_write(ni_dec_ctx->input_data_fifo);
        } while (p_ni_packet);
        // free all
        do
        {
            p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_read(ni_dec_ctx->input_data_fifo);
            if (p_ni_packet)
            {
                ni_packet_buffer_free(&(p_ni_packet->data.packet));
            }
        } while (p_ni_packet);
    }

    ni_fifo_free(ni_dec_ctx->input_data_fifo);
    ni_log(NI_LOG_DEBUG, "%s(): decoder frame fifo freed.\n", __FUNCTION__);
    if (p_session_ctx)
    {
        free(p_session_ctx);
    }
    if (p_param)
    {
        free(p_param);
    }
    if (p_ni_packet)
    {
        free(p_ni_packet);
    }
    if (p_session_data)
    {
        free(p_session_data);
    }
    if (p_session_data1)
    {
        free(p_session_data1);
    }
    if (p_session_data2)
    {
        free(p_session_data2);
    }
}

int ni_quadra_setup_decoder(ni_quadra_dec_context_t *ni_dec_ctx)
{
    ni_xcoder_params_t *p_param = NULL;
    ni_session_context_t *p_session_ctx = NULL;
    ni_session_data_io_t *p_session_data = NULL;
    ni_session_data_io_t *p_session_data1 = NULL;
    ni_session_data_io_t *p_session_data2 = NULL;
    int ret = NI_RETCODE_SUCCESS;
    int min_resolution_width, min_resolution_height;
    ni_log_set_level(ff_to_ni_log_level(ni_dec_ctx->ff_log_level));
    ni_log(NI_LOG_DEBUG, "XCoder setup device decoder\n");

    p_session_data = malloc(sizeof(ni_session_data_io_t));
    if (!p_session_data)
    {
        ni_log(NI_LOG_ERROR,
               "ERROR %d:%s() line %d, alloc p_session_data failed\n", NI_ERRNO,
               __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    } else
    {
        ni_dec_ctx->p_session_data = p_session_data;
    }
    p_session_data1 = malloc(sizeof(ni_session_data_io_t));
    if (!p_session_data1)
    {
        ni_log(NI_LOG_ERROR,
               "ERROR %d:%s() line %d, alloc p_session_data failed\n", NI_ERRNO,
               __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    } else
    {
        ni_dec_ctx->p_session_data1 = p_session_data1;
    }
    p_session_data2 = malloc(sizeof(ni_session_data_io_t));
    if (!p_session_data2)
    {
        ni_log(NI_LOG_ERROR,
               "ERROR %d:%s() line %d, alloc p_session_data failed\n", NI_ERRNO,
               __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    } else
    {
        ni_dec_ctx->p_session_data2 = p_session_data2;
    }
    // Xcoder Session Context
    p_session_ctx = malloc(sizeof(ni_session_context_t));
    if (!p_session_ctx)
    {
        ni_log(NI_LOG_ERROR,
               "ERROR %d:%s() line %d, alloc session context failed\n",
               NI_ERRNO, __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    } else
    {
        ni_dec_ctx->api_ctx = p_session_ctx;
        p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
    }
    if (ni_device_session_context_init(p_session_ctx) < 0)
    {
        ni_log(NI_LOG_ERROR, "Error XCoder init decoder context failure\n");
        ret = NI_RETCODE_FAILURE;
        goto end;
    }

    min_resolution_width = NI_MIN_RESOLUTION_WIDTH;
    min_resolution_height = NI_MIN_RESOLUTION_HEIGHT;

    // Check codec id or format as well as profile idc.
    p_session_ctx->codec_format = ni_dec_ctx->codec_format;
    if (p_session_ctx->codec_format == NI_CODEC_FORMAT_JPEG)
    {
        min_resolution_width = NI_MIN_RESOLUTION_WIDTH_JPEG;
        min_resolution_height = NI_MIN_RESOLUTION_HEIGHT_JPEG;
    }

    if (ni_dec_ctx->width > NI_MAX_RESOLUTION_WIDTH ||
        ni_dec_ctx->height > NI_MAX_RESOLUTION_HEIGHT ||
        ni_dec_ctx->width * ni_dec_ctx->height > NI_MAX_RESOLUTION_AREA)
    {
        ni_log(NI_LOG_ERROR, "Error XCoder resolution %dx%d not supported\n",
               ni_dec_ctx->width, ni_dec_ctx->height);
        ni_log(NI_LOG_ERROR, "Max Supported Width: %d Height %d Area %d\n",
               NI_MAX_RESOLUTION_WIDTH, NI_MAX_RESOLUTION_HEIGHT,
               NI_MAX_RESOLUTION_AREA);
        return NI_RETCODE_FAILURE;
    } else if (ni_dec_ctx->width < min_resolution_width ||
               ni_dec_ctx->height < min_resolution_height)
    {
        ni_log(NI_LOG_ERROR, "Error XCoder resolution %dx%d not supported\n",
               ni_dec_ctx->width, ni_dec_ctx->height);
        ni_log(NI_LOG_ERROR, "Min Supported Width: %d Height %d\n",
               min_resolution_width, min_resolution_height);
        return NI_RETCODE_FAILURE;
    }

    ni_dec_ctx->offset = 0LL;

    ni_dec_ctx->draining = 0;

    p_session_ctx->pic_reorder_delay = ni_dec_ctx->has_b_frames;
    p_session_ctx->bit_depth_factor = ni_dec_ctx->bit_depth_factor;

    ni_log(NI_LOG_TRACE, "%s pix_fmt %u bit_depth_factor %u\n", __FUNCTION__,
           ni_dec_ctx->codec_format, p_session_ctx->bit_depth_factor);

    // Xcoder User Configuration
    p_param = malloc(sizeof(ni_xcoder_params_t));
    if (!p_param)
    {
        ni_log(NI_LOG_ERROR,
               "ERROR %d:%s() line %d, alloc session context failed\n",
               NI_ERRNO, __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    } else
    {
        ni_dec_ctx->api_param = p_param;
        if (ni_decoder_init_default_params(
                p_param, ni_dec_ctx->fps_num, ni_dec_ctx->fps_denom,
                ni_dec_ctx->bit_rate, ni_dec_ctx->width,
                ni_dec_ctx->height) < 0)
        {
            ni_log(NI_LOG_INFO, "Error setting params\n");

            ret = NI_RETCODE_FAILURE;
            goto end;
        }
    }

    // Init input fifo
    ni_dec_ctx->input_data_fifo = ni_fifo_initialize(
        NI_DEC_PKT_BUF_NUM_MAX, sizeof(ni_session_data_io_t));
    if (!ni_dec_ctx->input_data_fifo)
    {
        ni_log(NI_LOG_ERROR,
               "%s(): line %d initialize enc frame buffer pool failed\n",
               __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        LRETURN;
    }
    ni_log(NI_LOG_DEBUG, "xcoder_setup success\n");
end:
    if (ret != NI_RETCODE_SUCCESS)
    {
        free(p_session_ctx);
        ni_dec_ctx->api_ctx = NULL;
        free(p_param);
        ni_dec_ctx->api_param = NULL;
        ni_fifo_free(ni_dec_ctx->input_data_fifo);
        free(p_session_data);
        ni_dec_ctx->p_session_data = NULL;
        free(p_session_data1);
        ni_dec_ctx->p_session_data1 = NULL;
        free(p_session_data2);
        ni_dec_ctx->p_session_data2 = NULL;
    }

    return 0;
}

int ni_quadra_dec_init(ni_quadra_dec_context_t *ni_dec_ctx)
{
    int ret;
    ni_xcoder_params_t *p_param = NULL;
    ni_session_context_t *p_session_ctx = NULL;

    if (!ni_dec_ctx)
    {
        ni_log(NI_LOG_ERROR, "ni_dec_ctx is null\n");
        return NI_RETCODE_INVALID_PARAM;
    }
    p_param = ni_dec_ctx->api_param;
    p_session_ctx = ni_dec_ctx->api_ctx;
    if (!p_param || !p_session_ctx)
    {
        ni_log(NI_LOG_ERROR, "%s(): line %d, Invalid Parameter, %p, %p\n",
               __FUNCTION__, __LINE__, p_param, p_session_ctx);
        return NI_RETCODE_INVALID_PARAM;
    }
    // overwrite keep alive timeout value here with a custom value if it was
    // provided
    // if xcoder option is set then overwrite the (legacy) decoder option
    uint32_t xcoder_timeout = p_param->dec_input_params.keep_alive_timeout;
    if (xcoder_timeout != NI_DEFAULT_KEEP_ALIVE_TIMEOUT)
    {
        p_session_ctx->keep_alive_timeout = xcoder_timeout;
    } else
    {
        p_session_ctx->keep_alive_timeout = ni_dec_ctx->keep_alive_timeout;
    }
    ni_log(NI_LOG_DEBUG, "Custom NVME Keep Alive Timeout set to %d\n",
           p_session_ctx->keep_alive_timeout);

    if (p_param->dec_input_params.decoder_low_delay != 0)
    {
        ni_dec_ctx->low_delay = p_param->dec_input_params.decoder_low_delay;
    } else
    {
        p_param->dec_input_params.decoder_low_delay = ni_dec_ctx->low_delay;
    }
    p_session_ctx->force_low_delay = p_param->dec_input_params.force_low_delay;
    p_session_ctx->enable_low_delay_check =
        ni_dec_ctx->enable_low_delay_check = 
            p_param->dec_input_params.enable_low_delay_check;
    ni_dec_ctx->skip_extra_headers = p_param->dec_input_params.skip_extra_headers;

    if (ni_dec_ctx->has_b_frames && ni_dec_ctx->enable_low_delay_check)
    {
        // If has B frame, must set lowdelay to 0
        ni_log(NI_LOG_ERROR,"Warning: decoder lowDelay mode "
              "is cancelled due to has_b_frames with enable_low_delay_check\n");
        ni_dec_ctx->low_delay = p_param->dec_input_params.decoder_low_delay = 0;
    }
    p_session_ctx->decoder_low_delay = ni_dec_ctx->low_delay;

    if (p_param->dec_input_params.hwframes)
    {   // need to set before open decoder
        p_session_ctx->hw_action = NI_CODEC_HW_ENABLE;
        ni_dec_ctx->hwframes = 1;
    } else
    {
        p_session_ctx->hw_action = NI_CODEC_HW_NONE;
    }
    p_session_ctx->bit_depth_factor = ni_dec_ctx->bit_depth_factor;

    if (ni_dec_ctx->custom_sei_type ==
            USER_DATA_UNREGISTERED_SEI_PAYLOAD_TYPE ||
        p_param->dec_input_params.custom_sei_passthru ==
            USER_DATA_UNREGISTERED_SEI_PAYLOAD_TYPE)
    {
        // use SW passthru only
        p_session_ctx->enable_user_data_sei_passthru = 0;
    } else
    {
        p_session_ctx->enable_user_data_sei_passthru =
            ni_dec_ctx->enable_user_data_sei_passthru;
    }

    ni_log(NI_LOG_TRACE,
           "api_ctx %p api_ctx/s: user_data_sei_passthru = %d/%d, custom_sei = "
           "%d\n",
           p_session_ctx, p_session_ctx->enable_user_data_sei_passthru,
           ni_dec_ctx->enable_user_data_sei_passthru,
           ni_dec_ctx->custom_sei_type);

    ni_dec_ctx->started = 0;
    ni_dec_ctx->extradata = NULL;
    ni_dec_ctx->extradata_size = 0;
    ni_dec_ctx->pkt_nal_bitmap = 0;
    ni_dec_ctx->svct_skip_next_packet = 0;

    ni_dec_ctx->eos_pkt_received = 0;
    ni_dec_ctx->dec_wait = NI_POLL_INTERVAL;

    ni_dec_ctx->pkt_duration = 0;
    ni_dec_ctx->ui32FramesCorrupted = 0;

    p_session_ctx->p_session_config = p_param;

    p_session_ctx->hw_id = ni_dec_ctx->dev_dec_idx;
    p_session_ctx->decoder_low_delay = 0;
    ni_quadra_xcoder_strncpy(p_session_ctx->blk_dev_name,
                             ni_dec_ctx->dev_blk_name,
                             NI_MAX_DEVICE_NAME_LEN);
    ni_quadra_xcoder_strncpy(p_session_ctx->dev_xcoder_name,
                             ni_dec_ctx->dev_xcoder,
                             MAX_CHAR_IN_DEVICE_NAME);

    ret = ni_device_session_open(p_session_ctx, NI_DEVICE_TYPE_DECODER);
    if (ret != 0)
    {
        ni_log(NI_LOG_ERROR,
               "Failed to open decoder (status = %d), "
               "resource unavailable\n",
               ret);
        ret = NI_RETCODE_FAILURE;
    } else
    {
        ni_dec_ctx->dev_xcoder_name = p_session_ctx->dev_xcoder_name;
        ni_dec_ctx->blk_xcoder_name = p_session_ctx->blk_xcoder_name;
        ni_dec_ctx->dev_dec_idx = p_session_ctx->hw_id;
        ni_log(NI_LOG_DEBUG,
               "XCoder %s Index %d (inst: %d) opened successfully\n",
               ni_dec_ctx->dev_xcoder_name, ni_dec_ctx->dev_dec_idx,
               p_session_ctx->session_id);

        if (p_param->dec_input_params.hwframes)
        {
            ni_dec_ctx->hwframes = 1;
            p_session_ctx->hw_action = NI_CODEC_HW_ENABLE;
        } else
        {
            // reassign in case above conditions alter value
            p_session_ctx->hw_action = NI_CODEC_HW_NONE;
        }
    }

    ni_dec_ctx->current_pts = 0;
    ni_dec_ctx->enable_follow_iframe = p_param->dec_input_params.enable_follow_iframe;
    return ret;
}

// reset and restart when xcoder decoder resets
int ni_quadra_dec_reset(ni_quadra_dec_context_t *ni_dec_ctx)
{
    ni_session_context_t *p_session_ctx = ni_dec_ctx->api_ctx;
    ni_session_data_io_t *p_ni_packet = NULL;
    ni_retcode_t ret = NI_RETCODE_FAILURE;
    ni_log(NI_LOG_DEBUG, "XCoder decode reset\n");

    int ori_run_state = p_session_ctx->session_run_state;
    ni_device_session_close(p_session_ctx, ni_dec_ctx->eos,
                            NI_DEVICE_TYPE_DECODER);

    ni_device_session_context_clear(p_session_ctx);

#ifdef _WIN32
    ni_device_close(p_session_ctx->device_handle);
#elif __linux__
    ni_device_close(p_session_ctx->device_handle);
    ni_device_close(p_session_ctx->blk_io_handle);
#endif
    p_session_ctx->device_handle = NI_INVALID_DEVICE_HANDLE;
    p_session_ctx->blk_io_handle = NI_INVALID_DEVICE_HANDLE;
    if (ni_dec_ctx->input_data_fifo)
    {
        // push back
        do
        {
            p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_write(ni_dec_ctx->input_data_fifo);
        } while (p_ni_packet);
        // free all
        do
        {
            p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_read(ni_dec_ctx->input_data_fifo);
            if (p_ni_packet)
            {
                ni_packet_buffer_free(&(p_ni_packet->data.packet));
            }
        } while (p_ni_packet);
    }
    ni_fifo_free(ni_dec_ctx->input_data_fifo);
    ni_log(NI_LOG_DEBUG, "%s(): decoder frame fifo freed.\n", __FUNCTION__);

    int64_t bcp_current_pts = ni_dec_ctx->current_pts;
    ret = ni_quadra_dec_init(ni_dec_ctx);
    ni_dec_ctx->current_pts = bcp_current_pts;
    // p_session_ctx->session_run_state = ori_run_state;
    p_session_ctx->session_run_state = ori_run_state;

    return ret;
}

// return 1 if need to prepend saved header to pkt data, 0 otherwise
int ni_quadra_add_headers(ni_quadra_dec_context_t *ni_dec_ctx, int pkt_flag,
                          const uint8_t *data, int size, uint8_t *extradata,
                          int extradata_size)
{
    int ret = 0;
    int vps_num, sps_num, pps_num;

    // check key frame packet only
    if (!(pkt_flag & NI_PKT_FLAG_KEY) || !data || !extradata || !extradata_size)
    {
        return ret;
    }

    if (ni_dec_ctx->extradata_size == extradata_size &&
        memcmp(ni_dec_ctx->extradata, extradata, extradata_size) == 0)
    {
        ni_log(NI_LOG_TRACE, "%s extradata unchanged.\n", __FUNCTION__);
        return ret;
    }

    if (NI_CODEC_FORMAT_H264 != ni_dec_ctx->codec_format &&
        NI_CODEC_FORMAT_H265 != ni_dec_ctx->codec_format)
    {
        ni_log(NI_LOG_TRACE, "%s not AVC/HEVC codec: %d, skip!\n",
              __FUNCTION__, ni_dec_ctx->codec_format);
        return ret;
    }

    // extradata (headers) non-existing or changed: save/update it in the
    // session storage
    free(ni_dec_ctx->extradata);
    ni_dec_ctx->extradata_size = 0;
    ni_dec_ctx->got_first_key_frame = 0;
    ni_dec_ctx->extradata = malloc(extradata_size);
    if (!ni_dec_ctx->extradata)
    {
        ni_log(NI_LOG_ERROR, "%s memory allocation failed !\n", __FUNCTION__);
        return ret;
    }

    memcpy(ni_dec_ctx->extradata, extradata, extradata_size);
    ni_dec_ctx->extradata_size = extradata_size;
    // prepend header by default (assuming no header found in the pkt itself)
    ret = 1;
    // and we've got the first key frame of this stream
    ni_dec_ctx->got_first_key_frame = 1;
    vps_num = sps_num = pps_num = 0;

    if (ni_dec_ctx->skip_extra_headers &&
        (ni_dec_ctx->extradata_size > 0) &&
        ni_dec_ctx->extradata)
    {
        const uint8_t *ptr = ni_dec_ctx->extradata;
        const uint8_t *end = ni_dec_ctx->extradata + ni_dec_ctx->extradata_size;
        uint32_t stc;
        int nalu_type;
        while (ptr < end)
        {
            stc = -1;
            ptr = ni_find_start_code(ptr, end, &stc);
            if (ptr == end)
            {
                break;
            }

            if (NI_CODEC_FORMAT_H264 == ni_dec_ctx->codec_format)
            {
                nalu_type = (int)stc & 0x1f;

                if (7 /*H264_NAL_SPS */ == nalu_type)
                {
                    sps_num++;
                } else if(8 /*H264_NAL_PPS */ == nalu_type)
                {
                    pps_num++;
                }
#define H264_MAX_SPS_COUNT          32
#define H264_MAX_PPS_COUNT         256
                if (sps_num > H264_MAX_SPS_COUNT ||
                    pps_num > H264_MAX_PPS_COUNT)
                {
                    ret = 0;
                    ni_log(NI_LOG_DEBUG, "Drop extradata because of repeated SPS/PPS\n");
                    break;
                }
            } else if (NI_CODEC_FORMAT_H265 == ni_dec_ctx->codec_format)
            {
                nalu_type = (int)(stc >> 1) & 0x3F;

                if (32 /* HEVC_NAL_VPS */ == nalu_type)
                {
                    vps_num++;
                } else if (33 /* HEVC_NAL_SPS */ == nalu_type)
                {
                    sps_num++;
                } else if(34 /* HEVC_NAL_PPS */ == nalu_type)
                {
                    pps_num++;
                }

#define HEVC_MAX_VPS_COUNT  16
#define HEVC_MAX_SPS_COUNT  16
#define HEVC_MAX_PPS_COUNT  64
                if (vps_num > HEVC_MAX_VPS_COUNT ||
                    sps_num > HEVC_MAX_SPS_COUNT ||
                    pps_num > HEVC_MAX_PPS_COUNT) {
                    ret = 0;
                    ni_log(NI_LOG_DEBUG, "Drop extradata because of repeated VPS/SPS/PPS\n");
                    break;
                }

            }
        }
    }

    return ret;
}

int ni_quadra_dec_send(ni_quadra_dec_context_t *ni_dec_ctx, int pkt_flags,
                       uint8_t *data, int pkt_size,
                       uint8_t *lonedata, int lone_size, int64_t pkt_pts,
                       int64_t pkt_dts, int pkt_pos, int width, int height,
                       uint8_t *extradata, int extradata_size)
{
    /* call xcoder_dec_send to send compressed video packet to the decoder
     instance */
    ni_session_context_t *p_session_ctx = ni_dec_ctx->api_ctx;
    ni_session_data_io_t *p_ni_packet = NULL;
    int need_draining = 0;
    size_t size;
    ni_packet_t *xpkt = NULL;
    int ret;
    int sent;
    int send_size = 0;
    int new_packet = 0;
    int extra_prev_size = 0;
    int svct_skip_packet = ni_dec_ctx->svct_skip_next_packet;

    size = pkt_size;

    if (ni_dec_ctx->flushing)
    {
        ni_log(NI_LOG_ERROR,
               "Decoder is flushing and cannot accept new "
               "buffer until all output buffers have been released\n");
        return NI_RETCODE_FAILURE;
    }

    if (pkt_size == 0)
    {
        need_draining = 1;
        // Once VPU reset, the s->draining would be recovered during session reset.
        // And so is the s->eos in case that it lost in the last session read.
        ni_dec_ctx->eos |= (ni_dec_ctx->draining && ni_dec_ctx->vpu_reset);
    }

    if (ni_dec_ctx->draining && ni_dec_ctx->eos)
    {
        ni_log(NI_LOG_DEBUG, "Decoder is draining, eos\n");
        return NI_RETCODE_FAILURE;
    }
    if (!ni_dec_ctx->eos_pkt_received)
    {
        if (ni_fifo_is_full(ni_dec_ctx->input_data_fifo))
        {
            ni_log(NI_LOG_ERROR, "fifo full,need lost 1 frame....\n");
            p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_read(ni_dec_ctx->input_data_fifo);
            ni_packet_buffer_free(&(p_ni_packet->data.packet));
        }
        p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_write(ni_dec_ctx->input_data_fifo);
        xpkt = &p_ni_packet->data.packet;
        {
            memset(xpkt, 0, sizeof(ni_packet_t));
            xpkt->pts = pkt_pts;
            xpkt->dts = pkt_dts;
            xpkt->flags = pkt_flags;
            // pkt->pos does not accurately reflect the position of the current pkt
            // in the bitstream; instead we track the size of data sent to decoder
            // and use that to track the returned decoded frame and its pts.
            // xpkt->pos = pkt->pos;
            xpkt->video_width = width;
            xpkt->video_height = height;
            xpkt->p_data = NULL;
            xpkt->data_len = pkt_size;
            xpkt->pkt_pos = pkt_pos;

            if (pkt_flags & NI_PKT_FLAG_KEY && extradata_size > 0 &&
                ni_quadra_add_headers(ni_dec_ctx, pkt_flags, data, pkt_size,
                                      extradata, extradata_size))
            {
                if (extradata_size > p_session_ctx->max_nvme_io_size * 2)
                {
                    ni_log(NI_LOG_ERROR,
                           "ff_xcoder_dec_send extradata_size %d "
                           "exceeding max size supported: %d\n",
                           extradata_size, p_session_ctx->max_nvme_io_size * 2);
                } else
                {
                    ni_log(NI_LOG_TRACE,
                           "ff_xcoder_dec_send extradata_size %d "
                           "copied to pkt start.\n",
                           ni_dec_ctx->extradata_size);

                    p_session_ctx->prev_size = ni_dec_ctx->extradata_size;
                    memcpy(p_session_ctx->p_leftover, ni_dec_ctx->extradata,
                           ni_dec_ctx->extradata_size);
                }
            }

            ni_dec_ctx->svct_skip_next_packet = 0;
            // If there was lone custom sei in the last packet and the firmware would
            // fail to recognize it. So passthrough the custom sei here.
            if (lone_size > 0)
            {
                // No need to check the return value here because the lone_sei_pkt was
                // parsed before. Here it is only to extract the SEI data.
                ni_dec_packet_parse(ni_dec_ctx->api_ctx, ni_dec_ctx->api_param,lonedata,lone_size,xpkt,
                                    ni_dec_ctx->low_delay,ni_dec_ctx->codec_format,ni_dec_ctx->pkt_nal_bitmap,
                                    ni_dec_ctx->custom_sei_type,&ni_dec_ctx->svct_skip_next_packet,
                                    &ni_dec_ctx->is_lone_sei_pkt);
            }

            ret = ni_dec_packet_parse(ni_dec_ctx->api_ctx, ni_dec_ctx->api_param,data, size, xpkt,
                        ni_dec_ctx->low_delay, ni_dec_ctx->codec_format, ni_dec_ctx->pkt_nal_bitmap,
                        ni_dec_ctx->custom_sei_type, &ni_dec_ctx->svct_skip_next_packet,
                        &ni_dec_ctx->is_lone_sei_pkt);
            if (ret != 0)
            {
                goto fail;
            }

            if (svct_skip_packet)
            {
                ni_log(NI_LOG_TRACE,
                       "ff_xcoder_dec_send packet: pts:%" PRIi64 ","
                       " size:%d\n",
                       pkt_pts, pkt_size);
                xpkt->data_len = 0;
                return pkt_size;
            }
            // If the current packet is a lone SEI, save it to be sent with the next
            // packet. And also check if getting the first packet containing key frame
            // in decoder low delay mode.
            if (ni_dec_ctx->is_lone_sei_pkt)
            {
                //  av_packet_ref(&s->lone_sei_pkt, pkt);
                xpkt->data_len = 0;
                free(xpkt->p_custom_sei_set);
                xpkt->p_custom_sei_set = NULL;
                if (ni_dec_ctx->low_delay && ni_dec_ctx->got_first_key_frame &&
                    !(ni_dec_ctx->pkt_nal_bitmap &
                      NI_GENERATE_ALL_NAL_HEADER_BIT))
                {
                    // Packets before the IDR is sent cannot be decoded. So
                    // set packet num to zero here.
                    p_session_ctx->decoder_low_delay = ni_dec_ctx->low_delay;
                    p_session_ctx->pkt_num = 0;
                    ni_dec_ctx->pkt_nal_bitmap |=
                        NI_GENERATE_ALL_NAL_HEADER_BIT;
                    ni_log(NI_LOG_TRACE,
                           "%s got first IDR in decoder low delay mode, "
                           "delay time %dms, pkt_nal_bitmap %d\n",
                           __FUNCTION__, ni_dec_ctx->low_delay,
                           ni_dec_ctx->pkt_nal_bitmap);
                }
                ni_log(NI_LOG_TRACE, "%s pkt lone SEI, saved, and return %d\n",
                       __FUNCTION__, pkt_size);
                return pkt_size;
            }
            // embed lone SEI saved previously (if any) to send to decoder
            if (lone_size > 0)
            {
                ni_log(NI_LOG_TRACE, "%s copy over lone SEI data size: %d\n",
                       __FUNCTION__, lone_size);

                memcpy(p_session_ctx->p_leftover + p_session_ctx->prev_size,
                       lonedata, lone_size);
                p_session_ctx->prev_size += lone_size;
                //  av_packet_unref(&s->lone_sei_pkt);
            }

            if (ni_dec_ctx->ppu_reconfig)
            {
                ni_ppu_config_t ppu_config = {0};

                for (int idx = 0; idx < NI_MAX_NUM_OF_DECODER_OUTPUTS; idx ++)
                {
                    if (ni_dec_ctx->ni_quadra_ppu_config.out_enabled[idx] == 1)
                    {
                        ppu_config.ppu_set_enable += (0x1 << idx);
                        ppu_config.ppu_w[idx] = 
                                    ni_dec_ctx->ni_quadra_ppu_config.ppu_w[idx];
                        ppu_config.ppu_h[idx] = 
                                    ni_dec_ctx->ni_quadra_ppu_config.ppu_h[idx];
                    }
                }
                ret = ni_reconfig_ppu_output(p_session_ctx, 
                        (ni_xcoder_params_t *)ni_dec_ctx->api_param, &ppu_config);
                if (ret < 0)
                {
                    ni_log(NI_LOG_ERROR, "ni_reconfig_ppu_output failed, ret = %d\n", ret);
                    goto fail;
                }
            }

            if ((pkt_size + p_session_ctx->prev_size) > 0)
            {
                ni_packet_buffer_alloc(xpkt,
                                       (pkt_size + p_session_ctx->prev_size));
                if (!xpkt->p_data)
                {
                    ret = NI_RETCODE_ERROR_MEM_ALOC;
                    goto fail;
                }
            }
            new_packet = 1;
        }

        ni_log(NI_LOG_DEBUG, "ni_quadra_dec_send: pkt size=%d dts=%lld\n",
               pkt_size, pkt_dts);
        if (ni_dec_ctx->started == 0)
        {
            xpkt->start_of_stream = 1;
            ni_dec_ctx->started = 1;
        }

        ni_log(NI_LOG_TRACE,
               "ni_packet_copy before: size=%d, s->prev_size=%d, send_size=%d, "
               "(end of stream)\n",
               pkt_size, p_session_ctx->prev_size, send_size);
        if (new_packet)
        {
            extra_prev_size = p_session_ctx->prev_size;
            send_size = ni_packet_copy(xpkt->p_data, data, pkt_size,
                                       p_session_ctx->p_leftover,
                                       &p_session_ctx->prev_size);
            // increment offset of data sent to decoder and save it
            xpkt->pos = (long long)ni_dec_ctx->offset;
            ni_dec_ctx->offset += pkt_size + extra_prev_size;
        }
        ni_log(NI_LOG_TRACE,
               "ni_packet_copy after: size=%d, s->prev_size=%d, send_size=%d, "
               "xpkt->data_len=%d (end of stream)\n",
               pkt_size, p_session_ctx->prev_size, send_size, xpkt->data_len);
        if (send_size < 0)
        {
            ni_log(NI_LOG_ERROR, "Failed to copy pkt (status = %d)\n",
                   send_size);
            ret = NI_RETCODE_FAILURE;
            goto fail;
        }

        xpkt->data_len += extra_prev_size;

        if (need_draining && !ni_dec_ctx->draining)
        {
            ni_log(NI_LOG_DEBUG, "Sending End Of Stream signal\n");
            xpkt->end_of_stream = 1;
            xpkt->data_len = 0;
            ni_dec_ctx->eos_pkt_received = 1;
        }
    }
    if (ni_dec_ctx->eos_pkt_received && !ni_dec_ctx->draining)
    {
        sent = 1;
        while (!ni_fifo_is_empty(ni_dec_ctx->input_data_fifo))
        {
            p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_read(ni_dec_ctx->input_data_fifo);
            ni_log(NI_LOG_TRACE, "%s pts=%" PRIi64 ", dts=%" PRIi64 "\n",
                   __FUNCTION__, p_ni_packet->data.packet.pts,
                   p_ni_packet->data.packet.dts);

            sent = ni_device_session_write(p_session_ctx,
                                           p_ni_packet,
                                           NI_DEVICE_TYPE_DECODER);

            if (sent == 0 && !ni_fifo_is_empty(ni_dec_ctx->input_data_fifo))
            {
                ni_log(NI_LOG_DEBUG,
                       "dec_send EAGAIN this better not happen "
                       "sleep %d us !\n",
                       ni_dec_ctx->dec_wait);
                usleep(ni_dec_ctx->dec_wait);
                ni_dec_ctx->dec_wait += NI_POLL_INTERVAL;
                ni_fifo_return_read(ni_dec_ctx->input_data_fifo);
                break;
            } else if (sent < 0)
            {
                ni_log(NI_LOG_ERROR,
                       "Failed to send eos signal (status = %d)\n", sent);
                if (NI_RETCODE_ERROR_VPU_RECOVERY == sent)
                {
                    ni_log(NI_LOG_INFO, "%s VPU recovery, need to reset\n",
                           __FUNCTION__);
                    p_session_ctx->session_run_state =
                        SESSION_RUN_STATE_RESETTING;
                    if (ni_quadra_dec_reset(ni_dec_ctx) < 0)
                    {
                        return NI_RETCODE_FAILURE;
                    }
                    return 0;
                }
                ret = NI_RETCODE_FAILURE;
                goto fail;
            } else
            {
                ni_packet_buffer_free(&(p_ni_packet->data.packet));
                if ((p_ni_packet->data.packet).data_len == 0)
                {
                    free((p_ni_packet->data.packet).p_custom_sei_set);
                    (p_ni_packet->data.packet).p_custom_sei_set = NULL;
                }
                // wait time is incremental, and reset after a successful send
                ni_dec_ctx->dec_wait = NI_POLL_INTERVAL;
            }
        }
        if (ni_fifo_is_empty(ni_dec_ctx->input_data_fifo))
        {
            ni_device_session_flush(p_session_ctx, NI_DEVICE_TYPE_DECODER);
            ni_dec_ctx->draining = 1;
            p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
            ni_dec_ctx->eos_pkt_received = 0;
        }
    } else
    {
        p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_read(ni_dec_ctx->input_data_fifo);
        ni_log(NI_LOG_DEBUG, "%s pts=%" PRIi64 ", dts=%" PRIi64 " \n",
               __FUNCTION__, p_ni_packet->data.packet.pts,
               p_ni_packet->data.packet.dts);

        sent = ni_device_session_write(
            p_session_ctx, p_ni_packet, NI_DEVICE_TYPE_DECODER);
        if (sent < 0)
        {
            ni_log(NI_LOG_ERROR, "Failed to send compressed pkt (status=%d)\n",
                   sent);
            if (NI_RETCODE_ERROR_VPU_RECOVERY == sent)
            {
                ni_log(NI_LOG_INFO, "%s VPU recovery, need to reset\n",
                       __FUNCTION__);
                p_session_ctx->session_run_state = SESSION_RUN_STATE_RESETTING;
                if (ni_quadra_dec_reset(ni_dec_ctx) < 0)
                {
                    return NI_RETCODE_FAILURE;
                }
                return 0;
            }
            ret = NI_RETCODE_FAILURE;
            goto fail;
        } else if (sent == 0)
        {
            ni_fifo_return_read(ni_dec_ctx->input_data_fifo);
            sent = size;
        } else
        {
            if (p_session_ctx->session_run_state == SESSION_RUN_STATE_RESETTING)
            {
                p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
            }
            ni_packet_buffer_free(&(p_ni_packet->data.packet));
            if ((p_ni_packet->data.packet).data_len == 0)
            {
                free((p_ni_packet->data.packet).p_custom_sei_set);
                (p_ni_packet->data.packet).p_custom_sei_set = NULL;
            }
        }
    }
    return sent;

fail:
    ni_packet_buffer_free(xpkt);
    if (xpkt)
    {
        free(xpkt->p_custom_sei_set);
        xpkt->p_custom_sei_set = NULL;
    }
    ni_dec_ctx->draining = 1;
    ni_dec_ctx->eos = 1;

    return ret;
}

void ni_quadra_dec_aux_data(ni_quadra_dec_context_t *ni_dec_ctx)
{
    ni_session_data_io_t *p_session_data = ni_dec_ctx->p_session_data;
    ni_frame_t *nifme = &p_session_data->data.frame;
    ni_dec_retrieve_aux_data(nifme);
}

void ni_quadra_dec_wipe_aux_data(ni_quadra_dec_context_t *ni_dec_ctx)
{
    ni_session_data_io_t *p_session_data = ni_dec_ctx->p_session_data;
    ni_frame_t *nifme = &p_session_data->data.frame;
    ni_frame_wipe_aux_data(nifme);
}

int ni_quadra_frame_get_aux_size(ni_quadra_dec_context_t *ni_dec_ctx, int type)
{
    ni_session_data_io_t *p_session_data = ni_dec_ctx->p_session_data;
    ni_frame_t *nifme = &p_session_data->data.frame;
    ni_aux_data_t *aux = NULL;
    aux = ni_frame_get_aux_data(nifme, type);
    if (aux)
        return aux->size;
    return 0;
}

void ni_quadra_frame_copy_aux_data(ni_quadra_dec_context_t *ni_dec_ctx,
                                   uint8_t *dst_data, int type, int size)
{
    ni_session_data_io_t *p_session_data = ni_dec_ctx->p_session_data;
    ni_frame_t *nifme = &p_session_data->data.frame;
    ni_aux_data_t *aux = NULL;
    aux = ni_frame_get_aux_data(nifme, type);
    memcpy(dst_data, aux->data, size);
}

void ni_quadra_frame_sync(ni_quadra_dec_context_t *ni_dec_ctx)
{
    ni_session_data_io_t *p_session_data = ni_dec_ctx->p_session_data;
    ni_frame_t *xfme = &p_session_data->data.frame;
    ni_log(NI_LOG_DEBUG,
           "Got output buffer pts=%lld "
           "dts=%lld eos=%d sos=%d\n",
           p_session_data->data.frame.pts, p_session_data->data.frame.dts,
           p_session_data->data.frame.end_of_stream,
           p_session_data->data.frame.start_of_stream);

    ni_dec_ctx->eos = xfme->end_of_stream;
    ni_dec_ctx->niframe.pts = xfme->pts;
    ni_dec_ctx->niframe.dts = xfme->dts;
    ni_dec_ctx->niframe.end_of_stream = xfme->end_of_stream;
    ni_dec_ctx->niframe.start_of_stream = xfme->start_of_stream;
    ni_dec_ctx->niframe.video_width = xfme->video_width;
    ni_dec_ctx->niframe.video_height = xfme->video_height;
    ni_dec_ctx->niframe.pkt_pos = xfme->pkt_pos;
    ni_dec_ctx->niframe.p_data[0] = xfme->p_data[0];
    ni_dec_ctx->niframe.data_len[0] = xfme->data_len[0];
    ni_dec_ctx->niframe.data_len[1] = xfme->data_len[1];
    ni_dec_ctx->niframe.data_len[2] = xfme->data_len[2];
    ni_dec_ctx->niframe.data_len[3] = xfme->data_len[3];
    ni_dec_ctx->niframe.ni_pict_type = xfme->ni_pict_type;
    ni_dec_ctx->niframe.vui_time_scale = xfme->vui_time_scale;
    ni_dec_ctx->niframe.vui_num_units_in_tick = xfme->vui_num_units_in_tick;
    ni_dec_ctx->niframe.sei_user_data_unreg_offset =
        xfme->sei_user_data_unreg_offset;
    ni_dec_ctx->niframe.sei_user_data_unreg_len = xfme->sei_user_data_unreg_len;
    ni_dec_ctx->niframe.sei_hdr_mastering_display_color_vol_offset =
        xfme->sei_hdr_mastering_display_color_vol_offset;
    ni_dec_ctx->niframe.sei_hdr_mastering_display_color_vol_len =
        xfme->sei_hdr_mastering_display_color_vol_len;
    ni_dec_ctx->niframe.sei_hdr_content_light_level_info_offset =
        xfme->sei_hdr_content_light_level_info_offset;
    ni_dec_ctx->niframe.sei_hdr_content_light_level_info_len =
        xfme->sei_hdr_content_light_level_info_len;
    ni_dec_ctx->niframe.sei_cc_offset = xfme->sei_cc_offset;
    ni_dec_ctx->niframe.sei_cc_len = xfme->sei_cc_len;
    ni_dec_ctx->niframe.p_custom_sei_set = xfme->p_custom_sei_set;
    ni_dec_ctx->niframe.p_buffer = xfme->p_buffer;
    ni_dec_ctx->niframe.crop_left = xfme->crop_left;
    ni_dec_ctx->niframe.crop_right = xfme->crop_right;
    ni_dec_ctx->niframe.crop_top = xfme->crop_top;
    ni_dec_ctx->niframe.crop_bottom = xfme->crop_bottom;
    ni_dec_ctx->niframe.dec_buf = xfme->dec_buf;
    ni_dec_ctx->niframe.color_primaries = xfme->color_primaries;
    ni_dec_ctx->niframe.color_trc = xfme->color_trc;
    ni_dec_ctx->niframe.color_space = xfme->color_space;
    ni_dec_ctx->niframe.video_full_range_flag = xfme->video_full_range_flag;
    ni_dec_ctx->niframe.vui_len = xfme->vui_len;
    ni_dec_ctx->niframe.orignal_pts = xfme->orignal_pts;
    ni_dec_ctx->niframe.error_ratio = xfme->error_ratio;
}

int ni_quadra_dec_receive(ni_quadra_dec_context_t *ni_dec_ctx, int avctx_width,
                          int avctx_height, int frame_planar)
{
    int ret;
    int retry_cnt = 0;
    int retry_read_cnt = 0;
    int fifo_flush_cnt = 0;
    uint64_t t0 = 0;
    int height, actual_width, alloc_mem;
    ni_xcoder_params_t *p_param = ni_dec_ctx->api_param;
    ni_session_context_t *p_session_ctx = ni_dec_ctx->api_ctx;
    ni_session_data_io_t *p_session_data = ni_dec_ctx->p_session_data;
    ni_dec_ctx->bSequenceChange = 0;
    ni_dec_ctx->pkt_pts_unchange = p_param->dec_input_params.pkt_pts_unchange;
    memset(p_session_data, 0, sizeof(ni_session_data_io_t));
    if (p_session_ctx->frame_num % 2 == 0)
    {
        p_session_ctx->burst_control =
            (p_session_ctx->burst_control == 0 ? 1 : 0);   // toggle
    }
    if (p_session_ctx->burst_control)
    {
        ni_log(NI_LOG_DEBUG,
               "ff_xcoder_dec_receive burst return%" PRId64 " frame\n",
               p_session_ctx->frame_num);
        return 0;
    }
    t0 = ni_gettime_ns();

to_malloc:
    // if active video resolution has been obtained we just use it as it's the
    // exact size of frame to be returned, otherwise we use what we are told by
    // upper stream as the initial setting and it will be adjusted.
    height = (int)(p_session_ctx->active_video_height > 0 ?
                       p_session_ctx->active_video_height :
                       avctx_height);
    actual_width = (int)(p_session_ctx->actual_video_width > 0 ?
                             p_session_ctx->actual_video_width :
                             avctx_width);
    // allocate memory only after resolution is known (buffer pool set up)
    alloc_mem = (p_session_ctx->active_video_width > 0 &&
                         p_session_ctx->active_video_height > 0 ?
                     1 :
                     0);

    if (ni_dec_ctx->hwframes != 1)
    {
        ret = ni_decoder_frame_buffer_alloc(
            p_session_ctx->dec_fme_buf_pool, &(p_session_data->data.frame),
            alloc_mem, actual_width, height,
            (ni_dec_ctx->codec_format == NI_CODEC_FORMAT_H264),
            p_session_ctx->bit_depth_factor, frame_planar);
    } else
    {
        ret = ni_frame_buffer_alloc(
            &(p_session_data->data.frame), actual_width, height,
            (ni_dec_ctx->codec_format == NI_CODEC_FORMAT_H264), 1,
            p_session_ctx->bit_depth_factor, 3, frame_planar);
    }
    if (NI_RETCODE_SUCCESS != ret)
    {
        return NI_RETCODE_FAILURE;
    }

to_read:
    if (ni_dec_ctx->hwframes != 1)
    {
        ret = ni_device_session_read(p_session_ctx, p_session_data,
                                     NI_DEVICE_TYPE_DECODER);
    } else
    {
        ret = ni_device_session_read_hwdesc(p_session_ctx, p_session_data,
                                            NI_DEVICE_TYPE_DECODER);
    }
    if (ret == 0)
    {
        ni_dec_ctx->eos = p_session_data->data.frame.end_of_stream;

        uint64_t diff = p_session_ctx->pkt_num - p_session_ctx->frame_num - p_session_ctx->session_statistic.ui32FramesDropped;
        while (!ni_dec_ctx->eos && p_param->dec_input_params.min_packets_delay &&
                diff < p_session_ctx->pkt_delay_cnt &&
                diff + ni_fifo_number_of_buffers(ni_dec_ctx->input_data_fifo) >= p_session_ctx->pkt_delay_cnt)
        {
            ni_session_data_io_t *p_ni_packet = NULL;
            p_ni_packet = (ni_session_data_io_t *)ni_fifo_generic_read(ni_dec_ctx->input_data_fifo);
            ni_log(NI_LOG_TRACE, "%s pts=%" PRIi64 ", dts=%" PRIi64 "\n",
                    __FUNCTION__, p_ni_packet->data.packet.pts,
                    p_ni_packet->data.packet.dts);
            int sent = ni_device_session_write(p_session_ctx, p_ni_packet, NI_DEVICE_TYPE_DECODER);
            if (sent > 0)
            {
                if (p_session_ctx->session_run_state == SESSION_RUN_STATE_RESETTING)
                {
                    p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
                }
                ni_packet_buffer_free(&(p_ni_packet->data.packet));
                if ((p_ni_packet->data.packet).data_len == 0)
                {
                    free((p_ni_packet->data.packet).p_custom_sei_set);
                    (p_ni_packet->data.packet).p_custom_sei_set = NULL;
                }
                diff = p_session_ctx->pkt_num - p_session_ctx->frame_num - p_session_ctx->session_statistic.ui32FramesDropped;
                fifo_flush_cnt++;
            } else if (sent == 0) {
                ni_fifo_return_read(ni_dec_ctx->input_data_fifo);
                break;
            } else {
                ni_fifo_return_read(ni_dec_ctx->input_data_fifo);
                ni_log(NI_LOG_ERROR, "Failed to send fifo's compressed pkt (status=%d)\n", sent);
                ret = sent;
                if (ret == NI_RETCODE_ERROR_VPU_RECOVERY)
                {
                    ni_log(NI_LOG_INFO, "%s VPU recovery, need to reset\n",
                        __FUNCTION__);
                    p_session_ctx->session_run_state = SESSION_RUN_STATE_RESETTING;
                    ni_decoder_frame_buffer_free(&(p_session_data->data.frame));

                    ret = ni_quadra_dec_reset(ni_dec_ctx);
                    if (ret == 0)
                    {
                        ret = NI_QUADRA_ERROR_VPU_RECOVERY;
                    }
                    return ret;
                } else if (ret == NI_RETCODE_ERROR_INVALID_SESSION ||
                           ret == NI_RETCODE_ERROR_NVME_CMD_FAILED)
                {
                    return NI_QUADRA_ERROR_INVALID_SESSION;
                }
                return NI_RETCODE_FAILURE;
            }
        }

        if (fifo_flush_cnt > 0 &&
            retry_read_cnt < 1 &&
            diff >= p_session_ctx->pkt_delay_cnt &&
            p_session_ctx->pkt_delay_cnt > 0)
        {
            fifo_flush_cnt = 0;
            retry_read_cnt++;
            ni_log(NI_LOG_DEBUG, "Re-try to read frame.\n");
            goto to_read;
        }

        if (!ni_dec_ctx->eos && (p_session_ctx->pkt_num - p_session_ctx->frame_num >
             ni_dec_ctx->num_reorder_pics + p_session_ctx->session_statistic.ui32FramesDropped)) {
            if (retry_cnt < 200 && (ni_gettime_ns() - t0) < 40000000LL) // 40ms timeout
            {
                retry_cnt++;
                if (ni_dec_ctx->hwframes != 1) {
                    if (p_session_data->data.frame.data_len[0] == 0) {
                        ni_decoder_frame_buffer_free(&(p_session_data->data.frame));
                        goto to_malloc;
                    }
                } else {
                    if (p_session_data->data.frame.data_len[3] == 0) {
                        ni_frame_buffer_free(&(p_session_data->data.frame));
                        goto to_malloc;
                    }
                }
                ni_usleep(200);  // sleep 200us
                goto to_read;
            }
            else
            {
                ni_dec_ctx->num_reorder_pics++;
                ni_log(NI_LOG_INFO, "Warning: timeout, increaing num_reorder_pics to %d\n",
                       ni_dec_ctx->num_reorder_pics);
            }
        }

        if (ni_dec_ctx->hwframes != 1)
        {
            ni_decoder_frame_buffer_free(&(p_session_data->data.frame));
        } else
        {
            ni_frame_buffer_free(&(p_session_data->data.frame));
        }
    } else if (ret > 0)
    {
        if (p_session_ctx->session_run_state == SESSION_RUN_STATE_RESETTING)
        {
            // On decoder VPU recovery the first received frame corresponding to the
            // cached seq_hdr_pkt should be dropped since the data is outdated.
            // p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
            p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
            if (ni_dec_ctx->hwframes != 1)
            {
                ni_decoder_frame_buffer_free(&(p_session_data->data.frame));
            } else
            {
                ni_frame_buffer_free(&(p_session_data->data.frame));
            }
            return 0;
        }
        // retrieve side data if available
        ni_quadra_dec_aux_data(ni_dec_ctx);
        ni_quadra_frame_sync(ni_dec_ctx);
        if (ni_dec_ctx->niframe.video_width != avctx_width ||
            ni_dec_ctx->niframe.video_height != avctx_height)
        {
            ni_log(NI_LOG_INFO,
                   "ff_xcoder_dec_receive: resolution "
                   "changed: %dx%d to %dx%d\n",
                   avctx_width, avctx_height, ni_dec_ctx->niframe.video_width,
                   ni_dec_ctx->niframe.video_height);
            ni_dec_ctx->bSequenceChange = 1;
            // set num_reorder_pics to default 3 when sequence change
            ni_dec_ctx->num_reorder_pics = 3;
        }
        if (p_session_ctx->frame_num == 1)
        {
            ni_log(NI_LOG_DEBUG, "NI:%s:out\n",
                   (frame_planar == 0) ?
                       "semiplanar" :
                       (frame_planar == 2) ? "tiled" : "planar");
        }
        ni_dec_ctx->pixel_format = p_session_ctx->pixel_format;
        ni_dec_ctx->num_extra_outputs =
            (p_param->dec_input_params.enable_out1 > 0) +
            (p_param->dec_input_params.enable_out2 > 0);
        ni_dec_ctx->device_handle = p_session_ctx->device_handle;
        ni_dec_ctx->active_video_width = p_session_ctx->active_video_width;
        ni_dec_ctx->active_video_height = p_session_ctx->active_video_height;
        ni_dec_ctx->bit_depth_factor = p_session_ctx->bit_depth_factor;
        ni_dec_ctx->ready_to_close = p_session_ctx->ready_to_close;
        memset(ni_dec_ctx->p_session_data1, 0, sizeof(ni_session_data_io_t));
        memset(ni_dec_ctx->p_session_data2, 0, sizeof(ni_session_data_io_t));
        if (ni_dec_ctx->num_extra_outputs)
        {
            ni_session_data_io_t *p_session_data1 = ni_dec_ctx->p_session_data1;
            ni_session_data_io_t *p_session_data2 = ni_dec_ctx->p_session_data2;
            // width height does not matter//codec id does not matter//no metadata
            ni_frame_buffer_alloc(&(p_session_data1->data.frame), 1, 1, 1, 0, 1,
                                  1, frame_planar);
            // buf1 = p_session_data1->data.frame.p_data[0];
            if (ni_dec_ctx->num_extra_outputs > 1)
            {
                // width height does not matter
                ni_frame_buffer_alloc(&(p_session_data2->data.frame), 1, 1, 1,
                                      0, 1, 1, frame_planar);
                // buf2 = p_session_data2->data.frame.p_data[0];
            }
        }
    } else if (ret < 0)
    {
        ni_log(NI_LOG_ERROR, "Failed to get output buffer (status = %d)\n",
               ret);
        if (NI_RETCODE_ERROR_VPU_RECOVERY == ret)
        {
            ni_log(NI_LOG_INFO,
                   "ni_quadra_dec_receive VPU recovery, need to reset ..\n");
            // p_session_ctx->session_run_state = SESSION_RUN_STATE_RESETTING;
            p_session_ctx->session_run_state = SESSION_RUN_STATE_RESETTING;
            ni_decoder_frame_buffer_free(&(p_session_data->data.frame));

            ret = ni_quadra_dec_reset(ni_dec_ctx);
            if (0 == ret)
            {
                ret = NI_QUADRA_ERROR_VPU_RECOVERY;
            }
            return ret;
        } else if (ret == NI_RETCODE_ERROR_INVALID_SESSION ||
                   ret == NI_RETCODE_ERROR_NVME_CMD_FAILED)
        {
            return NI_QUADRA_ERROR_INVALID_SESSION;
        }
        return NI_RETCODE_FAILURE;
    }

#ifdef MEASURE_LATENCY
    ni_log(NI_LOG_INFO, "DEC pkt_num %d, fme_num %d, fifo_pkt %d, latency is %d\n",
                        p_session_ctx->pkt_num, p_session_ctx->frame_num, ni_dec_ctx->input_data_fifo->number_of_buffers_used,
                        (p_session_ctx->pkt_num + ni_dec_ctx->input_data_fifo->number_of_buffers_used - p_session_ctx->frame_num));
#endif
    return ret;
}

int ni_quadra_dec_flush(ni_quadra_dec_context_t *ni_dec_ctx)
{
    ni_dec_ctx->draining = 0;
    ni_dec_ctx->flushing = 0;
    ni_dec_ctx->eos = 0;
    /* Future: for now, always return 1 to indicate the codec has been flushed
     and it leaves the flushing state and can process again ! will consider
     case of user retaining frames in HW "surface" usage */
    return 1;
}

int ni_quadra_dec_is_flushing(ni_quadra_dec_context_t *ni_dec_ctx)
{
    int ret = 1;
    if (ni_dec_ctx->flushing)
    {
        if (!ni_quadra_dec_flush(ni_dec_ctx))
        {
            ni_log(NI_LOG_DEBUG, "IS flushing and NOT dec_flush, AGAIN!!!\n");
            ret = 0;
        }
    }
    return ret;
}

int ni_quadra_dec_params_set_value(ni_quadra_dec_context_t *ni_dec_ctx,
                                   const char *name, char *value)
{
    ni_xcoder_params_t *p_param = ni_dec_ctx->api_param;
    int parse_ret = ni_decoder_params_set_value(p_param, name, value);
    if (parse_ret != NI_RETCODE_SUCCESS)
    {
        switch (parse_ret)
        {
            case NI_RETCODE_PARAM_INVALID_NAME:
                ni_log(NI_LOG_ERROR, "Unknown option: %s.\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_TOO_BIG:
                ni_log(NI_LOG_ERROR, "Invalid %s: too big, max char len = %d\n",
                       name, NI_MAX_PPU_PARAM_EXPR_CHAR);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_TOO_SMALL:
                ni_log(NI_LOG_ERROR, "Invalid %s: too small\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_OOR:
                ni_log(NI_LOG_ERROR, "Invalid %s: out of range\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_ZERO:
                ni_log(NI_LOG_ERROR, "Error setting option %s to value 0\n",
                       name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_INVALID_VALUE:
                ni_log(NI_LOG_ERROR, "Invalid value for %s: %s.\n", name,
                       value);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_WARNING_DEPRECATED:
                ni_log(NI_LOG_INFO, "Parameter %s is deprecated\n", name);
                parse_ret = NI_QUADRA_PARAM_WARNING_DEPRECATED;
                break;
            default:
                ni_log(NI_LOG_ERROR, "Invalid %s: ret %d\n", name, parse_ret);
                return NI_RETCODE_FAILURE;
        }
    }
    return parse_ret;
}

int ni_quadra_symbolic_decoder_param(ni_quadra_dec_context_t *ni_dec_ctx,
                                     int type)
{
    int i, j;
    ni_xcoder_params_t *p_param = ni_dec_ctx->api_param;
    ni_decoder_input_params_t *pdec_param = &p_param->dec_input_params;
    if (!pdec_param)
    {
        return NI_RETCODE_FAILURE;
    }

    if (type == 0)   // sync ni_dec_input_params_t to ni_decoder_input_params_t
    {
        memcpy(ni_dec_ctx->niparam.cr_expr, pdec_param->cr_expr,sizeof(pdec_param->cr_expr));
        memcpy(ni_dec_ctx->niparam.sc_expr, pdec_param->sc_expr,sizeof(pdec_param->sc_expr));
        memcpy(ni_dec_ctx->niparam.crop_whxy, pdec_param->crop_whxy,sizeof(pdec_param->crop_whxy));
    } else if (type ==
               1)   // sync ni_decoder_input_params_t to ni_xcoder_params_t
    {
        for (i = 0; i < NI_MAX_NUM_OF_DECODER_OUTPUTS; i++)
        {
            if (pdec_param->cr_expr[i][0][0] && pdec_param->cr_expr[i][1][0])
            {
                pdec_param->crop_whxy[i][0] =
                    (int)ni_dec_ctx->niparam.crop_whxy[i][0];
                pdec_param->crop_whxy[i][1] =
                    (int)ni_dec_ctx->niparam.crop_whxy[i][1];
            }
            if (pdec_param->cr_expr[i][2][0])
            {
                pdec_param->crop_whxy[i][2] =
                    (int)(ni_dec_ctx->niparam.crop_whxy[i][2]);
            }
            if (pdec_param->cr_expr[i][3][0])
            {
                pdec_param->crop_whxy[i][3] =
                    (int)(ni_dec_ctx->niparam.crop_whxy[i][3]);
            }
            if (pdec_param->sc_expr[i][0][0] && pdec_param->sc_expr[i][1][0])
            {
                pdec_param->scale_wh[i][0] = ni_dec_ctx->niparam.scale_wh[i][0];
                pdec_param->scale_wh[i][1] = ni_dec_ctx->niparam.scale_wh[i][1];
            }
        }
    } else if (type == 2)
    {
        for (i = 0; i < NI_MAX_NUM_OF_DECODER_OUTPUTS; i++)
        {
            if (p_param->dec_input_params.crop_mode[i] != NI_DEC_CROP_MODE_AUTO)
            {
                continue;
            }
            for (j = 0; j < 4; j++)
            {
                if (strlen(p_param->dec_input_params.cr_expr[i][j]))
                {
                    ni_log(NI_LOG_ERROR,
                           "Setting crop parameters without "
                           "setting crop mode to manual?\n");
                    return NI_RETCODE_FAILURE;
                }
            }
        }
        ni_dec_ctx->niparam.semi_planar[0] =
            p_param->dec_input_params.semi_planar[0];
        ni_dec_ctx->niparam.force_8_bit[0] =
            p_param->dec_input_params.force_8_bit[0];
    }
    return 0;
}

uint8_t *ni_quadra_dec_frame_buffer_alloc(ni_quadra_dec_context_t *ni_dec_ctx,
                                          void *p_session_data)
{
    ni_session_data_io_t *p_session_data1 = p_session_data;
    return p_session_data1->data.frame.p_data[0];
}
