/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
*  \file   ni_quadra_dec_api.h
*
*  \brief  NETINT decoder API header file
*
*******************************************************************************/
#pragma once
#ifdef __cplusplus
extern "C" {
#endif
#include <errno.h>
#include <stddef.h>
#include <unistd.h>
// #include "ni_defs.h"
// #include "ni_rsrc_api.h"
// #include "ni_util.h"
// #include "ni_device_api.h"
// #include "ni_av_codec.h"
// #include "ni_bitstream.h"
#include "ni_quadra_filter_api.h"

#define NI_NAL_VPS_BIT (0x01)
#define NI_NAL_SPS_BIT (0x01 << 1)
#define NI_NAL_PPS_BIT (0x01 << 2)
#define NI_GENERATE_ALL_NAL_HEADER_BIT (0x01 << 3)
#define NI_PKT_FLAG_KEY 0x0001       ///< The packet contains a keyframe
#define NI_PKT_FLAG_CORRUPT 0x0002   ///< The packet content is corrupted

/*!*
* \brief Session running state type.
*/
typedef enum _ni_quadra_session_run_state
{
    QUADRA_SESSION_RUN_STATE_NORMAL = 0,
    QUADRA_SESSION_RUN_STATE_SEQ_CHANGE_DRAINING = 1,
    QUADRA_SESSION_RUN_STATE_SEQ_CHANGE_OPENING = 2,
    QUADRA_SESSION_RUN_STATE_RESETTING = 3,
} ni_quadra_session_run_state_t;

typedef struct _ni_quadra_frame
{
    // codec of the source from which this frame is decoded
    int src_codec;

    long long dts;
    long long pts;
    uint32_t end_of_stream;
    uint32_t start_of_stream;
    uint32_t video_width;
    uint32_t video_height;

    uint32_t crop_top;
    uint32_t crop_bottom;
    uint32_t crop_left;
    uint32_t crop_right;

    // for decoding: frame picture type returned from decode
    int ni_pict_type;

    // SEI info of closed caption: returned by decoder or set by encoder
    unsigned int sei_cc_offset;
    unsigned int sei_cc_len;

    // SEI info of User Data Unregistered SEI: returned by decoder
    unsigned int sei_user_data_unreg_offset;
    unsigned int sei_user_data_unreg_len;
    unsigned int sei_hdr_mastering_display_color_vol_offset;
    unsigned int sei_hdr_mastering_display_color_vol_len;
    unsigned int sei_hdr_content_light_level_info_offset;
    unsigned int sei_hdr_content_light_level_info_len;

    uint8_t *p_data[NI_QUADRA_MAX_NUM_DATA_POINTERS];
    uint32_t data_len[NI_QUADRA_MAX_NUM_DATA_POINTERS];

    uint8_t *p_buffer;

    void *dec_buf;   // buffer pool entry (has memory pointed to by p_buffer)

    int pixel_format;

    void *p_custom_sei_set;

    // frame auxiliary data
    void *aux_data[NI_QUADRA_MAX_NUM_AUX_DATA_PER_FRAME];
    uint32_t vui_num_units_in_tick;
    uint32_t vui_time_scale;
    uint64_t pkt_pos;
    uint8_t color_primaries;
    uint8_t color_trc;
    uint8_t color_space;
    uint8_t video_full_range_flag;
    unsigned int vui_len;
    long long orignal_pts;
    uint32_t error_ratio;
} ni_quadra_frame_t;

typedef struct _ni_dec_input_params_t
{
    int force_8_bit[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS];
    int semi_planar[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS];
    int crop_mode[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS];
    int crop_whxy[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS][4];
    char cr_expr[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS][4]
                [NI_QUADRA_MAX_PPU_PARAM_EXPR_CHAR +
                 1]; /*cut-out of expression to be parsed*/
    int scale_wh[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS][2];
    char sc_expr[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS][2]
                [NI_QUADRA_MAX_PPU_PARAM_EXPR_CHAR +
                 1]; /*cut-out of expression to be parsed*/
} ni_dec_input_params_t;

typedef struct _ni_quadra_ppu_config
{
    uint16_t out_enabled[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS];
    uint16_t ppu_w[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS];
    uint16_t ppu_h[NI_QUADRA_MAX_NUM_OF_DECODER_OUTPUTS];
} ni_quadra_ppu_config_t;

typedef struct _ni_quadra_dec_context_t
{
    /* from the command line, which resource allocation method we use */
    char *dev_xcoder;
    char *dev_xcoder_name; /* dev name of the xcoder card to use */
    char *blk_xcoder_name; /* blk name of the xcoder card to use */
    int dev_dec_idx;       /* user-specified decoder index */
    char *dev_blk_name;    /* user-specified decoder block device name */
    int blk_io_handle;
    int device_handle;
    int keep_alive_timeout; /* keep alive timeout setting */

    int eos_pkt_received;
    int dec_wait;   // in us

    // stream header copied/saved from AVCodecContext.extradata
    int got_first_key_frame;
    uint8_t *extradata;
    int extradata_size;

    int64_t current_pts;
    /* offset of the cur. pkt relative to the start of the bit stream. used to */
    /* track the returned decoded frame and retrieve its saved pts */
    unsigned long long offset;
    int svct_skip_next_packet;

    int started;
    int draining;
    int flushing;
    int eos;
    int vpu_reset;
    int is_lone_sei_pkt;
    int bSequenceChange;
    int64_t pkt_duration;
    int num_extra_outputs;
    /* below are all command line options */
    char *xcoder_opts;
    int enable_user_data_sei_passthru;
    int enable_check_packet;   // check source packet. Skip SEI payloads after VCL
    int custom_sei_type;
    int low_delay;
    int force_low_delay;
    int enable_low_delay_check;
    int pkt_nal_bitmap;
    int skip_extra_headers;

    /* below are sync up from ffmpeg */
    int set_high_priority; /* set_high_priority */
    int ff_log_level;      /* ffmpeg log level */

    int fps_num;
    int fps_denom;
    long bit_rate;
    int width;
    int height;
    int codec_format;
    int bit_depth_factor;
    int has_b_frames;
    int hwframes;
    int pixel_format;
    int num_reorder_pics;

    //Current video width. this is used to do sequence change
    uint32_t active_video_width;
    //Current video height ,this is used to do sequence change
    uint32_t active_video_height;
    uint32_t ready_to_close;

    void *api_ctx;
    void *api_param;
    void *p_session_data;
    void *p_session_data1;
    void *p_session_data2;
    ni_quadra_frame_t niframe;
    ni_dec_input_params_t niparam;
    ni_fifo_buffer_t *input_data_fifo;
    int suspended;
    int pkt_pts_unchange;
    int enable_follow_iframe;
    u_int32_t ui32FramesCorrupted;
    bool ppu_reconfig;
    ni_quadra_ppu_config_t ni_quadra_ppu_config;
} ni_quadra_dec_context_t;

/*!*****************************************************************************
 *  \brief  Initialize decode parameters and open decoder device 
 *          including ni_quadra_dec_context_t,
 *                    ni_session_context_t,
 *                    ni_xcoder_params_t
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_quadra_dec_init(ni_quadra_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  Initialize init device session_context and init default params
 *          including ni_quadra_dec_context_t,
 *                    ni_session_context_t,
 *                    ni_decoder_params_t
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_quadra_setup_decoder(ni_quadra_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  Parse the packet and send the packet to decode device
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *
 *  \param[in/out] flag   flag=0 or 1 to do something.
 *  \param[in]  pkt_data  Pointer to input packet data
 *  \param[in]  pkt_size  packet size
 *  \param[in]  pkt_flags packet flags
 *  \param[in]  pkt_pts   packet pts
 *  \param[in]  pkt_dts   packet dts
 *  \param[in]  avctx_width avctx width
 *  \param[in]  avctx_height  avctx height
 *  \param[in]  extradata   Pointer to uint8_t
 *  \param[in]  extradata_size  extradata size
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 ******************************************************************************/
LIB_API int ni_quadra_dec_send(ni_quadra_dec_context_t *ni_dec_ctx,
                               int pkt_flags, uint8_t *data, int pkt_size,
                               uint8_t *lonedata, int lone_size,
                               int64_t pkt_pts, int64_t pkt_dts, int pkt_pos,
                               int width, int height, uint8_t *extradata,
                               int extradata_size);

/*!*****************************************************************************
 *  \brief  Decode is flushing or not, including ni_quadra_dec_context_t,
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *
 *  \return On 1
 *                        HW dosn't do decode flushing
 *          On 0
 *                        HW is doing decode flushing
 ******************************************************************************/
LIB_API int ni_quadra_dec_is_flushing(ni_quadra_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  Receive decoded data from decoder, including ni_quadra_dec_context_t,
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *  \param[in/out] p_session_data  Pointer to ni_session_data_io_t
 *  \param[in]  avctx_width avctx width
 *  \param[in]  avctx_height avctx height
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 ******************************************************************************/
LIB_API int ni_quadra_dec_receive(ni_quadra_dec_context_t *ni_dec_ctx,
                                  int avctx_width, int avctx_height,
                                  int frame_planar);

/*!*****************************************************************************
 *  \brief  Decode close, including ni_quadra_dec_context_t,
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *
 ******************************************************************************/
LIB_API void ni_quadra_dec_close(ni_quadra_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  buffer free, including void* 
 *
 *  \param[in]   opaque parameter to be got for processing or passed to free
 *  \param[in]   data  data array
 *
 ******************************************************************************/
LIB_API void ni_quadra_align_free(void *opaque, uint8_t *data);

/*!*****************************************************************************
 *  \brief  frame free, including void* 
 *
 *  \param[in]   opaque parameter to be got for processing or passed to free
 *  \param[in]   data  data array
 *
 ******************************************************************************/
LIB_API void ni_quadra_frame_free(void *opaque, uint8_t *data);

/*!*****************************************************************************
 *  \brief  Free and remove all auxiliary data from the ni_dec_ctx frame.
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *
 *  \return None
 ******************************************************************************/
LIB_API void ni_quadra_dec_wipe_aux_data(ni_quadra_dec_context_t *ni_dec_ctx);

/*!*****************************************************************************
 *  \brief  get size from the frame auxiliary data of a given type if exists
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *  \param[in] type   type of the auxiliary data to be retrieved
 *
 *  \return aux data size
 ******************************************************************************/
LIB_API int ni_quadra_frame_get_aux_size(ni_quadra_dec_context_t *ni_dec_ctx,
                                         int type);

/*!*****************************************************************************
 *  \brief  Retrieve from the frame auxiliary data of a given type if exists
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *  \param[in/out] dst_data  copy the auxiliary data
 *  \param[in] type   type of the auxiliary data
 *  \param[in] size   size of the auxiliary data
 *
 *  \return a pointer to the aux data of a given type on success, NULL otherwise
 ******************************************************************************/
LIB_API void ni_quadra_frame_copy_aux_data(ni_quadra_dec_context_t *ni_dec_ctx,
                                           uint8_t *dst_data, int type,
                                           int size);

/*!*****************************************************************************
 *  \brief  Set value referenced by name in decoder parameters structure
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *  \param[in] name       String represented parameter name to search
 *  \param[in] value      Parameter value to set
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 *                        NI_QUADRA_PARAM_WARNING_DEPRECATED
 *******************************************************************************/
LIB_API int ni_quadra_dec_params_set_value(ni_quadra_dec_context_t *ni_dec_ctx,
                                           const char *name, char *value);

/*!*****************************************************************************
 *  \brief  sync ni_decoder_input_params_t and ni_xcoder_params_t
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *  \param[in] type       0:sync ni_dec_input_params_t to ni_decoder_input_params_t
 *                        1:sync ni_decoder_input_params_t to ni_xcoder_params_t
 *                        2:copy semi_planar force_8_bit to ni_dec_input_params_t 
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 *                        NI_QUADRA_PARAM_WARNING_DEPRECATED
 *******************************************************************************/
LIB_API int
ni_quadra_symbolic_decoder_param(ni_quadra_dec_context_t *ni_dec_ctx, int type);

/*!*****************************************************************************
 *  \brief  alloc session p_data
 *
 *  \param[in/out] ni_dec_ctx  Pointer to ni_quadra_dec_context_t
 *  \param[in] p_session_data Pointer to void*
 *
 *  \return  uint8_t *p_data
 *
 ******************************************************************************/
LIB_API uint8_t *
ni_quadra_dec_frame_buffer_alloc(ni_quadra_dec_context_t *ni_dec_ctx,
                                 void *p_session_data);
#ifdef __cplusplus
}
#endif
