/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
 *  \file   ni_quadra_enc_api.c
 *
 *  \brief  Private definitions used by ni_quadra_enc_api.c for video processing
 *          tasks
 ******************************************************************************/

#include "ni_quadra_enc_api.h"
#include "ni_device_api.h"
#include "ni_defs.h"
#include "ni_rsrc_api.h"
#include "ni_util.h"
#include "ni_av_codec.h"

void ni_quadra_enc_gop_params_set(ni_quadra_enc_context_t *p_enc_ctx, char *key)
{
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    ni_gop_params_check_set(p_param, key);
}

bool ni_quadra_enc_gop_params_check(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    return ni_gop_params_check(p_param);
}

int ni_quadra_encoder_xcoder_opts_params_set_value(
    ni_quadra_enc_context_t *p_enc_ctx, const char *name, const char *value)
{
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    int parse_ret = ni_encoder_params_set_value(p_param, name, value);
    if (parse_ret != NI_RETCODE_SUCCESS)
    {
        switch (parse_ret)
        {
            case NI_RETCODE_PARAM_INVALID_NAME:
                ni_log(NI_LOG_ERROR, "Unknown option: %s.\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_TOO_BIG:
                ni_log(NI_LOG_ERROR, "Invalid %s: too big\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_TOO_SMALL:
                ni_log(NI_LOG_ERROR, "Invalid %s: too small\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_OOR:
                ni_log(NI_LOG_ERROR, "Invalid %s: out of range\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_ZERO:
                ni_log(NI_LOG_ERROR, "Error setting option %s to value 0\n",
                       name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_INVALID_VALUE:
                ni_log(NI_LOG_ERROR, "Invalid value for %s: %s.\n", name,
                       value);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_FAILURE:
                ni_log(NI_LOG_ERROR,
                       "Generic failure during xcoder-params setting for %s\n",
                       name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_WARNING_DEPRECATED:
                ni_log(NI_LOG_INFO, "Parameter %s is deprecated\n", name);
                parse_ret = NI_QUADRA_PARAM_WARNING_DEPRECATED;
                break;
            default:
                ni_log(NI_LOG_ERROR, "Invalid %s: ret %d\n", name, parse_ret);
                return NI_RETCODE_FAILURE;
        }
    }
    return parse_ret;
}

int ni_quadra_encoder_xcoder_gop_params_set_value(
    ni_quadra_enc_context_t *p_enc_ctx, const char *name, const char *value)
{
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    int parse_ret = ni_encoder_gop_params_set_value(p_param, name, value);
    if (parse_ret != NI_RETCODE_SUCCESS)
    {
        // deprecated param will trigger warning, not error
        // so do not free param dict for it
        // if (parse_ret != NI_RETCODE_PARAM_WARNING_DEPRECATED)
        // av_dict_free(&dict);
        switch (parse_ret)
        {
            case NI_RETCODE_PARAM_INVALID_NAME:
                ni_log(NI_LOG_ERROR, "Unknown option: %s.\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_TOO_BIG:
                ni_log(NI_LOG_ERROR,
                       "Invalid custom GOP parameters: %s too big\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_TOO_SMALL:
                ni_log(NI_LOG_ERROR,
                       "Invalid custom GOP parameters: %s too small\n", name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_OOR:
                ni_log(NI_LOG_ERROR,
                       "Invalid custom GOP parameters: %s out of range \n",
                       name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_ERROR_ZERO:
                ni_log(NI_LOG_ERROR,
                       "Invalid custom GOP paramaters: Error setting option %s "
                       "to value 0 \n",
                       name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_INVALID_VALUE:
                ni_log(NI_LOG_ERROR, "Invalid value for GOP param %s: %s.\n",
                       name, value);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_FAILURE:
                ni_log(NI_LOG_ERROR,
                       "Generic failure during xcoder-params setting for %s\n",
                       name);
                return NI_RETCODE_FAILURE;
            case NI_RETCODE_PARAM_WARNING_DEPRECATED:
                ni_log(NI_LOG_INFO, "Parameter %s is deprecated\n", name);
                parse_ret = NI_QUADRA_PARAM_WARNING_DEPRECATED;
                break;
            default:
                ni_log(NI_LOG_ERROR, "Invalid %s: ret %d\n", name, parse_ret);
                return NI_RETCODE_FAILURE;
        }
    }
    return parse_ret;
}

static int
ni_xcoder_encoder_header_check_set(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_xcoder_params_t *p_param;
    p_param = (ni_xcoder_params_t *)p_session_ctx->p_session_config;
    p_param->cfg_enc_params.videoFullRange = 0;

    // DolbyVision support
    if (5 == p_param->dolby_vision_profile)
    {
        ni_log(NI_LOG_ERROR,
               "dolbyVisionProfile is not supported on FFmpeg version 4.2.1 "
               "and earlier.\n");
        return -1;
    }

    if (p_enc_ctx->color_primaries != 2 || p_enc_ctx->color_trc != 2 ||
        p_enc_ctx->color_space != 2)
    {
        p_param->cfg_enc_params.colorDescPresent = 1;
        p_param->cfg_enc_params.colorPrimaries = p_enc_ctx->color_primaries;
        p_param->cfg_enc_params.colorTrc = p_enc_ctx->color_trc;
        p_param->cfg_enc_params.colorSpace = p_enc_ctx->color_space;

        ni_log(NI_LOG_INFO,
               "XCoder HDR color info color_primaries: %d "
               "color_trc: %d  color_space %d\n",
               p_enc_ctx->color_primaries, p_enc_ctx->color_trc,
               p_enc_ctx->color_space);
    }

    // for YUVJ420P pix fmt, indicate it's full range video
    if (p_enc_ctx->color_range == 2 || 12 == p_enc_ctx->pix_fmt ||
        NI_QUADRA_PIX_FMT_YUVJ420P == p_enc_ctx->sw_pix_fmt)
    {
        p_param->cfg_enc_params.videoFullRange = 1;
    }

    return 0;
}

int ni_quadra_xcoder_encoder_headers(ni_quadra_enc_context_t *p_enc_ctx)
{
    // use a copy of encoder context, take care to restore original config
    // cropping setting
    ni_xcoder_params_t *p_param;
    ni_session_context_t *p_orig_session_ctx = p_enc_ctx->api_ctx;
    ni_session_data_io_t *p_packet = p_enc_ctx->api_pkt;
    ni_session_context_t session_ctx = {0};
    ni_packet_t *xpkt;
    int orig_conf_win_right;
    int orig_conf_win_bottom;
    int linesize_aligned, height_aligned;
    int ret = 0, recv = 0, saved_ret = 0;

    memcpy(&session_ctx, p_enc_ctx->api_ctx, sizeof(ni_session_context_t));
    p_param = (ni_xcoder_params_t *)p_orig_session_ctx->p_session_config;

    orig_conf_win_right = p_param->cfg_enc_params.conf_win_right;
    orig_conf_win_bottom = p_param->cfg_enc_params.conf_win_bottom;
    //p_enc_ctx->api_ctx = &session_ctx;
    linesize_aligned = p_enc_ctx->width;
    if (linesize_aligned < NI_MIN_WIDTH)
    {
        p_param->cfg_enc_params.conf_win_right +=
            (NI_MIN_WIDTH - p_enc_ctx->width) / 2 * 2;
        linesize_aligned = NI_MIN_WIDTH;
    } else
    {
        if (p_enc_ctx->sw_pix_fmt ==
                NI_QUADRA_PIX_FMT_8_TILED4X4 ||   //AV_PIX_FMT_NI_QUAD_8_TILE_4X4 ||
            p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_10_TILED4X4)
        {   //AV_PIX_FMT_NI_QUAD_10_TILE_4X4) {
            linesize_aligned = NIALIGN(p_enc_ctx->width, 4);
            p_param->cfg_enc_params.conf_win_right +=
                (linesize_aligned - p_enc_ctx->width) / 2 * 2;
        } else
        {
            linesize_aligned = NIALIGN(p_enc_ctx->width, 2);
            p_param->cfg_enc_params.conf_win_right +=
                (linesize_aligned - p_enc_ctx->width) / 2 * 2;
        }
    }
    p_param->source_width = linesize_aligned;

    height_aligned = p_enc_ctx->height;
    if (height_aligned < NI_MIN_HEIGHT)
    {
        p_param->cfg_enc_params.conf_win_bottom +=
            (NI_MIN_HEIGHT - p_enc_ctx->height) / 2 * 2;
        height_aligned = NI_MIN_HEIGHT;
    } else
    {
        if (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_8_TILED4X4 ||
            p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_10_TILED4X4)
        {
            height_aligned = NIALIGN(p_enc_ctx->height, 4);
            p_param->cfg_enc_params.conf_win_bottom +=
                (height_aligned - p_enc_ctx->height) / 4 * 4;
        } else
        {
            height_aligned = NIALIGN(p_enc_ctx->height, 2);
            p_param->cfg_enc_params.conf_win_bottom +=
                (height_aligned - p_enc_ctx->height) / 2 * 2;
        }
    }
    p_param->source_height = height_aligned;

    session_ctx.hw_id = p_enc_ctx->dev_enc_idx;
    ni_xcoder_strncpy(session_ctx.blk_dev_name, p_enc_ctx->dev_blk_name,
                      NI_MAX_DEVICE_NAME_LEN);
    ni_xcoder_strncpy(session_ctx.dev_xcoder_name, p_enc_ctx->dev_xcoder,
                      MAX_CHAR_IN_DEVICE_NAME);

    ret = ni_device_session_open(&session_ctx, NI_DEVICE_TYPE_ENCODER);

    p_enc_ctx->dev_xcoder_name = session_ctx.dev_xcoder_name;
    p_enc_ctx->blk_xcoder_name = session_ctx.blk_xcoder_name;
    p_enc_ctx->dev_enc_idx = session_ctx.hw_id;

    if (ret != 0)
    {
        ni_log(NI_LOG_ERROR,
               "Failed to open encoder (status = %d), "
               "critical error or resource unavailable\n",
               ret);
        ret = NI_RETCODE_FAILURE;
        goto end;
    } else
    {
        ni_log(NI_LOG_DEBUG,
               "XCoder %s Index %d (inst: %d) opened successfully\n",
               p_enc_ctx->dev_xcoder_name, p_enc_ctx->dev_enc_idx,
               session_ctx.session_id);
    }

    xpkt = &p_packet->data.packet;
    ni_packet_buffer_alloc(xpkt, NI_MAX_TX_SZ);

    while (1)
    {
        recv = ni_device_session_read(&session_ctx, p_packet,
                                      NI_DEVICE_TYPE_ENCODER);

        if (recv > 0)
        {
            p_enc_ctx->extradata_size = recv - (int)session_ctx.meta_size;
            p_enc_ctx->extradata = malloc(p_enc_ctx->extradata_size);
            if (!p_enc_ctx->extradata)
            {
                ni_log(NI_LOG_ERROR,
                       "ERROR %d:%s() line %d, alloc extradata failed\n",
                       NI_ERRNO, __FUNCTION__, __LINE__);
                p_enc_ctx->extradata_size = 0;
                recv = NI_RETCODE_ERROR_MEM_ALOC;
            } else
            {
                memset(p_enc_ctx->extradata, 0, p_enc_ctx->extradata_size);
                memcpy(p_enc_ctx->extradata,
                       (uint8_t *)xpkt->p_data + session_ctx.meta_size,
                       p_enc_ctx->extradata_size);
            }
            ni_log(NI_LOG_DEBUG, "Xcoder encoder headers len: %d\n",
                   p_enc_ctx->extradata_size);
            break;
        } else if (recv == NI_RETCODE_SUCCESS)
        {
            continue;
        } else
        {
            ni_log(NI_LOG_ERROR, "Xcoder encoder headers error: %d", recv);
            break;
        }
    }

end:
    // close and clean up the temporary session
    saved_ret = ret;
    
    ret = ni_device_session_close(&session_ctx, p_enc_ctx->encoder_eof,
                                      NI_DEVICE_TYPE_ENCODER);
    ret = (ret < 0) ? ret : saved_ret; 

#ifdef _WIN32
    ni_device_close(session_ctx.device_handle);
#elif __linux__
    ni_device_close(session_ctx.device_handle);
    ni_device_close(session_ctx.blk_io_handle);
#endif
    session_ctx.device_handle = NI_INVALID_DEVICE_HANDLE;
    session_ctx.blk_io_handle = NI_INVALID_DEVICE_HANDLE;

    ni_packet_buffer_free(&(p_packet->data.packet));

    p_param->cfg_enc_params.conf_win_right = orig_conf_win_right;
    p_param->cfg_enc_params.conf_win_bottom = orig_conf_win_bottom;
    p_enc_ctx->api_ctx = p_orig_session_ctx;
    return (recv < 0 ? recv : ret);
}

static int ni_quadra_to_ni_pix_fmt(ni_quadra_enc_context_t *p_enc_ctx)
{
    int pixel_format;
    switch (p_enc_ctx->sw_pix_fmt)
    {
        case NI_QUADRA_PIX_FMT_YUV420P:
        case NI_QUADRA_PIX_FMT_YUVJ420P:
            pixel_format = NI_PIX_FMT_YUV420P;
            break;
        case NI_QUADRA_PIX_FMT_P010LE:
            pixel_format = NI_PIX_FMT_P010LE;
            break;
        case NI_QUADRA_PIX_FMT_YUV420P10LE:
            pixel_format = NI_PIX_FMT_YUV420P10LE;
            break;
        case NI_QUADRA_PIX_FMT_ARGB:
            pixel_format = NI_PIX_FMT_ARGB;
            break;
        case NI_QUADRA_PIX_FMT_ABGR:
            pixel_format = NI_PIX_FMT_ABGR;
            break;
        case NI_QUADRA_PIX_FMT_RGBA:
            pixel_format = NI_PIX_FMT_RGBA;
            break;
        case NI_QUADRA_PIX_FMT_BGRA:
            pixel_format = NI_PIX_FMT_BGRA;
            break;
        case NI_QUADRA_PIX_FMT_NV12:
            pixel_format = NI_PIX_FMT_NV12;
            break;
        case NI_QUADRA_PIX_FMT_8_TILED4X4:
            pixel_format = NI_PIX_FMT_8_TILED4X4;
            break;
        case NI_QUADRA_PIX_FMT_10_TILED4X4:
            pixel_format = NI_PIX_FMT_10_TILED4X4;
            break;
        default:
            ni_log(NI_LOG_ERROR, "Pixfmt %d not supported in Quadra encoder\n",
                p_enc_ctx->pix_fmt);
            return NI_RETCODE_FAILURE;
    }
    return pixel_format;
}

int ni_quadra_setup_encoder(ni_quadra_enc_context_t *p_enc_ctx)
{
    int ret = 0;
    uint32_t xcoder_timeout;
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    ni_xcoder_params_t *pparams = NULL;
    ni_session_data_io_t *p_packet = p_enc_ctx->api_pkt;
    ni_session_data_io_t *p_frame = p_enc_ctx->api_fme;
    p_session_ctx->codec_format = p_enc_ctx->codec_format;

    p_enc_ctx->firstPktArrived = 0;
    p_enc_ctx->spsPpsArrived = 0;
    p_enc_ctx->spsPpsHdrLen = 0;
    p_enc_ctx->p_spsPpsHdr = NULL;
    p_enc_ctx->xcode_load_pixel = 0;
    p_enc_ctx->reconfigCount = 0;
    p_enc_ctx->gotPacket = 0;
    p_enc_ctx->sentFrame = 0;
    p_enc_ctx->latest_dts = 0;
    p_enc_ctx->eos_fme_received = 0;

    #ifdef MEASURE_LATENCY
    p_enc_ctx->enable_test_latency = 1;
    #else
    p_enc_ctx->enable_test_latency = 0;
    #endif

    ret = ni_quadra_to_ni_pix_fmt(p_enc_ctx);
    if (ret < 0) {
        return NI_RETCODE_FAILURE;
    }
    else {
        p_session_ctx->pixel_format = ret;
    }

    if (p_param->enable_vfr)
    {
        // in the vfr mode, if the initial framerate is out of [5-120]
        // think the initial framerate is incorrect, set it to default 30 fps
        if (p_param->cfg_enc_params.frame_rate < 5 ||
            p_param->cfg_enc_params.frame_rate > 120)
        {
            p_param->cfg_enc_params.frame_rate = 30;
            p_session_ctx->prev_fps = 30;
        } else
        {
            p_session_ctx->prev_fps = p_param->cfg_enc_params.frame_rate;
        }
        p_session_ctx->last_change_framenum = 0;
        p_session_ctx->fps_change_detect_count = 0;
    }

    if (p_enc_ctx->nvme_io_size > 0 && p_enc_ctx->nvme_io_size % 4096 != 0)
    {
        ni_log(NI_LOG_ERROR, "Error XCoder iosize is not 4KB aligned!\n");
        return NI_RETCODE_FAILURE;
    }

    p_session_ctx->p_session_config = p_enc_ctx->api_param;
    pparams = (ni_xcoder_params_t *)p_session_ctx->p_session_config;

    // keep dtsOffset in case of sequence change
    if (SESSION_RUN_STATE_SEQ_CHANGE_DRAINING !=
        p_session_ctx->session_run_state)
    {
        switch (pparams->cfg_enc_params.gop_preset_index)
        {
            /* dtsOffset is the max number of non-reference frames in a GOP
        * (derived from x264/5 algo) In case of IBBBP the first dts of the I
        * frame should be input_pts-(3*ticks_per_frame) In case of IBP the
        * first dts of the I frame should be input_pts-(1*ticks_per_frame)
        * thus we ensure pts>dts in all cases
        * */
            case 1:
            case 9:
            case 10:
                p_enc_ctx->dts_offset = 0;
                break;
            /* ts requires dts/pts of I fraem not same when there are B frames in
        * streams */
            case 3:
            case 4:
            case 7:
                p_enc_ctx->dts_offset = 1;
                break;
            case 5:
                p_enc_ctx->dts_offset = 2;
                break;
            case -1:   // adaptive GOP
            case 8:
                p_enc_ctx->dts_offset = 3;
                break;
            default:
                p_enc_ctx->dts_offset = 7;
                break;
        }
        if (pparams->cfg_enc_params.custom_gop_params.custom_gop_size)
        {
            int dts_offset = 0;
            p_enc_ctx->dts_offset = 0;
            bool has_b_frame = false;
            for (int idx = 0; idx <
                 pparams->cfg_enc_params.custom_gop_params.custom_gop_size;
                 idx++)
            {
                if (pparams->cfg_enc_params.custom_gop_params.pic_param[idx]
                        .poc_offset < idx + 1)
                {
                    dts_offset = (idx + 1) -
                        pparams->cfg_enc_params.custom_gop_params.pic_param[idx]
                            .poc_offset;
                    if (p_enc_ctx->dts_offset < dts_offset)
                    {
                        p_enc_ctx->dts_offset = dts_offset;
                    }
                }
                if (!has_b_frame &&
                    (pparams->cfg_enc_params.custom_gop_params.pic_param[idx]
                         .pic_type == PIC_TYPE_B))
                {
                    has_b_frame = true;
                }
            }

            if (has_b_frame && !p_enc_ctx->dts_offset)
            {
                p_enc_ctx->dts_offset = 1;
            }
        }
        ni_log(NI_LOG_DEBUG, "dts offset set to %ld\n", p_enc_ctx->dts_offset);
    }

    p_enc_ctx->total_frames_received = 0;
    p_enc_ctx->gop_offset_count = 0;
    ni_log(NI_LOG_DEBUG, "dts offset: %" PRId64 ", gop_offset_count: %d\n",
           p_enc_ctx->dts_offset, p_enc_ctx->gop_offset_count);

    //overwrite the nvme io size here with a custom value if it was provided
    if (p_enc_ctx->nvme_io_size > 0)
    {
        p_session_ctx->max_nvme_io_size = p_enc_ctx->nvme_io_size;
        ni_log(NI_LOG_INFO, "Custom NVMe IO Size set to = %d\n",
               p_session_ctx->max_nvme_io_size);
    }
    // overwrite keep alive timeout value here with a custom value if it was
    // provided
    // if xcoder option is set then overwrite the (legacy) decoder option
    xcoder_timeout = p_param->cfg_enc_params.keep_alive_timeout;
    if (xcoder_timeout != NI_DEFAULT_KEEP_ALIVE_TIMEOUT)
    {
        p_session_ctx->keep_alive_timeout = xcoder_timeout;
    } else
    {
        p_session_ctx->keep_alive_timeout = p_enc_ctx->keep_alive_timeout;
    }
    ni_log(NI_LOG_DEBUG, "Custom NVME Keep Alive Timeout set to = %d\n",
           p_session_ctx->keep_alive_timeout);
    p_enc_ctx->encoder_eof = 0;
    p_enc_ctx->encoder_flushing = 0;
    if (!p_enc_ctx->is_swframes)
    {
        if ((p_enc_ctx->height >= NI_MIN_HEIGHT) &&
            (p_enc_ctx->width >= NI_MIN_WIDTH))
        {
            p_param->hwframes = 1;
        } else if (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_8_TILED4X4 ||
                   p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_10_TILED4X4)
        {
            ni_log(NI_LOG_ERROR,
                   "Invalid Picture Height or Width: too small\n");
            return NI_RETCODE_FAILURE;
        }
    }

    ni_log(NI_LOG_INFO, "p_param->hwframes = %d\n", p_param->hwframes);
    p_session_ctx->roi_len = 0;
    p_session_ctx->roi_avg_qp = 0;
    p_session_ctx->src_bit_depth = 8;
    p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
    p_session_ctx->bit_depth_factor = 1;

    if (NI_QUADRA_PIX_FMT_YUV420P10BE == p_enc_ctx->sw_pix_fmt ||
        NI_QUADRA_PIX_FMT_YUV420P10LE == p_enc_ctx->sw_pix_fmt || 
        NI_QUADRA_PIX_FMT_P010LE == p_enc_ctx->sw_pix_fmt ||
        NI_QUADRA_PIX_FMT_10_TILED4X4 == p_enc_ctx->sw_pix_fmt)
    {
        p_session_ctx->bit_depth_factor = 2;
        p_session_ctx->src_bit_depth = 10;
        if (NI_QUADRA_PIX_FMT_YUV420P10BE == p_enc_ctx->sw_pix_fmt)
        {
            p_session_ctx->src_endian = NI_FRAME_BIG_ENDIAN;
        }
    }

    switch (p_enc_ctx->sw_pix_fmt)
    {
        case NI_QUADRA_PIX_FMT_NV12:
        case NI_QUADRA_PIX_FMT_P010LE:
            pparams->cfg_enc_params.planar = NI_PIXEL_PLANAR_FORMAT_SEMIPLANAR;
            break;
        case NI_QUADRA_PIX_FMT_8_TILED4X4:
        case NI_QUADRA_PIX_FMT_10_TILED4X4:
            pparams->cfg_enc_params.planar = NI_PIXEL_PLANAR_FORMAT_TILED4X4;
            break;
        default:
            pparams->cfg_enc_params.planar = NI_PIXEL_PLANAR_FORMAT_PLANAR;
            break;
    }

    p_enc_ctx->user_sar_num = 0;
    p_enc_ctx->user_sar_den = 1;
    // check user set sar
    if (pparams->cfg_enc_params.aspectRatioWidth != p_enc_ctx->sar_num ||
        pparams->cfg_enc_params.aspectRatioHeight != p_enc_ctx->sar_den)
    {
        p_enc_ctx->user_sar_num = pparams->cfg_enc_params.aspectRatioWidth;
        p_enc_ctx->user_sar_den = pparams->cfg_enc_params.aspectRatioHeight;
    }

    // init HDR SEI stuff
    p_session_ctx->sei_hdr_content_light_level_info_len =
        p_session_ctx->light_level_data_len =
            p_session_ctx->sei_hdr_mastering_display_color_vol_len =
                p_session_ctx->mdcv_max_min_lum_data_len = 0;
    p_session_ctx->p_master_display_meta_data = NULL;

    p_packet->data.packet.av1_buffer_index = 0;
    if (p_enc_ctx->width > 0 && p_enc_ctx->height > 0)
    {
        bool isnv12frame = (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_NV12 ||
                            p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_P010LE);
        ni_frame_buffer_alloc(&p_frame->data.frame,
                              NIODD2EVEN(p_enc_ctx->width),
                              NIODD2EVEN(p_enc_ctx->height), 0, 0,
                              p_session_ctx->bit_depth_factor, 1, !isnv12frame);
    }

    if (ni_xcoder_encoder_header_check_set(p_enc_ctx) < 0)
    {
        return NI_RETCODE_FAILURE;
    }

    // original resolution this stream started with, this is used in encoder
    // sequence change
    p_session_ctx->ori_width = p_enc_ctx->width;
    p_session_ctx->ori_height = p_enc_ctx->height;
    p_session_ctx->ori_bit_depth_factor = p_session_ctx->bit_depth_factor;
    p_session_ctx->ori_pix_fmt = p_session_ctx->pixel_format;

    ni_log(NI_LOG_DEBUG,
           "ni_quadra_setup_encoder "
           "sw_pix_fmt %d ori_pix_fmt %d\n",
           p_enc_ctx->sw_pix_fmt, p_session_ctx->ori_pix_fmt);

    p_session_ctx->ori_luma_linesize = 0;
    p_session_ctx->ori_chroma_linesize = 0;
    p_enc_ctx->bit_rate = pparams->bitrate;
    if (p_param->cfg_enc_params.get_psnr_mode != 3)
    {
        p_enc_ctx->bget_psnr = 1;
    }
    else
    {
       p_enc_ctx->bget_psnr = 0;
    }

#define NI_CODEC_ID_MJPEG 8
    if ((pparams->generate_enc_hdrs || p_enc_ctx->generate_enc_hdrs) && \
        p_enc_ctx->codec_id != NI_CODEC_ID_MJPEG)
    {
        ret = ni_quadra_xcoder_encoder_headers(p_enc_ctx);
    }
#undef NI_CODEC_ID_MJPEG
    return ret;
}

int ni_quadra_enc_init(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_xcoder_params_t *p_param = NULL;
    ni_session_context_t *p_session_ctx = NULL;
    ni_session_data_io_t *p_session_data_fme = NULL;
    ni_session_data_io_t *p_session_data_pkt = NULL;
    ni_frame_t *p_input_frame = NULL;
    ni_session_run_state_t prev_state;
    int ret = NI_RETCODE_SUCCESS;
    int i;

    ni_log_set_level(ff_to_ni_log_level(p_enc_ctx->ff_log_level));
    ni_log(NI_LOG_DEBUG, "XCoder setup device encoder\n");

    if (!p_enc_ctx->api_ctx || !p_enc_ctx->api_param)
    {
        p_session_data_fme = malloc(sizeof(ni_session_data_io_t));
        if (!p_session_data_fme)
        {
            ni_log(NI_LOG_ERROR,
                "ERROR %d:%s() line %d, alloc p_session_data frame failed\n",
                NI_ERRNO, __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            LRETURN;
        } else
        {
            p_enc_ctx->api_fme = p_session_data_fme;
        }
        p_session_data_pkt = malloc(sizeof(ni_session_data_io_t));
        if (!p_session_data_pkt)
        {
            ni_log(NI_LOG_ERROR,
                "ERROR %d:%s() line %d, alloc p_session_data pkt failed\n",
                NI_ERRNO, __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            LRETURN;
        } else
        {
            p_enc_ctx->api_pkt = p_session_data_pkt;
        }
        p_input_frame = malloc(sizeof(ni_frame_t));
        if (!p_input_frame)
        {
            ni_log(NI_LOG_ERROR,
                "ERROR %d:%s() line %d, alloc p_input_frame failed\n", NI_ERRNO,
                __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            LRETURN;
        } else
        {
            p_enc_ctx->input_frame = p_input_frame;
        }
        // Xcoder Session Context
        p_session_ctx = malloc(sizeof(ni_session_context_t));
        if (!p_session_ctx)
        {
            ni_log(NI_LOG_ERROR,
                "ERROR %d:%s() line %d, alloc session context failed\n",
                NI_ERRNO, __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            LRETURN;
        } else
        {
            memset(p_session_ctx, 0, sizeof(ni_session_context_t));
            p_enc_ctx->api_ctx = p_session_ctx;
        }

        p_param = malloc(sizeof(ni_xcoder_params_t));
        if (!p_param)
        {
            ni_log(NI_LOG_ERROR,
                "ERROR %d:%s() line %d, alloc session context failed\n",
                NI_ERRNO, __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            LRETURN;
        } else
        {
            p_enc_ctx->api_param = p_param;
        }
    }
    else
    {
        ni_log(NI_LOG_DEBUG, "api_ctx and api_param had alloced\n");
        p_session_ctx = p_enc_ctx->api_ctx;
        p_param = p_enc_ctx->api_param;
    }

    memset(p_enc_ctx->api_fme, 0, sizeof(ni_session_data_io_t));
    memset(p_enc_ctx->api_pkt, 0, sizeof(ni_session_data_io_t));
    memset(p_enc_ctx->input_frame, 0, sizeof(ni_frame_t));
    prev_state = p_session_ctx->session_run_state;
    if (p_session_ctx->session_run_state ==
        SESSION_RUN_STATE_SEQ_CHANGE_DRAINING)
    {
        p_enc_ctx->dev_enc_idx = p_enc_ctx->orig_dev_enc_idx;
    } else
    {
        p_enc_ctx->orig_dev_enc_idx = p_enc_ctx->dev_enc_idx;
    }

    //p_session_ctx->session_id = NI_INVALID_SESSION_ID;
    if (ni_device_session_context_init(p_session_ctx) < 0)
    {
        ni_log(NI_LOG_ERROR, "Error XCoder init encoder context failure\n");
        return NI_RETCODE_FAILURE;
    }
    p_session_ctx->session_run_state = prev_state;

    //Xcoder User Configuration
    ret = ni_encoder_init_default_params(
        p_param, p_enc_ctx->framerate_num, p_enc_ctx->framerate_den,
        p_enc_ctx->bit_rate, p_enc_ctx->width, p_enc_ctx->height,
        p_enc_ctx->codec_format);

    switch (ret)
    {
        case NI_RETCODE_PARAM_ERROR_WIDTH_TOO_BIG:
            ni_log(NI_LOG_ERROR, "Invalid Picture Width: too big\n");
            return NI_RETCODE_FAILURE;
        case NI_RETCODE_PARAM_ERROR_WIDTH_TOO_SMALL:
            ni_log(NI_LOG_ERROR, "Invalid Picture Width: too small\n");
            return NI_RETCODE_FAILURE;
        case NI_RETCODE_PARAM_ERROR_HEIGHT_TOO_BIG:
            ni_log(NI_LOG_ERROR, "Invalid Picture Height: too big\n");
            return NI_RETCODE_FAILURE;
        case NI_RETCODE_PARAM_ERROR_HEIGHT_TOO_SMALL:
            ni_log(NI_LOG_ERROR, "Invalid Picture Height: too small\n");
            return NI_RETCODE_FAILURE;
        case NI_RETCODE_PARAM_ERROR_AREA_TOO_BIG:
            ni_log(NI_LOG_ERROR, "Invalid Picture Width x Height: exceeds %d\n",
                   NI_MAX_RESOLUTION_AREA);
            return NI_RETCODE_FAILURE;
        case NI_RETCODE_PARAM_ERROR_PIC_WIDTH:
            ni_log(NI_LOG_ERROR, "Invalid Picture Width\n");
            return NI_RETCODE_FAILURE;
        case NI_RETCODE_PARAM_ERROR_PIC_HEIGHT:
            ni_log(NI_LOG_ERROR, "Invalid Picture Height\n");
            return NI_RETCODE_FAILURE;
        default:
            if (ret < 0)
            {
                ni_log(NI_LOG_ERROR, "Error setting preset or log.\n");
                ni_log(NI_LOG_INFO, "Possible presets:");
                for (i = 0; g_xcoder_preset_names[i]; i++)
                    ni_log(NI_LOG_INFO, " %s", g_xcoder_preset_names[i]);
                ni_log(NI_LOG_INFO, "\n");

                ni_log(NI_LOG_INFO, "Possible log:");
                for (i = 0; g_xcoder_log_names[i]; i++)
                    ni_log(NI_LOG_INFO, " %s", g_xcoder_log_names[i]);
                ni_log(NI_LOG_INFO, "\n");

                return -22;
            }
            break;
    }
    if (p_param->cfg_enc_params.aspectRatioWidth == 0 &&
        p_param->cfg_enc_params.aspectRatioHeight == 1)
    {
        p_param->cfg_enc_params.aspectRatioWidth = p_enc_ctx->sar_num;
        p_param->cfg_enc_params.aspectRatioHeight = p_enc_ctx->sar_den;
        ni_log(NI_LOG_DEBUG, "%s sar set to %d:%d\n", __func__,
               p_param->cfg_enc_params.aspectRatioWidth,
               p_param->cfg_enc_params.aspectRatioHeight);
    }
    return ret;

end:
    if (ret != NI_RETCODE_SUCCESS)
    {
        free(p_session_ctx);
        p_enc_ctx->api_ctx = NULL;
        free(p_session_data_fme);
        p_enc_ctx->api_fme = NULL;
        free(p_input_frame);
        p_enc_ctx->input_frame = NULL;
        free(p_session_data_pkt);
        p_enc_ctx->api_pkt = NULL;
        free(p_param);
        p_enc_ctx->api_param = NULL;
    }
    return ret;
}

int ni_quadra_encode_close(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    int ret = 0;
    ni_session_data_io_t *ni_frame = p_enc_ctx->api_fme;
    ni_session_data_io_t *ni_packet = p_enc_ctx->api_pkt;
    ret = ni_device_session_close(p_session_ctx, p_enc_ctx->encoder_eof,
                                  NI_DEVICE_TYPE_ENCODER);
    if (NI_RETCODE_SUCCESS != ret)
    {
        ni_log(NI_LOG_ERROR, "Failed to close Encoder Session (status = %d)\n",
               ret);
    }

    if (ni_frame->data.frame.buffer_size ||
        ni_frame->data.frame.metadata_buffer_size ||
        ni_frame->data.frame.start_buffer_size)
    {
        ni_frame_buffer_free(&(ni_frame->data.frame));
    }
    ni_packet_buffer_free(&(ni_packet->data.packet));
    ni_log(NI_LOG_DEBUG, "XCoder encode close (status = %d)\n", ret);
    ni_device_session_context_clear(p_session_ctx);

    ni_log(NI_LOG_INFO, "XCoder encode close: session_run_state %d\n",
           p_session_ctx->session_run_state);
    if (p_session_ctx->session_run_state !=
        SESSION_RUN_STATE_SEQ_CHANGE_DRAINING)
    {
        ni_log(NI_LOG_INFO,
               "XCoder encode close: close blk_io_handle %d device_handle %d\n",
               p_session_ctx->blk_io_handle, p_session_ctx->device_handle);
#ifdef _WIN32
        ni_device_close(p_session_ctx->device_handle);
#elif __linux__
        ni_device_close(p_session_ctx->device_handle);
        ni_device_close(p_session_ctx->blk_io_handle);
#endif
        p_session_ctx->device_handle = NI_INVALID_DEVICE_HANDLE;
        p_session_ctx->blk_io_handle = NI_INVALID_DEVICE_HANDLE;
        p_session_ctx->auto_dl_handle = NI_INVALID_DEVICE_HANDLE;
        p_session_ctx->sender_handle = NI_INVALID_DEVICE_HANDLE;
        // ni_device_session_context_clear(p_session_ctx);
        if (p_enc_ctx->api_ctx)
            free(p_enc_ctx->api_ctx);
        p_enc_ctx->api_ctx = NULL;
        p_session_ctx = NULL;
        if (p_enc_ctx->api_fme)
            free(p_enc_ctx->api_fme);
        p_enc_ctx->api_fme = NULL;
        if (p_enc_ctx->input_frame)
            free(p_enc_ctx->input_frame);
        p_enc_ctx->input_frame = NULL;
        if (p_enc_ctx->api_pkt)
            free(p_enc_ctx->api_pkt);
        p_enc_ctx->api_pkt = NULL;
        if (p_enc_ctx->api_param)
            free(p_enc_ctx->api_param);
        p_enc_ctx->api_param = NULL;
    }

    free(p_enc_ctx->g_enc_change_params);
    p_enc_ctx->g_enc_change_params = NULL;
    free(p_enc_ctx->p_spsPpsHdr);
    p_enc_ctx->p_spsPpsHdr = NULL;
    p_enc_ctx->started = 0;

    return 0;
}

int ni_quadra_encode_sequence_change(ni_quadra_enc_context_t *p_enc_ctx,
                                     int width, int height,
                                     int bit_depth_factor)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    ni_xcoder_params_t *pparams =
        (ni_xcoder_params_t *)p_session_ctx->p_session_config;
    ni_session_data_io_t *ni_packet = p_enc_ctx->api_pkt;
    int ret = 0;
    ni_log(NI_LOG_INFO,
           "ni_quadra_encode_sequence_change encode sequence change: "
           "session_run_state: %d\n",
           p_session_ctx->session_run_state);
    ret = ni_device_session_sequence_change(
        p_session_ctx, width, height, bit_depth_factor, NI_DEVICE_TYPE_ENCODER);

    if (NI_RETCODE_SUCCESS != ret)
    {
        ni_log(NI_LOG_ERROR,
               "Failed to send Sequence Change to Encoder "
               "Session (status = %d)\n",
               ret);
        return ret;
    }
    if (!p_enc_ctx->is_swframes)
    {
        if (p_enc_ctx->height >= NI_MIN_HEIGHT &&
            p_enc_ctx->width >= NI_MIN_WIDTH)
        {
            p_param->hwframes = 1;
        }
    }

    // update session context
    p_session_ctx->bit_depth_factor = bit_depth_factor;
    p_session_ctx->src_bit_depth = (bit_depth_factor == 1) ? 8 : 10;
    p_session_ctx->src_endian =
        (NI_QUADRA_PIX_FMT_YUV420P10BE == p_enc_ctx->sw_pix_fmt) ?
        NI_FRAME_BIG_ENDIAN :
        NI_FRAME_LITTLE_ENDIAN;
    p_session_ctx->ready_to_close = 0;
    // need to reset frame_num because pkt_num is set to 1 when header received
    // after sequnce change, and low delay mode compares frame_num and pkt_num
    p_session_ctx->frame_num = 0;
    // also need to reset pkt_num because before header received,
    // pkt_num > frame_num will also cause low delay mode stuck
    p_session_ctx->pkt_num = 0;
    ni_packet->data.packet.end_of_stream = 0;

    switch (p_enc_ctx->sw_pix_fmt)
    {
        case NI_QUADRA_PIX_FMT_NV12:
        case NI_QUADRA_PIX_FMT_P010LE:
            pparams->cfg_enc_params.planar = NI_PIXEL_PLANAR_FORMAT_SEMIPLANAR;
            break;
        case NI_QUADRA_PIX_FMT_10_TILED4X4:
        case NI_QUADRA_PIX_FMT_8_TILED4X4:
            pparams->cfg_enc_params.planar = NI_PIXEL_PLANAR_FORMAT_TILED4X4;
            break;
        default:
            pparams->cfg_enc_params.planar = NI_PIXEL_PLANAR_FORMAT_PLANAR;
            break;
    }
    return ret;
}

int ni_quadra_encode_session_open(ni_quadra_enc_context_t *p_enc_ctx,
                                  int frame_width, int frame_height,
                                  const int *frame_linesize,
                                  niquadraFrameSurface1_t *p_surface,
                                  bool hw_frames_ctx, int frame_queqe,
                                  int avctx_flags)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    ni_session_data_io_t *p_niframe = p_enc_ctx->api_fme;
    int ret = 0;
    int bIsCenter = 0;
    if (p_enc_ctx->started == 0)
    {
        // adjust sample aspect ratio based on: user config, incoming decoded
        // frame and session context, in that order.
        if (p_enc_ctx->user_sar_num == 0 && p_enc_ctx->user_sar_den == 1)
        {
            if (p_param->cfg_enc_params.aspectRatioWidth !=
                    p_enc_ctx->sar_num ||
                p_param->cfg_enc_params.aspectRatioHeight != p_enc_ctx->sar_den)
            {
                p_param->cfg_enc_params.aspectRatioWidth = p_enc_ctx->sar_num;
                p_param->cfg_enc_params.aspectRatioHeight = p_enc_ctx->sar_den;
                ni_log(NI_LOG_DEBUG, "%s sar set to %d:%d\n", __func__,
                       p_param->cfg_enc_params.aspectRatioWidth,
                       p_param->cfg_enc_params.aspectRatioHeight);
            }
        }

        int linesize_aligned = frame_width;
        if (linesize_aligned < NI_MIN_WIDTH)
        {
            p_param->cfg_enc_params.conf_win_right +=
                (NI_MIN_WIDTH - frame_width) / 2 * 2;
            linesize_aligned = NI_MIN_WIDTH;
        } else
        {
            linesize_aligned = NIALIGN(frame_width, 2);
            p_param->cfg_enc_params.conf_win_right +=
                (linesize_aligned - frame_width) / 2 * 2;
        }
        p_param->source_width = linesize_aligned;

        int height_aligned = frame_height;
        if (height_aligned < NI_MIN_HEIGHT)
        {
            p_param->cfg_enc_params.conf_win_bottom +=
                (NI_MIN_HEIGHT - frame_height) / 2 * 2;
            height_aligned = NI_MIN_HEIGHT;
        } else
        {
            height_aligned = NIALIGN(frame_height, 2);
            p_param->cfg_enc_params.conf_win_bottom +=
                (height_aligned - frame_height) / 2 * 2;
        }
        p_param->source_height = height_aligned;
        ret = ni_quadra_to_ni_pix_fmt(p_enc_ctx);
        if (ret < 0) {
            return NI_RETCODE_FAILURE;
        }
        else {
            p_session_ctx->pixel_format = ret;
        }
        if (ni_xcoder_encoder_header_check_set(p_enc_ctx) < 0)
        {
            return NI_RETCODE_FAILURE;
        }
        ni_log(
            NI_LOG_INFO,
            "width/height %dx%d conf_win_right %d  conf_win_bottom %d , "
            "color primaries %u trc %u space %u\n",
            frame_width, frame_height, p_param->cfg_enc_params.conf_win_right,
            p_param->cfg_enc_params.conf_win_bottom, p_enc_ctx->color_primaries,
            p_enc_ctx->color_trc, p_enc_ctx->color_space);
        if (SESSION_RUN_STATE_SEQ_CHANGE_OPENING !=
            p_session_ctx->session_run_state)
        {
            // sequence change backup / restore encoder device handles, hw_id
            // and block device name, so no need to overwrite
            // hw_id/blk_dev_name to user set values
            p_session_ctx->hw_id = p_enc_ctx->dev_enc_idx;
            ni_xcoder_strncpy(p_session_ctx->dev_xcoder_name,
                              p_enc_ctx->dev_xcoder, MAX_CHAR_IN_DEVICE_NAME);

            ni_xcoder_strncpy(p_session_ctx->blk_dev_name,
                              p_enc_ctx->dev_blk_name, NI_MAX_DEVICE_NAME_LEN);
        }
        p_param->rootBufId =
            p_enc_ctx->is_swframes ? 0 : p_surface->ui16FrameIdx;

        if (!p_enc_ctx->is_swframes)
        {
            p_param->hwframes = 1;
            // if HW frame does not meet minimal resolution requirement, set
            // to SW frame encode.
            if (p_enc_ctx->height < NI_MIN_HEIGHT ||
                p_enc_ctx->width < NI_MIN_WIDTH)
            {
                p_param->hwframes = 0;
            }
            p_session_ctx->hw_action = NI_CODEC_HW_ENABLE;
            p_session_ctx->sender_handle =
                (ni_device_handle_t)((int64_t)p_surface->device_handle);
        }

        if (hw_frames_ctx && p_session_ctx->hw_id == -1 &&
            0 == strcmp(p_session_ctx->blk_dev_name, ""))
        {
            p_session_ctx->hw_id = frame_queqe;
            ni_log(
                NI_LOG_INFO,
                "xcoder_send_frame: hw_id -1, empty blk_dev_name, collocated "
                "to %d\n",
                p_session_ctx->hw_id);
        }
        // AUD insertion has to be handled differently in the firmware
        // if it is global header
        if (p_param->cfg_enc_params.EnableAUD)
        {
            if (avctx_flags & NI_CODEC_FLAG_GLOBAL_HEADER)
            {
                p_param->cfg_enc_params.EnableAUD =
                    NI_ENABLE_AUD_FOR_GLOBAL_HEADER;
            }

            ni_log(NI_LOG_INFO, "%s: EnableAUD %d global header flag %d\n",
                   __FUNCTION__, (p_param->cfg_enc_params.EnableAUD),
                   (avctx_flags & NI_CODEC_FLAG_GLOBAL_HEADER) ? 1 : 0);
        }

        // config linesize for zero copy (if input resolution is zero copy compatible)
        ni_encoder_frame_zerocopy_check(p_session_ctx, p_param, frame_width,
                                        frame_height,
                                        (const int *)frame_linesize, true);

        ret = ni_device_session_open(p_session_ctx, NI_DEVICE_TYPE_ENCODER);
        // // As the file handle may change we need to assign back
        p_enc_ctx->dev_xcoder_name = p_session_ctx->dev_xcoder_name;
        p_enc_ctx->blk_xcoder_name = p_session_ctx->blk_xcoder_name;
        p_enc_ctx->dev_enc_idx = p_session_ctx->hw_id;
        switch (ret)
        {
            case NI_RETCODE_SUCCESS:
                ni_log(NI_LOG_DEBUG,
                       "XCoder %s.%d (inst: %d) opened successfully\n",
                       p_enc_ctx->dev_xcoder_name, p_enc_ctx->dev_enc_idx,
                       p_session_ctx->session_id);
                break;
            case NI_RETCODE_INVALID_PARAM:
                ni_log(
                    NI_LOG_ERROR,
                    "Failed to open encoder (status = %d), invalid parameter "
                    "values given: %s\n",
                    ret, p_session_ctx->param_err_msg);
                ni_device_session_close(p_session_ctx, p_enc_ctx->encoder_eof,
                        NI_DEVICE_TYPE_DECODER);
                ni_device_session_context_clear(p_session_ctx);
                ret = NI_RETCODE_FAILURE;
                return ret;
            default:
                ni_log(NI_LOG_ERROR,
                       "Failed to open encoder (status = %d), "
                       "resource unavailable\n",
                       ret);
                ni_device_session_close(p_session_ctx, p_enc_ctx->encoder_eof,
                        NI_DEVICE_TYPE_DECODER);
                ni_device_session_context_clear(p_session_ctx);
                ret = NI_RETCODE_FAILURE;
                return ret;
        }

        if (p_param->cfg_enc_params.roi_enable &&
            (1 == p_param->roi_demo_mode || 2 == p_param->roi_demo_mode))
        {
            if (ni_set_demo_roi_map(p_enc_ctx->api_ctx) < 0)
            {
                return NI_RETCODE_ERROR_MEM_ALOC;
            }
        }   // roi_enable && roi_demo_mode
    }       // p_enc_ctx->started == 0

    if (p_enc_ctx->started == 0)
    {
        p_niframe->data.frame.start_of_stream = 1;
        p_enc_ctx->started = 1;
    } else if (p_session_ctx->session_run_state ==
               SESSION_RUN_STATE_SEQ_CHANGE_OPENING)
    {
        p_niframe->data.frame.start_of_stream = 1;
    } else
    {
        p_niframe->data.frame.start_of_stream = 0;
    }
    p_niframe->data.frame.preferred_characteristics_data_len = 0;
    p_niframe->data.frame.end_of_stream = 0;
    p_niframe->data.frame.force_key_frame =
        p_niframe->data.frame.use_cur_src_as_long_term_pic =
            p_niframe->data.frame.use_long_term_ref = 0;

    p_niframe->data.frame.sei_total_len = p_niframe->data.frame.sei_cc_offset =
        p_niframe->data.frame
            .sei_cc_len = p_niframe->data.frame
                              .sei_hdr_mastering_display_color_vol_offset =
            p_niframe->data.frame.sei_hdr_mastering_display_color_vol_len =
                p_niframe->data.frame.sei_hdr_content_light_level_info_offset =
                    p_niframe->data.frame.sei_hdr_content_light_level_info_len =
                        p_niframe->data.frame.sei_hdr_plus_offset =
                            p_niframe->data.frame.sei_hdr_plus_len = 0;

    p_niframe->data.frame.roi_len = 0;
    p_niframe->data.frame.reconf_len = 0;
    p_niframe->data.frame.force_pic_qp = 0;

    return ret;
}

int ni_quadra_encode_send_prepare(ni_quadra_enc_context_t *p_enc_ctx,
                                  niquadraFrameSurface1_t *p_surface,
                                  int frame_width, int frame_height)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    ni_session_data_io_t *p_niframe = p_enc_ctx->api_fme;
    ni_frame_t *p_input_frame = p_enc_ctx->input_frame;
    ni_aux_data_t *aux_data = NULL;
    int ret = 0;
    memset(p_input_frame, 0, sizeof(ni_frame_t));
    p_enc_ctx->need_free_frame = 0;
    //pass the frame_pts to niframe
    p_niframe->data.frame.pts = p_enc_ctx->frame_pts;
    p_niframe->data.frame.dts = p_enc_ctx->pkt_dts;
    if (!p_enc_ctx->eos_fme_received)
    {
        int8_t bit_depth = 1;
        if (!p_enc_ctx->is_swframes)
        {
            if (p_session_ctx->pixel_format == NI_PIX_FMT_RGBA ||
                p_session_ctx->pixel_format == NI_PIX_FMT_BGRA ||
                p_session_ctx->pixel_format == NI_PIX_FMT_ARGB ||
                p_session_ctx->pixel_format == NI_PIX_FMT_ABGR)
                bit_depth = 1;
            else
                bit_depth = p_surface->bit_depth;

            switch (bit_depth)
            {
                case 1:
                case 2:
                    break;
                default:
                    ni_log(NI_LOG_ERROR, "ERROR: Unknown bit depth %d!\n",
                           bit_depth);
                    return NI_RETCODE_FAILURE;
            }
        } else
        {
            if (NI_QUADRA_PIX_FMT_YUV420P10BE == p_enc_ctx->sw_pix_fmt ||
                NI_QUADRA_PIX_FMT_YUV420P10LE == p_enc_ctx->sw_pix_fmt ||
                NI_QUADRA_PIX_FMT_P010LE == p_enc_ctx->sw_pix_fmt)
            {
                bit_depth = 2;
            }
        }
        if ((frame_height && frame_width) &&
                (frame_height != p_enc_ctx->height ||
                 frame_width != p_enc_ctx->width) ||
            bit_depth != p_session_ctx->bit_depth_factor ||
            p_enc_ctx->color_change == 1)
        {
            ni_log(NI_LOG_INFO,
                   "xcoder_send_frame resolution change %dx%d "
                   "-> %dx%d or bit depth change %d -> %d "
                   "or color changed 0 -> %d\n",
                   p_enc_ctx->width, p_enc_ctx->height, frame_width,
                   frame_height, p_session_ctx->bit_depth_factor,
                   bit_depth, p_enc_ctx->color_change);
            p_session_ctx->session_run_state =
                SESSION_RUN_STATE_SEQ_CHANGE_DRAINING;
            p_enc_ctx->eos_fme_received = 1;
            if (p_enc_ctx->is_fifo_empty)
            {
                p_enc_ctx->need_free_frame = 1;
            }
        }
    }

    if (SESSION_RUN_STATE_SEQ_CHANGE_DRAINING ==
            p_session_ctx->session_run_state ||
        (p_enc_ctx->eos_fme_received && p_enc_ctx->is_fifo_empty))
    {
        ni_log(NI_LOG_DEBUG, "XCoder start flushing\n");
        p_niframe->data.frame.end_of_stream = 1;
        p_enc_ctx->encoder_flushing = 1;
    } else
    {
        // extra data starts with metadata header, various aux data sizes
        // have been reset above
        p_niframe->data.frame.extra_data_len = NI_APP_ENC_FRAME_META_DATA_SIZE;
        p_niframe->data.frame.ni_pict_type = 0;
        p_enc_ctx->reconf_demo_mode = XCODER_TEST_RECONF_OFF;
        ret = ni_enc_prep_reconf_demo_data(p_enc_ctx->api_ctx, p_input_frame);
        if (ret < 0)
        {
          return ret;
        }
        
        // support VFR
        if (p_param->enable_vfr)
        {
            uint32_t cur_fps, pre_fps;

            pre_fps = p_session_ctx->prev_fps;

            if (p_enc_ctx->frame_pts > p_session_ctx->prev_pts)
            {
                p_session_ctx->passed_time_in_timebase_unit +=
                    p_enc_ctx->frame_pts - p_session_ctx->prev_pts;
                p_session_ctx->count_frame_num_in_sec++;

                // change the FrameRate for VFR
                // 1. Only when the fps change, setting the new bitrate
                // 2. The interval between two framerate change settings shall be
                //    greater than 1 seconds,  or at the start of the transcoding
                if (p_session_ctx->passed_time_in_timebase_unit >=
                    (p_enc_ctx->timebase_den / p_enc_ctx->timebase_num))
                {
                    cur_fps = p_session_ctx->count_frame_num_in_sec;
                    if ((p_session_ctx->frame_num != 0) &&
                        (pre_fps != cur_fps) &&
                        ((p_session_ctx->frame_num <
                          p_param->cfg_enc_params.frame_rate) ||
                         (p_session_ctx->frame_num -
                              p_session_ctx->last_change_framenum >=
                          p_param->cfg_enc_params.frame_rate)))
                    {
                        aux_data = ni_frame_new_aux_data(
                            p_input_frame, NI_FRAME_AUX_DATA_FRAMERATE,
                            sizeof(ni_framerate_t));
                        if (aux_data)
                        {
                            ni_framerate_t *framerate =
                                (ni_framerate_t *)aux_data->data;
                            framerate->framerate_num = cur_fps;
                            framerate->framerate_denom = 1;
                        }

                        p_session_ctx->last_change_framenum =
                            p_session_ctx->frame_num;
                        p_session_ctx->prev_fps = cur_fps;
                    }
                    p_session_ctx->count_frame_num_in_sec = 0;
                    p_session_ctx->passed_time_in_timebase_unit = 0;
                }
                p_session_ctx->prev_pts = p_enc_ctx->frame_pts;
            } else if (p_enc_ctx->frame_pts < p_session_ctx->prev_pts)
            {
                // Error handle for the case that pts jump back,
                // this may cause a little error in the bitrate setting which is
                // acceptable.
                // As long as the subsequent PTS is normal, it will be back to
                // normal quickly.
                p_session_ctx->prev_pts = p_enc_ctx->frame_pts;
            } else
            {
                // do nothing, when the pts of two adjacent frames are the same;
                // this may cause a little error in the bitrate setting which is
                // acceptable.
                // As long as the subsequent PTS is normal, it will be back to
                // normal quickly.
            }
        }

        // force pic qp demo mode: initial QP (200 frames) -> QP value specified by
        // ForcePicQpDemoMode (100 frames) -> initial QP (remaining frames)
        if (p_param->force_pic_qp_demo_mode)
        {
            if (p_session_ctx->frame_num >= 300)
            {
                p_niframe->data.frame.force_pic_qp =
                    p_param->cfg_enc_params.rc.intra_qp;
            } else if (p_session_ctx->frame_num >= 200)
            {
                p_niframe->data.frame.force_pic_qp =
                    p_param->force_pic_qp_demo_mode;
            }
        }
        // END NETINT_INTERNAL - currently only for internal testing

        // roi data, if enabled, with demo mode
        // Quadra case: when ROI demo enabled, always have a data buffer for ROI
        // Note: when ROI demo modes enabled, supply ROI map for the specified range
        //       frames, and 0 map for others
        if (p_param->roi_demo_mode && p_param->cfg_enc_params.roi_enable)
        {
            if (p_session_ctx->frame_num > 90 && p_session_ctx->frame_num < 300)
            {
                p_niframe->data.frame.roi_len = p_session_ctx->roi_len;
            } else
            {
                p_niframe->data.frame.roi_len = 0;
            }
            // when ROI enabled, always have a data buffer for ROI
            // Note: this is handled separately from ROI through side/aux data
            p_niframe->data.frame.extra_data_len += p_session_ctx->roi_len;
        }

        p_enc_ctx->enable_all_sei_passthru = 
                        p_param->cfg_enc_params.enable_all_sei_passthru;
        if (!p_param->cfg_enc_params.enable_all_sei_passthru)
        {
            // SEI (HDR)
            // content light level info (only user-set since FFmpeg 3.1.1 has no support
            // of that frame side data type yet !)
            p_enc_ctx->HDR10CLLEnable = p_param->cfg_enc_params.HDR10CLLEnable;
            if (p_param->cfg_enc_params.HDR10CLLEnable &&
                (NI_CODEC_FORMAT_H264 == p_enc_ctx->codec_format ||
                p_session_ctx->bit_depth_factor == 1) &&
                p_session_ctx->light_level_data_len == 0)
            {
                // User input maxCLL so create SEIs for h264 and don't touch for (h265
                // && hdr10) since that is conveyed in config step
                // Quadra autoset only for hdr10 format with hevc
                aux_data = ni_frame_new_aux_data(
                    p_input_frame, NI_FRAME_AUX_DATA_CONTENT_LIGHT_LEVEL,
                    sizeof(ni_content_light_level_t));
                if (aux_data)
                {
                    ni_content_light_level_t *cll =
                        (ni_content_light_level_t *)(aux_data->data);

                    cll->max_cll = p_param->cfg_enc_params.HDR10MaxLight;
                    cll->max_fall = p_param->cfg_enc_params.HDR10AveLight;
                }
            }
            p_enc_ctx->HDR10Enable = p_param->cfg_enc_params.HDR10Enable;
            // mastering display color volume
            if (!p_param->cfg_enc_params.HDR10Enable)   // not user set
            {
                p_enc_ctx->HDR10Enable = p_param->cfg_enc_params.HDR10Enable;
            } else if ((NI_CODEC_FORMAT_H264 == p_enc_ctx->codec_format ||
                        p_session_ctx->bit_depth_factor == 1) &&
                    p_session_ctx->sei_hdr_mastering_display_color_vol_len == 0)
            {
                // User input masterDisplay so create SEIs for h264 and don't touch for
                // (h265 && hdr10) since that is conveyed in config step
                // Quadra autoset only for hdr10 format with hevc
                aux_data = ni_frame_new_aux_data(
                    p_input_frame, NI_FRAME_AUX_DATA_MASTERING_DISPLAY_METADATA,
                    sizeof(ni_mastering_display_metadata_t));
                if (aux_data)
                {
                    ni_mastering_display_metadata_t *mst_dsp =
                        (ni_mastering_display_metadata_t *)(aux_data->data);

                    // X, Y display primaries for RGB channels and white point(WP) in
                    // units of 0.00002 and max, min luminance(L) values in units of
                    // 0.0001 nits, xy are denom = 50000 num = HDR10dx0/y
                    mst_dsp->display_primaries[0][0].den =
                        MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->display_primaries[0][1].den =
                        MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->display_primaries[1][0].den =
                        MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->display_primaries[1][1].den =
                        MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->display_primaries[2][0].den =
                        MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->display_primaries[2][1].den =
                        MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->white_point[0].den = MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->white_point[1].den = MASTERING_DISP_CHROMA_DEN;
                    mst_dsp->min_luminance.den = MASTERING_DISP_LUMA_DEN;
                    mst_dsp->max_luminance.den = MASTERING_DISP_LUMA_DEN;
                    // ni_mastering_display_metadata_t has to be filled with R,G,B
                    // values, in that order, while HDR10d is filled in order of G,B,R,
                    // so do the conversion here.
                    mst_dsp->display_primaries[0][0].num =
                        p_param->cfg_enc_params.HDR10dx2;
                    mst_dsp->display_primaries[0][1].num =
                        p_param->cfg_enc_params.HDR10dy2;
                    mst_dsp->display_primaries[1][0].num =
                        p_param->cfg_enc_params.HDR10dx0;
                    mst_dsp->display_primaries[1][1].num =
                        p_param->cfg_enc_params.HDR10dy0;
                    mst_dsp->display_primaries[2][0].num =
                        p_param->cfg_enc_params.HDR10dx1;
                    mst_dsp->display_primaries[2][1].num =
                        p_param->cfg_enc_params.HDR10dy1;
                    mst_dsp->white_point[0].num = p_param->cfg_enc_params.HDR10wx;
                    mst_dsp->white_point[1].num = p_param->cfg_enc_params.HDR10wy;
                    mst_dsp->min_luminance.num =
                        p_param->cfg_enc_params.HDR10minluma;
                    mst_dsp->max_luminance.num =
                        p_param->cfg_enc_params.HDR10maxluma;
                    mst_dsp->has_primaries = 1;
                    mst_dsp->has_luminance = 1;
                }
            }
        }
    }
    return ret;
}

void ni_quadra_frame_new_aux_data(ni_quadra_enc_context_t *p_enc_ctx,
                                  void *side_data, int data_size, int type)
{
    ni_aux_data_t *aux_data = NULL;
    ni_frame_t *p_input_frame = p_enc_ctx->input_frame;
    int aux_size;
    switch (type)
    {
        case NI_FRAME_AUX_DATA_LONG_TERM_REF:
            aux_size = sizeof(ni_long_term_ref_t);
            break;
        case NI_FRAME_AUX_DATA_MASTERING_DISPLAY_METADATA:
            aux_size = sizeof(ni_mastering_display_metadata_t);
            break;
        case NI_FRAME_AUX_DATA_CONTENT_LIGHT_LEVEL:
            aux_size = sizeof(ni_content_light_level_t);
            break;
        case NI_FRAME_AUX_DATA_BITRATE:
        case NI_FRAME_AUX_DATA_CRF:
        case NI_FRAME_AUX_DATA_VBV_MAX_RATE:
        case NI_FRAME_AUX_DATA_VBV_BUFFER_SIZE:
            aux_size = sizeof(int32_t);
            break;
        case NI_FRAME_AUX_DATA_CRF_FLOAT:
            aux_size = sizeof(float);
            break;
        default:
            aux_size = data_size;
            break;
    }
    if (aux_size == 0)
    {
        return;
    }
    if (type == NI_FRAME_AUX_DATA_REGIONS_OF_INTEREST)
    {
        ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
        if (p_param->roi_demo_mode || !p_param->cfg_enc_params.roi_enable)
        {
            return ;
        }
    }
    aux_data = ni_frame_new_aux_data(p_input_frame, type, aux_size);
    if (aux_data)
    {
        memcpy(aux_data->data, side_data, data_size);
    }
}
uint8_t mdcv_data[NI_MAX_SEI_DATA];
uint8_t cll_data[NI_MAX_SEI_DATA];
uint8_t cc_data[NI_MAX_SEI_DATA];
uint8_t udu_data[NI_MAX_SEI_DATA];
uint8_t hdrp_data[NI_MAX_SEI_DATA];

int ni_quadra_enc_prep_aux_data(ni_quadra_enc_context_t *p_enc_ctx,
                                void *side_data, int data_size)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    ni_session_data_io_t *p_niframe = p_enc_ctx->api_fme;
    ni_frame_t *p_input_frame = p_enc_ctx->input_frame;

    int ret = 0;
    int i, j;
    p_niframe->data.frame.ni_pict_type = 0;
    if (p_session_ctx->force_frame_type)
    {
        switch (p_enc_ctx->pict_type)
        {
            case 1:
                p_niframe->data.frame.ni_pict_type = PIC_TYPE_IDR;
                break;
            case 2:
                p_niframe->data.frame.ni_pict_type = PIC_TYPE_P;
                break;
            default:;
        }
    } else if (p_enc_ctx->pict_type == 1)
    {
        p_niframe->data.frame.force_key_frame = 1;
        p_niframe->data.frame.ni_pict_type = PIC_TYPE_IDR;
    }

    p_enc_ctx->forced_header_enable = p_param->cfg_enc_params.forced_header_enable;
    ni_log(NI_LOG_TRACE,
           "xcoder_send_frame: #%" PRIu64 " ni_pict_type %d"
           " forced_header_enable %d intraPeriod %d\n",
           p_session_ctx->frame_num, p_niframe->data.frame.ni_pict_type,
           p_param->cfg_enc_params.forced_header_enable,
           p_param->cfg_enc_params.intra_period);
    // whether should send SEI with this frame
    int send_sei_with_idr = ni_should_send_sei_with_frame(
        p_session_ctx, p_niframe->data.frame.ni_pict_type, p_param);
    // prep for auxiliary data (various SEI, ROI) in encode frame, based on the
    // data returned in decoded frame
    ni_enc_prep_aux_data(p_session_ctx, &p_niframe->data.frame, p_input_frame,
                         p_session_ctx->codec_format, send_sei_with_idr,
                         mdcv_data, cll_data, cc_data, udu_data, hdrp_data);
    if (side_data && data_size > 0)
    {
        int64_t local_pts = p_enc_ctx->frame_pts;
        uint8_t *p_src_sei_data, *p_dst_sei_data;
        int sei_size;
        uint8_t sei_type;
        int size;
        ni_custom_sei_set_t *src_custom_sei_set, *dst_custom_sei_set;
        ni_custom_sei_t *p_src_custom_sei, *p_dst_custom_sei;

        // if one picture can be skipped, nienc will send that frame but will
        // not receive packet, therefore it will skip the free in receive
        // packet as well and cause memory leak. So check the last
        // pkt_custom_sei_set has been released or not.
        dst_custom_sei_set =
            p_session_ctx->pkt_custom_sei_set[local_pts % NI_FIFO_SZ];
        if (dst_custom_sei_set)
        {
            free(dst_custom_sei_set);
        }

        /* copy the whole SEI data */
        src_custom_sei_set = (ni_custom_sei_set_t *)side_data;
        dst_custom_sei_set = malloc(sizeof(ni_custom_sei_set_t));
        if (dst_custom_sei_set == NULL)
        {
            ni_log(NI_LOG_ERROR,
                   "%s failed to allocate memory for custom sei data\n",
                   __FUNCTION__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            return ret;
        }
        memset(dst_custom_sei_set, 0, sizeof(ni_custom_sei_set_t));

        /* fill sei data */
        for (i = 0; i < src_custom_sei_set->count; i++)
        {
            int len;
            p_src_custom_sei = &src_custom_sei_set->custom_sei[i];
            sei_size = p_src_custom_sei->size;
            sei_type = p_src_custom_sei->type;
            p_src_sei_data = &p_src_custom_sei->data[0];

            p_dst_custom_sei = &dst_custom_sei_set->custom_sei[i];
            p_dst_sei_data = &p_dst_custom_sei->data[0];
            size = 0;

            // long start code
            p_dst_sei_data[size++] = 0x00;
            p_dst_sei_data[size++] = 0x00;
            p_dst_sei_data[size++] = 0x00;
            p_dst_sei_data[size++] = 0x01;

            if (NI_CODEC_FORMAT_H264 == p_enc_ctx->codec_format)
            {
                p_dst_sei_data[size++] = 0x06;   //nal type: SEI
            } else
            {
                p_dst_sei_data[size++] = 0x4e;   //nal type: SEI
                p_dst_sei_data[size++] = 0x01;
            }

            // SEI type
            p_dst_sei_data[size++] = sei_type;

            // original payload size
            len = sei_size;
            while (len >= 0)
            {
                p_dst_sei_data[size++] = len > 0xff ? 0xff : len;
                len -= 0xff;
            }

            // payload data
            for (j = 0; j < sei_size && size < NI_MAX_CUSTOM_SEI_DATA - 1; j++)
            {
                if (j >= 2 && !p_dst_sei_data[size - 2] &&
                    !p_dst_sei_data[size - 1] && p_src_sei_data[j] <= 0x03)
                {
                    /* insert 0x3 as emulation_prevention_three_byte */
                    p_dst_sei_data[size++] = 0x03;
                }
                p_dst_sei_data[size++] = p_src_sei_data[j];
            }

            if (j != sei_size)
            {
                ni_log(NI_LOG_INFO,
                       "%s: sei RBSP size exceeds limit "
                       "(%d), idx=%u, type=%u, size=%d, custom_sei_loc=%d.\n",
                       __FUNCTION__, NI_MAX_CUSTOM_SEI_DATA, i, sei_type, sei_size,
                       p_src_custom_sei->location);
                free(dst_custom_sei_set);
                break;
            }

            // trailing byte
            p_dst_sei_data[size++] = 0x80;

            p_dst_custom_sei->size = size;
            p_dst_custom_sei->type = sei_type;
            p_dst_custom_sei->location = p_src_custom_sei->location;
            ni_log(NI_LOG_TRACE,
                   "%s: custom sei idx %d type %u len %d "
                   "loc %d.\n",
                   __FUNCTION__, i, sei_type, size, p_dst_custom_sei->location);
        }

        dst_custom_sei_set->count = src_custom_sei_set->count;
        p_session_ctx->pkt_custom_sei_set[local_pts % NI_FIFO_SZ] =
            dst_custom_sei_set;
        ni_log(NI_LOG_TRACE, "%s: sei number %d pts %" PRId64 ".\n", __func__,
               dst_custom_sei_set->count, local_pts);
    }

    if (p_niframe->data.frame.sei_total_len > NI_ENC_MAX_SEI_BUF_SIZE)
    {
        ni_log(NI_LOG_ERROR,
               "xcoder_send_frame: sei total length %u exceeds maximum sei "
               "size %u.\n",
               p_niframe->data.frame.sei_total_len, NI_ENC_MAX_SEI_BUF_SIZE);
        ret = NI_RETCODE_FAILURE;
        return ret;
    }

    p_niframe->data.frame.extra_data_len += p_niframe->data.frame.sei_total_len;

    // data layout requirement: leave space for reconfig data if at least one of
    // reconfig, SEI or ROI is present
    // Note: ROI is present when enabled, so use encode config flag instead of
    //       frame's roi_len as it can be 0 indicating a 0'd ROI map setting !
    if (p_niframe->data.frame.reconf_len ||
        p_niframe->data.frame.sei_total_len ||
        p_param->cfg_enc_params.roi_enable)
    {
        p_niframe->data.frame.extra_data_len +=
            sizeof(ni_encoder_change_params_t);
    }

    p_niframe->data.frame.pts = p_enc_ctx->frame_pts;
    p_niframe->data.frame.dts = p_enc_ctx->pkt_dts;

    p_niframe->data.frame.video_width = p_enc_ctx->width;
    p_niframe->data.frame.video_height = p_enc_ctx->height;

    p_enc_ctx->auto_dl_handle = 0;
    if (p_session_ctx->auto_dl_handle != 0 ||
        (p_enc_ctx->height < NI_MIN_HEIGHT) ||
        (p_enc_ctx->width < NI_MIN_WIDTH))
    {
        p_session_ctx->hw_action = 0;
        p_enc_ctx->auto_dl_handle = 1;
        p_enc_ctx->ishwframe = 0;
    }
    return ret;
}

int ni_quadra_enc_make_niframe(ni_quadra_enc_context_t *p_enc_ctx,
                               uint8_t **data, uint8_t *nidata, int *linesize,
                               int frame_width, int frame_height,
                               NIFramesContext *nif_src_ctx)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    ni_session_data_io_t *p_niframe = p_enc_ctx->api_fme;
    ni_frame_t *p_input_frame = p_enc_ctx->input_frame;
    int ret;
    int need_to_copy = 1;
    int isnv12frame = (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_NV12 ||
                       p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_P010LE);
    bool alignment_2pass_wa = ((p_param->cfg_enc_params.lookAheadDepth ||
                            p_param->cfg_enc_params.crf >= 0 ||
                            p_param->cfg_enc_params.crfFloat >= 0) &&
                            p_enc_ctx->codec_format == NI_CODEC_FORMAT_H265);
    int dst_stride[NI_MAX_NUM_DATA_POINTERS] = {0};
    int height_aligned[NI_MAX_NUM_DATA_POINTERS] = {0};
    int src_height[NI_MAX_NUM_DATA_POINTERS] = {0};

    src_height[0] = frame_height;
    src_height[1] = frame_height / 2;
    src_height[2] = (isnv12frame) ? 0 : (frame_height / 2);
    if (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_RGBA ||
        p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_BGRA ||
        p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_ARGB ||
        p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_ABGR)
    {
        src_height[0] = frame_height;
        src_height[1] = 0;
        src_height[2] = 0;
        alignment_2pass_wa = 0;
    }
    ni_get_hw_yuv420p_dim(frame_width, frame_height,
                          p_session_ctx->bit_depth_factor, isnv12frame,
                          dst_stride, height_aligned);

    ni_log(NI_LOG_TRACE,
           "xcoder_send_frame frame->width %d "
           "p_session_ctx->bit_depth_factor %d dst_stride[0/1/2] %d/%d/%d\n",
           frame_width, p_session_ctx->bit_depth_factor, dst_stride[0],
           dst_stride[1], dst_stride[2]);

    /* The main reason for the problem is that when using out=sw and noautoscale=0,
    * the incoming stream initially conforms to zerocopy, so the firmware will allocate memory according to zerocopy.
    * The resolution after this changes, but the software inserts autoscale,
    * so the encoder cannot detect the change in resolution and cannot reopen.
    * However, due to ffmpeg being 64 bit aligned, So the linesize is not consistent with the linesize we initially decoded,
    * so the encoding will take the path of non-zero copy. At this time, due to lookahead>0,
    * the software will send the firmware a size larger than the originally requested size,
    * so it will not be able to send it. Moreover, at this point,
    * the firmware is unable to perceive changes in linesize and respond accordingly.
    * For fix this, we can check the linesize and disable 2-pass workaround.
    */
    if (p_param->luma_linesize)
    {
        alignment_2pass_wa = false;
    }

    if (alignment_2pass_wa && !p_enc_ctx->ishwframe)
    {
        if (isnv12frame)
        {
            // for 2-pass encode output mismatch WA, need to extend (and
            // pad) CbCr plane height, because 1st pass assume input 32
            // align
            height_aligned[1] = NIALIGN(height_aligned[0], 32) / 2;
        } else
        {
            // for 2-pass encode output mismatch WA, need to extend (and
            // pad) Cr plane height, because 1st pass assume input 32 align
            height_aligned[2] = NIALIGN(height_aligned[0], 32) / 2;
        }
    }

    // alignment(16) extra padding for H.264 encoding
    if (p_enc_ctx->ishwframe)
    {
        uint8_t *dsthw;
        const uint8_t *srchw;

        ni_frame_buffer_alloc_hwenc(&(p_niframe->data.frame), frame_width,
                                    frame_height,
                                    (int)p_niframe->data.frame.extra_data_len);
        if (!p_niframe->data.frame.p_data[3])
        {
            return NI_RETCODE_ERROR_MEM_ALOC;
        }
        dsthw = p_niframe->data.frame.p_data[3];
        srchw = (const uint8_t *)nidata;
        ni_log(NI_LOG_TRACE, "dst=%p src=%p len=%d\n", dsthw, srchw,
               p_niframe->data.frame.data_len[3]);
        memcpy(dsthw, srchw, p_niframe->data.frame.data_len[3]);
        ni_log(NI_LOG_TRACE, "frame_data[3] %p memcpy to %p\n", nidata, dsthw);
    } else   // traditional yuv transfer
    {
        ni_log(NI_LOG_TRACE,
               "%s %s %d buffered_fme.data[0] %p data[3] %p wxh %u %u "
               "dst_stride[0] %d %d  linesize[0] %d %d data[1] %p %p data[2] "
               "%p %p data[3] %p avctx(w:h:cw:ch) %u:%u\n",
               __FILE__, __FUNCTION__, __LINE__, data[0], data[3], frame_width,
               frame_height, dst_stride[0], dst_stride[1], linesize[0],
               linesize[1], data[1], data[0] + dst_stride[0] * frame_height,
               data[2], data[1] + dst_stride[1] * frame_height / 2, data[3],
               p_enc_ctx->width, p_enc_ctx->height);

        // check input resolution zero copy compatible or not
        if (ni_encoder_frame_zerocopy_check(p_session_ctx, p_param, frame_width,
                                            frame_height, linesize,
                                            false) == NI_RETCODE_SUCCESS)
        {
            need_to_copy = 0;
            // alloc metadata buffer etc. (if needed)
            ret = ni_encoder_frame_zerocopy_buffer_alloc(
                &(p_niframe->data.frame), frame_width, frame_height, linesize,
                (const uint8_t **)data,
                (int)p_niframe->data.frame.extra_data_len);
            if (ret != NI_RETCODE_SUCCESS)
                return NI_RETCODE_ERROR_MEM_ALOC;
        } else
        {
            // if linesize changes (while resolution remains the same), copy to
            // previously configured linesizes
            if (p_param->luma_linesize && p_param->chroma_linesize)
            {
                dst_stride[0] = p_param->luma_linesize;
                dst_stride[1] = dst_stride[2] = p_param->chroma_linesize;
            }

            ni_encoder_sw_frame_buffer_alloc(
                !isnv12frame, &(p_niframe->data.frame), frame_width,
                height_aligned[0], dst_stride,
                p_enc_ctx->codec_format == NI_CODEC_FORMAT_H264,
                (int)p_niframe->data.frame.extra_data_len, alignment_2pass_wa);
        }

        ni_log(NI_LOG_TRACE, "%p need_to_copy %d! pts = %ld\n",
               p_niframe->data.frame.p_buffer, need_to_copy,
               p_enc_ctx->frame_pts);

        if (!p_niframe->data.frame.p_data[0])
            return NI_RETCODE_ERROR_MEM_ALOC;

        // if this is indeed sw frame, do the YUV data layout, otherwise may
        // need to do frame download
        if (p_enc_ctx->is_swframes)
        {
            ni_log(
                NI_LOG_TRACE,
                "xcoder_send_frame: fme.data_len[0]=%d, "
                "buf_fme->linesize=%d/%d/%d, dst alloc linesize = %d/%d/%d, "
                "src height = %d/%d/%d, dst height aligned = %d/%d/%d, "
                "force_key_frame=%d, extra_data_len=%d sei_size=%d "
                "(hdr_content_light_level %u hdr_mastering_display_color_vol "
                "%u "
                "hdr10+ %u cc %u udu %u prefC %u) roi_size=%u reconf_size=%u "
                "force_pic_qp=%u "
                "use_cur_src_as_long_term_pic %u use_long_term_ref %u\n",
                p_niframe->data.frame.data_len[0], linesize[0], linesize[1],
                linesize[2], dst_stride[0], dst_stride[1], dst_stride[2],
                src_height[0], src_height[1], src_height[2], height_aligned[0],
                height_aligned[1], height_aligned[2],
                p_niframe->data.frame.force_key_frame,
                p_niframe->data.frame.extra_data_len,
                p_niframe->data.frame.sei_total_len,
                p_niframe->data.frame.sei_hdr_content_light_level_info_len,
                p_niframe->data.frame.sei_hdr_mastering_display_color_vol_len,
                p_niframe->data.frame.sei_hdr_plus_len,
                p_niframe->data.frame.sei_cc_len,
                p_niframe->data.frame.sei_user_data_unreg_len,
                p_niframe->data.frame.preferred_characteristics_data_len,
                (p_param->cfg_enc_params.roi_enable ? p_session_ctx->roi_len :
                                                      0),
                p_niframe->data.frame.reconf_len,
                p_niframe->data.frame.force_pic_qp,
                p_niframe->data.frame.use_cur_src_as_long_term_pic,
                p_niframe->data.frame.use_long_term_ref);

            // YUV part of the encoder input data layout
            if (need_to_copy)
            {
                ni_copy_hw_yuv420p(
                    (uint8_t **)(p_niframe->data.frame.p_data), data,
                    frame_width, frame_height, p_session_ctx->bit_depth_factor,
                    isnv12frame, p_param->cfg_enc_params.conf_win_right,
                    dst_stride, height_aligned, linesize, src_height);
            }
        } else
        {
            ni_session_data_io_t *p_session_data;
            ni_session_data_io_t niframe;

            ni_log(NI_LOG_DEBUG,
                   "xcoder_send_frame:Autodownload to be run: hdl: %d w: %d h: "
                   "%d\n",
                   p_session_ctx->auto_dl_handle, p_enc_ctx->width,
                   p_enc_ctx->height);

            niquadraFrameSurface1_t *src_surf =
                (niquadraFrameSurface1_t *)data[3];

            if (p_enc_ctx->height < NI_MIN_HEIGHT ||
                p_enc_ctx->width < NI_MIN_WIDTH)
            {
                int bit_depth;
                int is_planar;

                p_session_data = &niframe;
                memset(&niframe, 0, sizeof(niframe));
                bit_depth =
                    ((p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_YUV420P10LE) ||
                     (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_P010LE)) ?
                    2 :
                    1;
                is_planar =
                    (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_YUV420P) ||
                    (p_enc_ctx->sw_pix_fmt == NI_QUADRA_PIX_FMT_YUV420P10LE);

                /* Allocate a minimal frame */
                ni_frame_buffer_alloc(&niframe.data.frame, p_enc_ctx->width,
                                      p_enc_ctx->height, 0, /* alignment */
                                      1,                    /* metadata */
                                      bit_depth, 0,         /* hw_frame_count */
                                      is_planar);
            } else
            {
                p_session_data = p_niframe;
            }
            ni_quadra_filter_session_dl_set(nif_src_ctx->api_ctx, true);
            ret = ni_device_session_hwdl(nif_src_ctx->api_ctx, p_session_data,
                                         (niFrameSurface1_t *)src_surf);
            p_enc_ctx->ishwframe = false;
            if (ret <= 0)
            {
                ni_log(NI_LOG_ERROR,
                       "nienc.c:ni_hwdl_frame() failed to retrieve frame\n");
                return NI_RETCODE_FAILURE;
            }

            if ((p_enc_ctx->height < NI_MIN_HEIGHT) ||
                (p_enc_ctx->width < NI_MIN_WIDTH))
            {
                ni_expand_frame(&p_niframe->data.frame,
                                &p_session_data->data.frame, dst_stride,
                                p_enc_ctx->width, p_enc_ctx->height,
                                p_session_ctx->pixel_format, p_enc_ctx->nb_planes);

                ni_frame_buffer_free(&niframe.data.frame);
            }
        }
    }   // traditional yuv transfer
    ni_enc_copy_aux_data(p_session_ctx, &p_niframe->data.frame, p_input_frame,
                         p_session_ctx->codec_format, mdcv_data, cll_data,
                         cc_data, udu_data, hdrp_data, p_enc_ctx->ishwframe,
                         isnv12frame);
    ni_frame_buffer_free(p_input_frame);

    return 0;
}

int ni_quadra_enc_send_frame(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_session_data_io_t *p_niframe = p_enc_ctx->api_fme;
    int sent = 0;
    sent = ni_device_session_write(p_session_ctx, p_niframe,
                                   NI_DEVICE_TYPE_ENCODER);
    ni_log(NI_LOG_DEBUG,
           "ni_quadra_enc_send_frame: pts %lld dts %lld "
           "size %d sent to xcoder\n",
           p_niframe->data.frame.pts, p_niframe->data.frame.dts, sent);
    // return EIO at error
    if (NI_RETCODE_ERROR_VPU_RECOVERY == sent)
    {
        p_enc_ctx->retstatus = NI_QUADRA_ERROR_VPU_RECOVERY;
    }
    else if (sent <= 0)
    {
        p_enc_ctx->start_of_stream = p_niframe->data.frame.start_of_stream;
    }
    else
    {
        p_enc_ctx->auto_dl_handle = p_session_ctx->auto_dl_handle;

        // pushing input pts in circular FIFO
        p_session_ctx->enc_pts_list[p_session_ctx->enc_pts_w_idx % NI_FIFO_SZ] =
            p_niframe->data.frame.pts;
        p_session_ctx->enc_pts_w_idx++;
    }
    return sent;
}

int ni_quadra_enc_device_session_flush(ni_quadra_enc_context_t *p_enc_ctx)
{
    return ni_device_session_flush(p_enc_ctx->api_ctx, NI_DEVICE_TYPE_ENCODER);
}

int ni_quadra_encode_reinit(ni_quadra_enc_context_t *p_enc_ctx, int frame_width,
                            int frame_height, const int *linesize)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    p_enc_ctx->tmp_device_handle = p_session_ctx->device_handle;
    p_enc_ctx->tmp_blk_io_handle = p_session_ctx->blk_io_handle;
    p_enc_ctx->tmp_hw_id = p_session_ctx->hw_id;
    int stride, ori_stride;
    bool bIsSmallPicture = false;

    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;

    ni_xcoder_strncpy(p_enc_ctx->tmp_blk_dev_name, p_session_ctx->blk_dev_name,
                      NI_MAX_DEVICE_NAME_LEN);

    p_enc_ctx->eos_fme_received = 0;
    p_enc_ctx->encoder_eof = 0;
    p_enc_ctx->encoder_flushing = 0;
    p_enc_ctx->firstPktArrived = 0;
    p_enc_ctx->spsPpsArrived = 0;
    p_enc_ctx->spsPpsHdrLen = 0;
    free(p_enc_ctx->p_spsPpsHdr);
    p_enc_ctx->p_spsPpsHdr = NULL;
    p_enc_ctx->seqChangeCount++;
    // check if resolution is zero copy compatible and set linesize according to
    // new resolution
    if (ni_encoder_frame_zerocopy_check(p_session_ctx, p_param, frame_width,
                                        frame_height, linesize,
                                        true) == NI_RETCODE_SUCCESS)
    {
        stride =
            p_param->luma_linesize;   // new sequence is zero copy compatible
    } else
    {
        stride = NIALIGN(frame_width * p_enc_ctx->bit_depth, 128);
    }

    if (p_session_ctx->ori_luma_linesize && p_session_ctx->ori_chroma_linesize)
    {
        // previous sequence was zero copy compatible
        ori_stride = p_session_ctx->ori_luma_linesize;
    } else
    {
        ori_stride =
            NIALIGN(p_session_ctx->ori_width * p_enc_ctx->bit_depth, 128);
    }

    if (p_enc_ctx->pix_fmt == NI_PIX_FMT_ARGB ||
        p_enc_ctx->pix_fmt == NI_PIX_FMT_ABGR)
    {
        stride = frame_width;
        ori_stride = p_session_ctx->ori_width;
    }

    if (p_param->cfg_enc_params.lookAheadDepth ||
        p_param->cfg_enc_params.crf >= 0 ||
        p_param->cfg_enc_params.crfFloat >= 0)
    {
        ni_log(NI_LOG_DEBUG,
               "ni_quadra_encode_reinit 2-pass lookaheadDepth %d "
               "and/or CRF %d and/or CRFFloat %f\n",
               p_param->cfg_enc_params.lookAheadDepth,
               p_param->cfg_enc_params.crf,
               p_param->cfg_enc_params.crfFloat);
        if ((frame_width < NI_2PASS_ENCODE_MIN_WIDTH) ||
            (frame_height < NI_2PASS_ENCODE_MIN_HEIGHT))
        {
            bIsSmallPicture = true;
        }
    } else
    {
        if ((frame_width < NI_MIN_WIDTH) || (frame_height < NI_MIN_HEIGHT))
        {
            bIsSmallPicture = true;
        }
    }

    if (p_param->cfg_enc_params.multicoreJointMode)
    {
        ni_log(NI_LOG_DEBUG,
               "ni_quadra_encode_reinit multicore "
               "joint mode\n");
        if ((frame_width < 256) || (frame_height < 256))
        {
            bIsSmallPicture = true;
        }
    }
    p_enc_ctx->reinit = 0;
    if ((ori_stride * p_session_ctx->ori_height < stride * frame_height) ||
        (p_session_ctx->ori_pix_fmt != p_enc_ctx->pix_fmt) || bIsSmallPicture ||
        p_param->cfg_enc_params.disable_adaptive_buffers)
    {
        p_enc_ctx->reinit = 1;
    }

    return 0;
}
int ni_quadra_encode_after_reinit(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    // keep device handle(s) open during sequence change to fix mem bin buffer
    // not recycled
    p_session_ctx->device_handle = p_enc_ctx->tmp_device_handle;
    p_session_ctx->blk_io_handle = p_enc_ctx->tmp_blk_io_handle;
    p_session_ctx->hw_id = p_enc_ctx->tmp_hw_id;
    ni_xcoder_strncpy(p_session_ctx->blk_dev_name, p_enc_ctx->tmp_blk_dev_name,
                      NI_MAX_DEVICE_NAME_LEN);
    // this state is referenced when sending first frame after sequence change
    p_session_ctx->session_run_state = SESSION_RUN_STATE_SEQ_CHANGE_OPENING;
    return 0;
}

int ni_quadra_encoder_receive_packet(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_session_data_io_t *p_packet = p_enc_ctx->api_pkt;
    ni_xcoder_params_t *p_param = p_enc_ctx->api_param;
    int i, ret = 0;
    int recv;
    ni_packet_t *xpkt = &p_packet->data.packet;

    while (1)
    {
        p_enc_ctx->retstatus = NI_QUADRA_RET_SUCCESS;
        xpkt->recycle_index = -1;
        recv = ni_device_session_read(p_session_ctx, p_packet,
                                      NI_DEVICE_TYPE_ENCODER);
        ni_log(
            NI_LOG_TRACE,
            "XCoder receive packet: xpkt.end_of_stream=%d, xpkt.data_len=%d, "
            "recv=%d, encoder_flushing=%d, encoder_eof=%d\n",
            xpkt->end_of_stream, xpkt->data_len, recv,
            p_enc_ctx->encoder_flushing, p_enc_ctx->encoder_eof);
        p_enc_ctx->end_of_stream = xpkt->end_of_stream;

        if (recv <= 0)
        {
            p_enc_ctx->encoder_eof = xpkt->end_of_stream;
            /* not ready ?? */
            if (p_enc_ctx->encoder_eof)
            {
                if (SESSION_RUN_STATE_SEQ_CHANGE_DRAINING ==
                    p_session_ctx->session_run_state)
                {
                    // after sequence change completes, reset codec state
                    ni_log(NI_LOG_INFO,
                           "xcoder_receive_packet 1: sequence "
                           "change completed, return AVERROR(EAGAIN) and will "
                           "reopen codec!\n");
                    p_enc_ctx->retstatus = NI_QUADRA_SEQ_CHANGE_DRAINING;
                }

                ni_log(NI_LOG_TRACE,
                       "xcoder_receive_packet: got encoder_eof, "
                       "return AVERROR_EOF\n");
                break;
            } else
            {
                if (NI_RETCODE_ERROR_VPU_RECOVERY == recv)
                {
                    p_enc_ctx->retstatus = NI_QUADRA_ERROR_VPU_RECOVERY;
                    break;
                } else if (recv < 0)
                {
                    if ((NI_RETCODE_ERROR_INVALID_SESSION == recv) &&
                        !p_enc_ctx->started)   // session may be in recovery state, return
                                               // EAGAIN
                    {
                        p_enc_ctx->retstatus = NI_QUADRA_ERROR_INVALID_SESSION;
                    } else
                    {
                        ;
                    }
                    p_enc_ctx->gotPacket = 0;
                    p_enc_ctx->sentFrame = 0;
                    break;
                } else if (p_param->low_delay_mode && p_enc_ctx->sentFrame &&
                           !p_enc_ctx->gotPacket)
                {
                    ni_log(NI_LOG_TRACE,
                           "XCoder receive packet: low delay mode,"
                           " keep reading until pkt arrives\n");
                    continue;
                }

                p_enc_ctx->gotPacket = 0;
                p_enc_ctx->sentFrame = 0;
                p_enc_ctx->retstatus = NI_QUADRA_EAGAIN;
                if (!p_enc_ctx->encoder_flushing &&
                    !p_enc_ctx->eos_fme_received)
                {
                    ni_log(NI_LOG_TRACE,
                           "xcoder_receive_packet: NOT "
                           "encoder_flushing, return AVERROR(EAGAIN)\n");
                    break;
                }
            }
        } else
        {
            p_enc_ctx->need_to_recycle = false;
            /* got encoded data back */
            int meta_size = p_session_ctx->meta_size;
            if (p_enc_ctx->is_swframes == 0 && xpkt->recycle_index >= 0 &&
                xpkt->recycle_index <
                    NI_GET_MAX_HWDESC_FRAME_INDEX(p_session_ctx->ddr_config) &&
                p_enc_ctx->height >= NI_MIN_HEIGHT &&
                p_enc_ctx->width >= NI_MIN_WIDTH)
            {
                p_enc_ctx->recycle_index = xpkt->recycle_index;
                p_enc_ctx->need_to_recycle = true;
                xpkt->recycle_index = -1;
            }

            if (!p_enc_ctx->spsPpsArrived)
            {
                p_enc_ctx->retstatus = NI_QUADRA_EAGAIN;
                //   ret                = AVERROR(EAGAIN);
                p_enc_ctx->spsPpsArrived = 1;
                p_enc_ctx->spsPpsHdrLen = recv - meta_size;
                p_enc_ctx->p_spsPpsHdr = malloc(p_enc_ctx->spsPpsHdrLen);
                if (!p_enc_ctx->p_spsPpsHdr)
                {
                    p_enc_ctx->retstatus = NI_QUADRA_ERROR;
                    //   ret = NI_RETCODE_ERROR_MEM_ALOC;
                    break;
                }

                memcpy(p_enc_ctx->p_spsPpsHdr,
                       (uint8_t *)xpkt->p_data + meta_size,
                       xpkt->data_len - meta_size);

                // start pkt_num counter from 1 to get the real first frame
                p_session_ctx->pkt_num = 1;
                // for low-latency mode, keep reading until the first frame is
                // back
                if (p_param->low_delay_mode)
                {
                    ni_log(NI_LOG_TRACE,
                           "XCoder receive packet: low delay mode,"
                           " keep reading until 1st pkt arrives\n");
                    continue;
                }
                break;
            }
            // handle pic skip
            if (xpkt->frame_type == 3)   // 0=I, 1=P, 2=B, 3=not coded / skip
            {
                p_enc_ctx->retstatus = NI_QUADRA_EAGAIN;
                // ret = AVERROR(EAGAIN);
                if (p_enc_ctx->first_frame_pts == INT_MIN)
                    p_enc_ctx->first_frame_pts = xpkt->pts;
                /* no support for AV1 now
                if (AV_CODEC_ID_AV1 == avctx->codec_id) {
                    ctx->latest_dts = xpkt->pts;
                    } else */
                if (p_enc_ctx->total_frames_received < p_enc_ctx->dts_offset)
                {
                    // guess dts
                    int64_t bframes_delay = 0;
                    if (p_session_ctx->enc_pts_w_idx > p_enc_ctx->dts_offset)
                        bframes_delay = p_session_ctx->enc_pts_list[p_enc_ctx->dts_offset] - p_session_ctx->enc_pts_list[0];
                    ni_log(NI_LOG_INFO, "%s xcoder_bframes_delay=%"PRId64"\n", __func__, bframes_delay);
                    p_enc_ctx->latest_dts = p_enc_ctx->first_frame_pts +
                        ((p_enc_ctx->gop_offset_count - p_enc_ctx->dts_offset) * (bframes_delay ? bframes_delay / p_enc_ctx->dts_offset : p_enc_ctx->ticks_per_frame));
                    p_enc_ctx->gop_offset_count++;
                } else
                {
                    // get dts from pts FIFO
                    p_enc_ctx->latest_dts =
                        p_session_ctx
                            ->enc_pts_list[p_session_ctx->enc_pts_r_idx %
                                           NI_FIFO_SZ];
                    p_session_ctx->enc_pts_r_idx++;
                }
                if (p_enc_ctx->latest_dts > xpkt->pts)
                {
                    p_enc_ctx->latest_dts = xpkt->pts;
                }
                p_enc_ctx->total_frames_received++;

                if (!p_enc_ctx->encoder_flushing &&
                    !p_enc_ctx->eos_fme_received)
                {
                    ni_log(NI_LOG_TRACE,
                           "xcoder_receive_packet: skip"
                           " picture output, return AVERROR(EAGAIN)\n");
                    break;
                } else
                    continue;
            }

            p_enc_ctx->gotPacket = 1;
            p_enc_ctx->sentFrame = 0;

            uint32_t nalu_type = 0;
            const uint8_t *p_start_code;
            uint32_t stc = -1;
            p_enc_ctx->copy_len = 0;
            uint8_t *p_src = (uint8_t *)xpkt->p_data + meta_size;
            uint8_t *p_end = p_src + (xpkt->data_len - meta_size);
            p_enc_ctx->local_pts = xpkt->pts;
            p_enc_ctx->total_custom_sei_size = 0;
            p_enc_ctx->custom_sei_count = 0;
            ni_custom_sei_set_t *p_custom_sei_set;
            p_enc_ctx->data_len = xpkt->data_len - meta_size;
            p_custom_sei_set =
                p_session_ctx
                    ->pkt_custom_sei_set[p_enc_ctx->local_pts % NI_FIFO_SZ];
            if (p_custom_sei_set != NULL)
            {
                p_enc_ctx->custom_sei_count = p_custom_sei_set->count;
                for (i = 0; i < p_custom_sei_set->count; i++)
                {
                    p_enc_ctx->total_custom_sei_size +=
                        p_custom_sei_set->custom_sei[i].size;
                }
            }

            if (p_enc_ctx->custom_sei_count)
            {
                // if HRD or custom sei enabled, search for pic_timing or custom
                // SEI insertion point by skipping non-VCL until video data is
                // found.
                p_start_code = p_src;
                if (NI_CODEC_FORMAT_H265 == p_enc_ctx->codec_format)
                {
                    do
                    {
                        stc = -1;
                        p_start_code =
                            ni_find_start_code(p_start_code, p_end, &stc);
                        nalu_type = (stc >> 1) & 0x3F;
                    } while (nalu_type > 31 /* HEVC_NAL_RSV_VCL31 */);

                    // calc. length to copy
                    p_enc_ctx->copy_len = p_start_code - 5 - p_src;
                } else if (NI_CODEC_FORMAT_H264 == p_enc_ctx->codec_format)
                {
                    do
                    {
                        stc = -1;
                        p_start_code =
                            ni_find_start_code(p_start_code, p_end, &stc);
                        nalu_type = stc & 0x1F;
                    } while (nalu_type > 5 /* H264_NAL_IDR_SLICE */);

                    // calc. length to copy
                    p_enc_ctx->copy_len = p_start_code - 5 - p_src;
                } else
                {
                    ni_log(NI_LOG_ERROR,
                           "%s: codec %d not supported for "
                           "SEI !\n",
                           __FUNCTION__, p_enc_ctx->codec_format);
                }
            }
#define NI_CODEC_ID_MJPEG 8
            if (p_enc_ctx->codec_id == NI_CODEC_ID_MJPEG &&
                (!p_enc_ctx->firstPktArrived))
            {
                // there is no header for Jpeg, so skip header copy
                p_enc_ctx->firstPktArrived = 1;
                p_enc_ctx->first_frame_pts = xpkt->pts;
            }
#undef NI_CODEC_ID_MJPEG
            if (!p_enc_ctx->firstPktArrived)
            {
                //    p_enc_ctx->firstPktArrived = 1;
                p_enc_ctx->first_frame_pts = xpkt->pts;
            }

            break;
        }
    }
    return recv;
}

int ni_quadra_enc_get_extra_data(ni_quadra_enc_context_t *p_enc_ctx,
                                 uint8_t *p_dst)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_session_data_io_t *p_packet = p_enc_ctx->api_pkt;
    ni_packet_t *xpkt = &p_packet->data.packet;
    uint8_t *p_src = (uint8_t *)xpkt->p_data + p_session_ctx->meta_size;
    int i;
    ni_custom_sei_set_t *p_custom_sei_set;
    p_custom_sei_set =
        p_session_ctx->pkt_custom_sei_set[p_enc_ctx->local_pts % NI_FIFO_SZ];
    if (p_enc_ctx->custom_sei_count)
    {
        // copy buf_period
        memcpy(p_dst, p_src, p_enc_ctx->copy_len);
        p_dst += p_enc_ctx->copy_len;

        for (i = 0; i < p_enc_ctx->custom_sei_count; i++)
        {
            // copy custom sei
            ni_custom_sei_t *p_custom_sei = &p_custom_sei_set->custom_sei[i];
            if (p_custom_sei->location == NI_CUSTOM_SEI_LOC_AFTER_VCL)
            {
                break;
            }
            memcpy(p_dst, &p_custom_sei->data[0], p_custom_sei->size);
            p_dst += p_custom_sei->size;
        }

        // copy the IDR data
        memcpy(p_dst, p_src + p_enc_ctx->copy_len,
               xpkt->data_len - p_session_ctx->meta_size - p_enc_ctx->copy_len);
        p_dst +=
            xpkt->data_len - p_session_ctx->meta_size - p_enc_ctx->copy_len;

        // copy custom sei after slice
        for (; i < p_enc_ctx->custom_sei_count; i++)
        {
            ni_custom_sei_t *p_custom_sei = &p_custom_sei_set->custom_sei[i];
            memcpy(p_dst, &p_custom_sei->data[0], p_custom_sei->size);
            p_dst += p_custom_sei->size;
        }
    } else
    {
        memcpy(p_dst, (uint8_t *)xpkt->p_data + p_session_ctx->meta_size,
               xpkt->data_len - p_session_ctx->meta_size);
    }
    return 0;
}

int ni_quadra_encoder_prepare_avpkt(ni_quadra_enc_context_t *p_enc_ctx,
                                    int is_pkt, int64_t *pkt_pts,
                                    int64_t *pkt_dts, int *pkt_flag)
{
    ni_session_context_t *p_session_ctx = p_enc_ctx->api_ctx;
    ni_session_data_io_t *p_packet = p_enc_ctx->api_pkt;
    ni_packet_t *xpkt = &p_packet->data.packet;
    ni_custom_sei_set_t *p_custom_sei_set;
    p_custom_sei_set =
        p_session_ctx->pkt_custom_sei_set[p_enc_ctx->local_pts % NI_FIFO_SZ];
    // free buffer
    if (p_enc_ctx->custom_sei_count)
    {
        free(p_custom_sei_set);
        p_session_ctx->pkt_custom_sei_set[p_enc_ctx->local_pts % NI_FIFO_SZ] =
            NULL;
    }

    if (!is_pkt)
    {
        if (PIC_TYPE_I == xpkt->frame_type)
        {
            *pkt_flag |= 0x0001;
        }
        *pkt_pts = xpkt->pts;
        /* to ensure pts>dts for all frames, we assign a guess pts for the
        * first 'dtsOffset' frames and then the pts from input stream is
        * extracted from input pts FIFO. if GOP = IBBBP and PTSs = 0 1 2
        * 3 4 5 .. then out DTSs = -3 -2 -1 0 1 ... and -3 -2 -1 are the
        * guessed values if GOP = IBPBP and PTSs = 0 1 2 3 4 5 .. then
        * out DTSs = -1 0 1 2 3 ... and -1 is the guessed value the
        * number of guessed values is equal to dtsOffset
        */
        if (p_enc_ctx->total_frames_received < p_enc_ctx->dts_offset)
        {
            // guess dts
            int64_t bframes_delay = 0;
            if (p_session_ctx->enc_pts_w_idx > p_enc_ctx->dts_offset)
                bframes_delay = p_session_ctx->enc_pts_list[p_enc_ctx->dts_offset] - p_session_ctx->enc_pts_list[0];
            ni_log(NI_LOG_INFO, "%s xcoder_bframes_delay=%"PRId64"\n", __func__, bframes_delay);
            *pkt_dts = p_enc_ctx->first_frame_pts +
            ((p_enc_ctx->gop_offset_count - p_enc_ctx->dts_offset) * (bframes_delay ? bframes_delay / p_enc_ctx->dts_offset : p_enc_ctx->ticks_per_frame));
            p_enc_ctx->gop_offset_count++;
        } else
        {
            // get dts from pts FIFO
            *pkt_dts =
                p_session_ctx
                    ->enc_pts_list[p_session_ctx->enc_pts_r_idx % NI_FIFO_SZ];
            p_session_ctx->enc_pts_r_idx++;
        }
        if (p_enc_ctx->total_frames_received >= 1)
        {
            if (*pkt_dts < p_enc_ctx->latest_dts)
            {
                ni_log(NI_LOG_INFO, "dts: %ld < latest_dts: %ld.\n", *pkt_dts,
                       p_enc_ctx->latest_dts);
            }
        }

        if (*pkt_pts < p_enc_ctx->first_frame_pts)
        {
            ni_log(NI_LOG_INFO, "pts %ld less than first frame pts %ld. Force it to first frame pts\n",
                    *pkt_pts, p_enc_ctx->first_frame_pts);
            *pkt_pts = p_enc_ctx->first_frame_pts;
        }
        if (*pkt_dts > *pkt_pts)
        {
            ni_log(NI_LOG_INFO, "dts: %ld, pts: %ld. Forcing dts = pts \n",
                   *pkt_dts, *pkt_pts);
            *pkt_dts = *pkt_pts;
        }

        // For test script video checker purpose
        ni_log(NI_LOG_DEBUG,
               "XCoder recv pkt #%" PRId64 ""
               " pts %" PRId64 "  dts %" PRId64 "  frame_type %u avg qp %u\n",
               p_session_ctx->pkt_num - 1, *pkt_dts, *pkt_pts, xpkt->frame_type,
               xpkt->avg_frame_qp);

        p_enc_ctx->total_frames_received++;
        p_enc_ctx->latest_dts = *pkt_dts;
    }
    p_enc_ctx->encoder_eof = xpkt->end_of_stream;
    if (SESSION_RUN_STATE_SEQ_CHANGE_DRAINING ==
        p_session_ctx->session_run_state)
    {
        // p_enc_ctx->retstatus = NI_QUADRA_SEQ_CHANGE_DRAINING;
    }
    return 0;
}

void ni_quadra_set_session_wh(void *session_ctx, int width, int height)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    p_session_ctx->ori_width = width;
    p_session_ctx->ori_height = height;
}

int ni_quadra_packet_buffer_alloc(ni_quadra_enc_context_t *p_enc_ctx)
{
    ni_session_data_io_t *p_packet = p_enc_ctx->api_pkt;
    int ret = 0;
    ni_packet_t *xpkt = &p_packet->data.packet;
    if ((ret = ni_packet_buffer_alloc(xpkt, NI_MAX_TX_SZ))!= NI_RETCODE_SUCCESS)
    {
        ni_log(
            NI_LOG_ERROR,
            "ni_quadra_packet_buffer_alloc: packet buffer size %d allocation failed\n",
            NI_MAX_TX_SZ);
        return NI_RETCODE_ERROR_MEM_ALOC;
     }
     return ret;
}

int ni_quadra_get_pkt_info_size(void)
{
    return sizeof(ni_pkt_info);
}

void ni_quadra_get_pkt_info(void *session_ctx, void *data)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    ni_pkt_info *pkt_info = data;
    pkt_info->psnr_y = p_session_ctx->psnr_y;
    pkt_info->psnr_u = p_session_ctx->psnr_u;
    pkt_info->psnr_v = p_session_ctx->psnr_v;
    pkt_info->average_psnr = p_session_ctx->average_psnr;
    ni_log(NI_LOG_DEBUG,"Add PSNR Y:%lf U:%lf V:%lf Ava:%lf to packet side data\n",
            p_session_ctx->psnr_y,p_session_ctx->psnr_u, p_session_ctx->psnr_v,p_session_ctx->average_psnr);
}
