/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
*  \file   ni_quadra_enc_api.h
*
*  \brief  NETINT encoder API header file
*
*******************************************************************************/

#pragma once

#ifdef __cplusplus
extern "C" {
#endif

// #include "ni_defs.h"
// #include "ni_rsrc_api.h"
// #include "ni_util.h"
// #include "ni_device_api.h"
// #include "ni_av_codec.h"
#include "ni_quadra_filter_api.h"
#include <limits.h>

// copy maximum number of bytes of a string from src to dst, ensuring null byte
// terminated
static inline void ni_xcoder_strncpy(char *dst, const char *src, int max)
{
    if (dst && src && max)
    {
        *dst = '\0';
        strncpy(dst, src, max);
        *(dst + max - 1) = '\0';
    }
}
#define NIALIGN(x, a) (((x) + (a)-1) & ~((a)-1))
#define NIODD2EVEN(X) ((X & 1) && (X > 31)) ? (X + 1) : (X)
/**
 * Place global headers in extradata instead of every keyframe.
 */
#define NI_CODEC_FLAG_GLOBAL_HEADER (1 << 22)

#define QUADRA_XCODER_TEST_RECONF_LONG_TERM_REF 4

/*!*
* \brief Session running state type.
*/
typedef enum _ni_quadra_session_run_state
{
    QUADRA_SESSION_RUN_STATE_NORMAL = 0,
    QUADRA_SESSION_RUN_STATE_SEQ_CHANGE_DRAINING = 1,
    QUADRA_SESSION_RUN_STATE_SEQ_CHANGE_OPENING = 2,
    QUADRA_SESSION_RUN_STATE_RESETTING = 3,
} ni_quadra_session_run_state_t;

typedef struct _ni_quadra_enc_context
{
    char *dev_xcoder;
    char *dev_xcoder_name; /* dev name of the xcoder card to use */
    char *blk_xcoder_name; /* blk name of the xcoder card to use */
    int dev_enc_idx;       /* user-specified encoder index */
    char *dev_blk_name;    /* user-specified encoder block device name */
    int keep_alive_timeout;
    uint64_t xcode_load_pixel; /* xcode load in pixels by this encode task */
    char tmp_blk_dev_name[32];
    int32_t tmp_device_handle;
    int32_t tmp_blk_io_handle;
    int tmp_hw_id;
    int eos_fme_received;

    void *api_pkt; /* used for receiving bitstream from xcoder */
    void *api_fme; /* used for sending YUV data to xcoder */
    void *api_ctx;
    void *api_param;
    void *input_frame;

    uint8_t *extradata;
    int extradata_size;

    int timebase_num;
    int timebase_den;
    int framerate_num;
    int framerate_den;
    int ticks_per_frame;
    int64_t bit_rate;
    int width;
    int height;
    int ff_log_level;
    int codec_format;
    int codec_id;
    int pix_fmt; // used when don reinit
    int sw_pix_fmt; // the pix_fmt from ffmpeg
    int pict_type;
    int64_t frame_pts;
    int64_t pkt_dts;
    int auto_dl_handle;
    int ishwframe;
    int64_t local_pts;
    uint32_t data_len;

    // color metrics
    int color_primaries;
    int color_trc;
    int color_space;
    int color_range;
    int color_change;
    int user_color_primaries;
    int user_color_trc;
    int user_color_space;
    int user_color_range; // Need to keep for backward compatibility

    // sample_aspect_ratio
    int sar_num;
    int sar_den;
    int user_sar_num;
    int user_sar_den;

    /* below are all command line options */
    char *xcoder_opts;

    char *xcoder_gop;
    /* Variables that need to be initialized */
    int started;
    uint8_t *p_spsPpsHdr;
    int spsPpsHdrLen;
    int spsPpsAttach; /* Attach SPS&PPS to Packet */
    int spsPpsArrived;
    int firstPktArrived; /* specially treat first pkt of the encoded bitstream */
    int64_t dts_offset;
    int gop_offset_count; /*this is a counter to guess the pts only dtsOffset
                                times */
    int seqChangeCount;
    uint64_t total_frames_received;
    int64_t first_frame_pts;
    int64_t latest_dts;
    /* backup copy of original values of -enc command line option */
    int orig_dev_enc_idx;
    // low delay mode flags
    int gotPacket; /* used to stop receiving packets when a packet is already
                    received */
    int sentFrame; /* used to continue receiving packets when a frame is sent and
                    a packet is not yet received */
    int nvme_io_size; /* custom nvme io size */
    int vpu_reset;
    int encoder_flushing;
    int encoder_eof;
    bool is_swframes;
    void *g_enc_change_params;
    int is_fifo_empty;
    void *aux_data;
    int reconfigCount; /* count of reconfig */
    int reconf_demo_mode;
    int use_cur_src_as_long_term_pic;
    int use_long_term_ref;
    int HDR10CLLEnable;
    int HDR10Enable;
    int start_of_stream;
    int reinit;
    int bit_depth;
    int end_of_stream;
    int recycle_index;
    bool need_to_recycle;
    ni_quadra_retcode_t retstatus;
    int total_custom_sei_size;
    int custom_sei_count;
    uint32_t copy_len;
    int need_free_frame;
    int nb_planes;
    int forced_header_enable;
    int enable_all_sei_passthru;
    bool enable_test_latency;
    bool bget_psnr;
    bool generate_enc_hdrs;
} ni_quadra_enc_context_t;

/*!*****************************************************************************
 *  \brief  Set value referenced by name in encode parameters structure
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_quadra_enc_context_t
 *  \param[in] name       String represented parameter name to search
 *  \param[in] value      Parameter value to set
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 *                        NI_QUADRA_PARAM_WARNING_DEPRECATED
 *******************************************************************************/
LIB_API int ni_quadra_encoder_xcoder_opts_params_set_value(
    ni_quadra_enc_context_t *p_enc_ctx, const char *name, const char *value);

/*!*****************************************************************************
 *  \brief  Set value referenced by name in encode parameters structure
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_quadra_enc_context_t
 *  \param[in] name       String represented parameter name to search
 *  \param[in] value      Parameter value to set
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 *                        NI_QUADRA_PARAM_WARNING_DEPRECATED
 *******************************************************************************/
LIB_API int ni_quadra_encoder_xcoder_gop_params_set_value(
    ni_quadra_enc_context_t *p_enc_ctx, const char *name, const char *value);

/*!*****************************************************************************
 *  \brief  Initialize init device session_context and init default params
 *          including ni_quadra_enc_context_t,
 *                    ni_session_context_t,
 *                    ni_xcoder_params_t
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 *******************************************************************************/
LIB_API int ni_quadra_setup_encoder(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Initialize malloc device session_context and malloc params
 *          including ni_quadra_enc_context_t,
 *                    ni_session_context_t,
 *                    ni_xcoder_params_t
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        NI_LOGAN_RETCODE_SUCCESS
 *          On failure
 *                        NI_LOGAN_RETCODE_FAILURE
 *                        NI_LOGAN_RETCODE_INVALID_PARAM,
 *                        NI_LOGAN_RETCODE_ERROR_MEM_ALOC, etc
 *******************************************************************************/
LIB_API int ni_quadra_enc_init(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Close device session_context
 *          including ni_quadra_enc_context_t,
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_quadra_enc_context_t
 *
 *  \return 0
 *
 *******************************************************************************/
LIB_API int ni_quadra_encode_close(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Set encode sequence change
 *
 *  \param[in/out] p_enc_ctx  Pointer to ni_quadra_enc_context_t
 *  \param[in] width       frame width
 *  \param[in] height      frame height
 *  \param[in] bit_depth_factor  frame bit_depth
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_encode_sequence_change(ni_quadra_enc_context_t *p_enc_ctx,
                                             int width, int height,
                                             int bit_depth_factor);

/*!*****************************************************************************
 *  \brief  Set encode session open and set niframe default info
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in] frame_width       frame width
 *  \param[in] frame_height      frame height
 *  \param[in] frame_linesize    frame linesize
 *  \param[in] p_surface         pointer to niquadraFrameSurface1_t
 *  \param[in] hw_frames_ctx     frame hw_frames_ctx is exist
 *  \param[in] frame_queqe       frame opaque
 *  \param[in] avctx_flags       avctx flags
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_encode_session_open(ni_quadra_enc_context_t *p_enc_ctx,
                                          int frame_width, int frame_height,
                                          const int *frame_linesize,
                                          niquadraFrameSurface1_t *p_surface,
                                          bool hw_frames_ctx, int frame_queqe,
                                          int avctx_flags);

/*!*****************************************************************************
 *  \brief  Add a new auxiliary data to a frame
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in/out] side_data     a frame to which the auxiliary data should be added
 *  \param[in]     data_size size of the added auxiliary data 
 *  \param[in]     type   type of the added auxiliary data
 *  
 *  \return void
 ******************************************************************************/
LIB_API void ni_quadra_frame_new_aux_data(ni_quadra_enc_context_t *p_enc_ctx,
                                          void *side_data, int data_size,
                                          int type);
/*!*****************************************************************************
 *  \brief  Prepare a niframe which to send to encoder
 *
 *  \param[in/out] p_enc_ctx         Pointer to ni_quadra_enc_context_t
 *  \param[in]     p_surface         pointer to niquadraFrameSurface1_t
 *  \param[in]     frame_width       frame width
 *  \param[in]     frame_height      frame height
 *  
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 ******************************************************************************/
LIB_API int ni_quadra_encode_send_prepare(ni_quadra_enc_context_t *p_enc_ctx,
                                          niquadraFrameSurface1_t *p_surface,
                                          int frame_width, int frame_height);

/*!*****************************************************************************
 *  \brief  Prepare auxiliary data that should be sent together with this frame
 *          to encoder based on the auxiliary data of the decoded frame.
 *
 *  \param[in/out]           Pointer to ni_quadra_enc_context_t
 *  \param[out] side_data    frame side data
 *  \param[in]  data_size    frame side data size
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 ******************************************************************************/
LIB_API int ni_quadra_enc_prep_aux_data(ni_quadra_enc_context_t *p_enc_ctx,
                                        void *side_data, int data_size);

/*!*****************************************************************************
 *  \brief  Copy avframe to niframe
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in] data              frame data
 *  \param[in] nidata            frame data[3]
 *  \param[in] linesize          frame linesize
 *  \param[in] frame_width       frame width
 *  \param[in] frame_height      frame height
 *  \param[in] nif_src_ctx       pointer to niquadraFrameSurface1_t
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_enc_make_niframe(ni_quadra_enc_context_t *p_enc_ctx,
                                       uint8_t **data, uint8_t *nidata,
                                       int *linesize, int frame_width,
                                       int frame_height,
                                       NIFramesContext *nif_src_ctx);

/*!*****************************************************************************
 *  \brief  Send niframe to do encoder
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_enc_send_frame(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Flush encode session context
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int
ni_quadra_enc_device_session_flush(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Reinit encode context
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in] linesize          frame linesize
 *  \param[in] frame_width       frame width
 *  \param[in] frame_height      frame height
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_encode_reinit(ni_quadra_enc_context_t *p_enc_ctx,
                                    int frame_width, int frame_height,
                                    const int *linesize);

/*!*****************************************************************************
 *  \brief  Reinit after need to sync context
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_encode_after_reinit(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Recive packet frome encode
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int
ni_quadra_encoder_receive_packet(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Copy niframe to make avframe
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in] p_dst             Pointer to frame data
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_enc_get_extra_data(ni_quadra_enc_context_t *p_enc_ctx,
                                         uint8_t *p_dst);

/*!*****************************************************************************
 *  \brief  Make avpkt from encode
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in] is_pkt            Pointer to frame data
 *  \param[in] pkt_pts           Pointer to frame data
 *  \param[in] pkt_dts           Pointer to frame data
 *  \param[in] pkt_flag          Pointer to frame data
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API int ni_quadra_encoder_prepare_avpkt(ni_quadra_enc_context_t *p_enc_ctx,
                                            int is_pkt, int64_t *pkt_pts,
                                            int64_t *pkt_dts, int *pkt_flag);

/*!*****************************************************************************
 *  \brief  Set width and height to session context
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in] width             frame width
 *  \param[in] height            frame height
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_FAILURE
 * 
 *******************************************************************************/
LIB_API void ni_quadra_set_session_wh(void *session_ctx, int width, int height);

/*!*****************************************************************************
 *  \brief  Allocate memory for the packet buffer based on provided packet size
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        NI_RETCODE_SUCCESS
 *          On failure
 *                        NI_RETCODE_ERROR_MEM_ALOC
 *
 *******************************************************************************/
LIB_API int ni_quadra_packet_buffer_alloc(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Check gop_params
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *
 *  \return On success
 *                        ture
 *          On failure
 *                        false
 * 
 *******************************************************************************/
LIB_API bool ni_quadra_enc_gop_params_check(ni_quadra_enc_context_t *p_enc_ctx);

/*!*****************************************************************************
 *  \brief  Set gop_params
 *
 *  \param[in/out] p_enc_ctx     Pointer to ni_quadra_enc_context_t
 *  \param[in/out] key     Pointer params string
 *
 *  \return none
 * 
 *******************************************************************************/
LIB_API void ni_quadra_enc_gop_params_set(ni_quadra_enc_context_t *p_enc_ctx,
                                        char *key);

/*!*****************************************************************************
 *  \brief  Get struct ni_pkt_info size
 *
 *  \param[in/out] void
 *
 *  \return int
 *
 *******************************************************************************/
LIB_API int ni_quadra_get_pkt_info_size(void);

/*!*****************************************************************************
 *  \brief  Get ni_pkt_info psnr info
 *
 *  \param[in/out] session_ctx     Pointer to ni_session_context_t
 *  \param[in/out] data            Pointer void* which has been malloced
 *
 *  \return none
 *
 *******************************************************************************/
LIB_API void ni_quadra_get_pkt_info(void *session_ctx, void *data);
#ifdef __cplusplus
}
#endif
