/*******************************************************************************
 *
 * Copyright (C) 2022 NETINT Technologies
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 ******************************************************************************/

/*!*****************************************************************************
 *  \file   ni_quadra_filter_api.c
 *
 *  \brief  Private definitions used by ni_quadra_filter_api.c for video processing
 *          tasks
 ******************************************************************************/
#include "ni_quadra_filter_api.h"
#include "ni_defs.h"
#include "ni_rsrc_api.h"
#include "ni_util.h"
#include "ni_device_api.h"
#include "ni_av_codec.h"
#include "ni_bitstream.h"

/*!******************************************************************************
 *  \brief Initialize a fifo buffer
 *
 *  \param[in] number_of_buffers  Total number of buffers in this fifo
 *  \param[in] size               The size of every buffer
 *
 *  \return On success return fifo pointer, On failure return NULL
 *******************************************************************************/
ni_fifo_buffer_t *ni_fifo_initialize(uint32_t number_of_buffers, uint32_t size)
{
    ni_fifo_buffer_t *p_fifo = NULL;

    ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

    if ((p_fifo = (ni_fifo_buffer_t *)malloc(sizeof(ni_fifo_buffer_t))) == NULL)
    {
        ni_log(NI_LOG_ERROR, "Error %d: %s alloc for fifo\n", NI_ERRNO,
               __FUNCTION__);
        return NULL;
    }

    //initialise the struct
    memset(p_fifo, 0, sizeof(ni_fifo_buffer_t));
    ni_pthread_mutex_init(&(p_fifo->mutex));
    p_fifo->number_of_buffers = number_of_buffers;
    p_fifo->buffer_size = size;

    // alloc fifo buffer
    p_fifo->buffer = malloc(number_of_buffers * size);
    if (p_fifo->buffer == NULL)
    {
        free(p_fifo);
        p_fifo = NULL;
        ni_log(NI_LOG_ERROR, "Error %d: %s alloc for fifo buffer\n", NI_ERRNO,
               __FUNCTION__);
        return NULL;
    }
    memset(p_fifo->buffer, 0, (number_of_buffers * size));

    return p_fifo;
}

/*!******************************************************************************
 *  \brief Free a fifo
 *
 *  \param[in] p_fifo  The pointer of ni_fifo_buffer_t
 *
 *  \return
 *******************************************************************************/
void ni_fifo_free(ni_fifo_buffer_t *p_fifo)
{
    ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

    if (!p_fifo)
    {
        ni_log(NI_LOG_TRACE, "%s: null fifo, return\n", __FUNCTION__);
        return;
    }
    ni_pthread_mutex_destroy(&(p_fifo->mutex));

    free(p_fifo->buffer);
    p_fifo->buffer = NULL;

    free(p_fifo);

    return;
}

/*!******************************************************************************
 *  \brief Check if a fifo is full
 *
 *  \param[in] p_fifo  The pointer of ni_fifo_buffer_t
 *
 *  \return 1 is full; 0 is not full
 *******************************************************************************/
int ni_fifo_is_full(ni_fifo_buffer_t *p_fifo)
{
    return (p_fifo->number_of_buffers_used == p_fifo->number_of_buffers);
}

/*!******************************************************************************
 *  \brief Check if a fifo is empty
 *
 *  \param[in] p_fifo  The pointer of ni_fifo_buffer_t
 *
 *  \return 1 is empty; 0 is not empty
 *******************************************************************************/
int ni_fifo_is_empty(ni_fifo_buffer_t *p_fifo)
{
    return (p_fifo->number_of_buffers_used == 0);
}

/*!******************************************************************************
 *  \brief Get number of buffers used in fifo
 *
 *  \param[in] p_fifo  The pointer of ni_fifo_buffer_t
 *
 *  \return fifo's number_of_buffers_used
 *******************************************************************************/
uint32_t ni_fifo_number_of_buffers(ni_fifo_buffer_t *p_fifo)
{
    return p_fifo->number_of_buffers_used;
}

/*!******************************************************************************
 *  \brief Get free buffer to write in the fifo
 *
 *  \param[in] p_fifo  The pointer of ni_fifo_buffer_t
 *
 *  \return On success return the pointer of free buffer,
 *          On failure return NULL
 *******************************************************************************/
uint8_t *ni_fifo_generic_write(ni_fifo_buffer_t *p_fifo)
{
    uint8_t *p_buffer;
    ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

    if (NULL == p_fifo || ni_fifo_is_full(p_fifo))
    {
        ni_log(NI_LOG_TRACE, "Warning pointer is null or fifo is full, %p\n",
               p_fifo);
        return NULL;
    }

    ni_pthread_mutex_lock(&(p_fifo->mutex));

    p_buffer = p_fifo->buffer + p_fifo->wndx * p_fifo->buffer_size;
    p_fifo->wndx = (p_fifo->wndx + 1) % p_fifo->number_of_buffers;
    p_fifo->number_of_buffers_used++;

    ni_pthread_mutex_unlock(&(p_fifo->mutex));
    return p_buffer;
}

/*!******************************************************************************
 *  \brief Get first filled buffer to read in the fifo
 *
 *  \param[in] p_fifo  The pointer of ni_fifo_buffer_t
 *
 *  \return On success return the pointer of filled buffer in fifo,
 *          On failure return NULL
 *******************************************************************************/
uint8_t *ni_fifo_generic_read(ni_fifo_buffer_t *p_fifo)
{
    uint8_t *p_buffer;
    ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

    if (NULL == p_fifo || ni_fifo_is_empty(p_fifo))
    {
        ni_log(NI_LOG_TRACE, "Warning pointer is null or fifo is empty, %p\n",
               p_fifo);
        return NULL;
    }

    ni_pthread_mutex_lock(&(p_fifo->mutex));
    p_buffer = p_fifo->buffer + p_fifo->rndx * p_fifo->buffer_size;
    p_fifo->rndx = (p_fifo->rndx + 1) % p_fifo->number_of_buffers;
    p_fifo->number_of_buffers_used--;

    ni_pthread_mutex_unlock(&(p_fifo->mutex));
    return p_buffer;
}

/*!******************************************************************************
 *  \brief Push back the last read buffer to the fifo
 *
 *  \param[in] p_fifo  The pointer of ni_fifo_buffer_t
 *
 *  \return On success return 0,
 *          On failure return -1
 *******************************************************************************/
int ni_fifo_return_read(ni_fifo_buffer_t *p_fifo)
{
    ni_log(NI_LOG_TRACE, "%s: enter\n", __FUNCTION__);

    if (NULL == p_fifo || ni_fifo_is_full(p_fifo))
    {
        ni_log(NI_LOG_TRACE, "Warning pointer is null or fifo is full, %p\n",
               p_fifo);
        return -1;
    }

    ni_pthread_mutex_lock(&(p_fifo->mutex));

    p_fifo->rndx =
        p_fifo->rndx > 0 ? (p_fifo->rndx - 1) : (p_fifo->number_of_buffers - 1);
    p_fifo->number_of_buffers_used++;

    ni_pthread_mutex_unlock(&(p_fifo->mutex));
    return 0;
}

int ni_quadra_init_filters(ni_quadra_filter_t *nifilter)
{
    ni_scaler_params_t *p_param = NULL;
    ni_session_context_t *p_session_ctx = NULL;
    ni_session_data_io_t *dst_frame = NULL;
    ni_session_data_io_t *dst_pkt = NULL;

    int ret = NI_RETCODE_SUCCESS;
    ni_log(NI_LOG_DEBUG, "XCoder setup device decoder\n");

    if (!nifilter->dst_frame)
    {
        dst_frame = malloc(sizeof(ni_session_data_io_t));
        if (!dst_frame)
        {
            ni_log(NI_LOG_ERROR,
                   "ERROR %d:%s() line %d, alloc dst_frame failed\n", NI_ERRNO,
                   __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            goto end;
        } else
        {
            nifilter->dst_frame = dst_frame;
        }
        memset(dst_frame, 0, sizeof(ni_frame_t));
    }
    if (nifilter->device_type == NI_QUADRA_DEVICE_TYPE_AI && !nifilter->dst_pkt)
    {
        dst_pkt = malloc(sizeof(ni_session_data_io_t));
        if (!dst_pkt)
        {
            ni_log(NI_LOG_ERROR,
                   "ERROR %d:%s() line %d, alloc dst_pkt failed\n", NI_ERRNO,
                   __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            goto end;
        } else
        {
            nifilter->dst_pkt = dst_pkt;
        }
        memset(dst_pkt, 0, sizeof(ni_packet_t));
    }
    // Xcoder Session Context
    if (!nifilter->api_ctx)
    {
        p_session_ctx = malloc(sizeof(ni_session_context_t));
        if (!p_session_ctx)
        {
            ni_log(NI_LOG_ERROR,
                   "ERROR %d:%s() line %d, alloc session context failed\n",
                   NI_ERRNO, __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            goto end;
        } else
        {
            nifilter->api_ctx = p_session_ctx;
            p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
        }
    }
    else
    {
        p_session_ctx = (ni_session_context_t *)nifilter->api_ctx;
    }

    if (ni_device_session_context_init(p_session_ctx) < 0)
    {
        ni_log(NI_LOG_ERROR, "ni filter session context init failure\n");
        ret = NI_RETCODE_FAILURE;
        goto end;
    }

    if (nifilter->hw_action)
    {
        p_session_ctx->device_handle = nifilter->device_handle;
        p_session_ctx->blk_io_handle = nifilter->blk_io_handle;
    }
    p_session_ctx->hw_id = nifilter->hw_id;
    p_session_ctx->device_type = nifilter->device_type;
    p_session_ctx->scaler_operation = nifilter->scaler_operation;
    p_session_ctx->keep_alive_timeout = nifilter->keep_alive_timeout;
    p_session_ctx->isP2P = nifilter->is_p2p;
    p_session_ctx->hw_action = nifilter->hw_action;
    ni_log(NI_LOG_ERROR, "Open filter session, hdl %d, blk_hdl %d\n",
           p_session_ctx->device_handle, p_session_ctx->blk_io_handle);

    ret = ni_device_session_open(p_session_ctx, p_session_ctx->device_type);
    if (ret != NI_RETCODE_SUCCESS)
    {
        ni_log(NI_LOG_ERROR, "Can't open device session\n");
        ni_device_session_close(p_session_ctx, 1, p_session_ctx->device_type);
        ni_device_session_context_clear(p_session_ctx);
        goto end;
    }
    nifilter->session_opened = 1;
    if (p_session_ctx->scaler_operation == NI_QUADRA_SCALER_OPCODE_SCALE ||
        p_session_ctx->scaler_operation == NI_QUADRA_SCALER_OPCODE_DRAWBOX ||
        p_session_ctx->scaler_operation == NI_QUADRA_SCALER_OPCODE_MERGE)
    {
        p_param = malloc(sizeof(ni_scaler_params_t));
        if (!p_param)
        {
            ni_log(NI_LOG_ERROR,
                   "ERROR %d:%s() line %d, alloc session context failed\n",
                   NI_ERRNO, __FUNCTION__, __LINE__);
            ret = NI_RETCODE_ERROR_MEM_ALOC;
            goto end;
        } else
        {
            nifilter->params = p_param;
            // init ni_scaler_params_t data
            p_param->nb_inputs = 0;
            p_param->filterblit = 0;
        }
        p_param->scaler_param_b = nifilter->scaler_param_b;
        p_param->scaler_param_c = nifilter->scaler_param_c;
        if (nifilter->scaler_param_b != 0 || nifilter->scaler_param_c != 0.75)
        {
            p_param->enable_scaler_params = true;
        }
        else
        {
            p_param->enable_scaler_params = false;
        }

        if (nifilter->filterblit)
        {
            p_param->filterblit = nifilter->filterblit;
            ret = ni_scaler_set_params(nifilter->api_ctx, nifilter->params);
            if (ret < 0)
                goto end;
        }
    }

    return ret;
end:
    if (ret != NI_RETCODE_SUCCESS)
    {
        if (p_session_ctx)
            free(p_session_ctx);
        nifilter->api_ctx = NULL;
        if (p_param)
            free(p_param);
        nifilter->params = NULL;
        if (dst_frame)
            free(dst_frame);
        nifilter->dst_frame = NULL;
        if (nifilter->dst_pkt)
            free(nifilter->dst_pkt);
        nifilter->dst_pkt = NULL;
    }
    return ret;
}

int ni_quadra_build_frame_pool(ni_quadra_filter_t *nifilter, int width,
                               int height, int out_format, int pool_size)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    int ret;
    int options;

    options = NI_SCALER_FLAG_IO | NI_SCALER_FLAG_PC;
    if (p_session_ctx->isP2P)
    {
        options |= NI_SCALER_FLAG_P2;
    }
    /* Allocate a pool of frames by the scaler */
    ret = ni_device_alloc_frame(p_session_ctx, NIALIGN(width, 2),
                                NIALIGN(height, 2), out_format, options,
                                0,           // rec width
                                0,           // rec height
                                0,           // rec X pos
                                0,           // rec Y pos
                                pool_size,   // rgba color/pool size
                                0,           // frame index
                                NI_DEVICE_TYPE_SCALER);
    return ret;
}

int ni_quadra_filter_frame_buffer_alloc_hwenc(ni_quadra_filter_t *nifilter,
                                              int video_width, int video_height,
                                              int extra_len)
{
    ni_session_data_io_t *niframe = nifilter->dst_frame;
    return ni_frame_buffer_alloc_hwenc(&niframe->data.frame, video_width,
                                       video_height, extra_len);
}

int ni_quadra_filter_device_alloc_frame(ni_quadra_filter_t *nifilter, int width,
                                        int height, int format, int options,
                                        int rectangle_width,
                                        int rectangle_height, int rectangle_x,
                                        int rectangle_y, int rgba_color,
                                        int frame_index, int type)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    return ni_device_alloc_frame(p_session_ctx, width, height, format, options,
                                 rectangle_width, rectangle_height, rectangle_x,
                                 rectangle_y, rgba_color, frame_index, type);
}

int ni_quadra_filter_device_alloc_dst_frame(ni_quadra_filter_t *nifilter,
                                            niquadraFrameSurface1_t *p_hwdesc,
                                            int type)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    return ni_device_alloc_dst_frame(p_session_ctx,
                                    (niFrameSurface1_t *)p_hwdesc,
                                    type);
}

int ni_quadra_filter_device_clone_hwframe(
                    ni_quadra_filter_t *nifilter,
                    ni_quadra_frameclone_desc_t *p_frameclone_desc)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    return ni_device_clone_hwframe(p_session_ctx,
                (ni_frameclone_desc_t *)p_frameclone_desc);
}

int ni_quadra_device_session_read_hwdesc(void *p_ctx, void *p_data,
                                                int device_type)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    ni_session_data_io_t *niframe = p_data;
    return ni_device_session_read_hwdesc(p_session_ctx, niframe, device_type);
}

int ni_quadra_device_session_read(void *p_ctx, void *p_data,
                                         int device_type)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    ni_session_data_io_t *ni_pkt = p_data;
    return ni_device_session_read(p_session_ctx, ni_pkt, device_type);
}

int ni_quadra_scaler_set_drawbox_params(
    ni_quadra_filter_t *nifilter, ni_quadra_scaler_drawbox_params_t *p_params)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    return ni_scaler_set_drawbox_params(p_session_ctx, (ni_scaler_drawbox_params_t *)p_params);
}

int ni_quadra_scaler_set_watermark_params(
    ni_quadra_filter_t *nifilter, ni_quadra_scaler_watermark_params_t *p_params)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    return ni_scaler_set_watermark_params(p_session_ctx, (ni_scaler_watermark_params_t* )p_params);
}

int ni_quadra_filter_get_output_data(ni_quadra_filter_t *nifilter,
                                     uint8_t *data, int width, int height)
{
    ni_session_data_io_t *niframe = nifilter->dst_frame;
    niFrameSurface1_t *frame_surface = (niFrameSurface1_t *)data;
    niFrameSurface1_t *new_frame_surface = (niFrameSurface1_t *)niframe->data.frame.p_data[3];
    frame_surface->ui16FrameIdx = new_frame_surface->ui16FrameIdx;
    frame_surface->ui16session_ID = new_frame_surface->ui16session_ID;
    frame_surface->device_handle = new_frame_surface->device_handle;
    frame_surface->output_idx = new_frame_surface->output_idx;
    frame_surface->src_cpu = new_frame_surface->src_cpu;
    frame_surface->dma_buf_fd = 0;
    frame_surface->ui32nodeAddress = 0;
    frame_surface->ui16width = width;
    frame_surface->ui16height = height;
    return 0;
}

uint8_t *ni_quadra_filter_get_data3(void *p_frame)
{
    ni_session_data_io_t *niframe = p_frame;
    return niframe->data.frame.p_data[3];
}

int ni_quadra_filter_drawtext_prep_frames(void **niframe)
{
    int ret = 0;
    if (!*niframe)
    {
        ni_session_data_io_t *p_data = NULL;
        *niframe = malloc(sizeof(ni_session_data_io_t));
        if (!*niframe)
        {
            ni_log(NI_LOG_ERROR,
                   "ERROR %d:%s() line %d, alloc session data failed\n",
                   NI_ERRNO, __FUNCTION__, __LINE__);
            return NI_RETCODE_ERROR_MEM_ALOC;
        }
        memset(*niframe, 0, sizeof(ni_session_data_io_t));
    }
    return ret;
}

int ni_quadra_filter_frame_buffer_alloc_dl(void *niframe, int video_width,
                                           int video_height, int pixel_format)
{
    ni_session_data_io_t *p_frame = niframe;

    return ni_frame_buffer_alloc_dl(&(p_frame->data.frame), video_width,
                                    video_height, pixel_format);
}

void ni_quadra_filter_close(ni_quadra_filter_t *nifilter)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    ni_session_data_io_t *niframe = nifilter->dst_frame;
//    ni_scaler_params_t *p_param = nifilter->params;
    if (niframe && niframe->data.frame.p_buffer)
        ni_frame_buffer_free(&niframe->data.frame);

    if (nifilter->dl_frame)
    {
        ni_session_data_io_t *p_frame = nifilter->dl_frame;
        ni_frame_buffer_free(&p_frame->data.frame);
        if (nifilter->dl_frame)
            free(nifilter->dl_frame);
    }
    if (nifilter->txt_frame)
    {
        ni_session_data_io_t *p_frame = nifilter->txt_frame;
        ni_frame_buffer_free(&p_frame->data.frame);
        if (nifilter->txt_frame)
            free(nifilter->txt_frame);
    }
    if (nifilter->dst_pkt)
    {
        ni_session_data_io_t *p_packet = nifilter->dst_pkt;
        ni_packet_buffer_free(&p_packet->data.packet);
        if (nifilter->dst_pkt)
            free(nifilter->dst_pkt);
    }
    if (nifilter->session_opened)
    {
        /* Close operation will free the device frames */
        ni_device_session_close(p_session_ctx, 1, nifilter->device_type);
        ni_device_session_context_clear(p_session_ctx);
    }
    if (nifilter->api_ctx)
        free(nifilter->api_ctx);
    if (nifilter->dst_frame)
        free(nifilter->dst_frame);
    if (nifilter->params)
        free(nifilter->params);
    nifilter->api_ctx = NULL;
    nifilter->params = NULL;
    nifilter->dst_frame = NULL;
}

int ni_quadra_ai_config_network_binary(ni_quadra_filter_t *nifilter,
                                       ni_quadra_network_data_t *p_network,
                                       const char *file)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;
    return ni_ai_config_network_binary(p_session_ctx, (ni_network_data_t *)p_network, file);
}

int ni_quadra_ai_config_hvsplus(ni_quadra_filter_t *nifilter,
                                ni_quadra_network_data_t *p_network,
                                int nb_width,
                                int nb_height,
                                int level,
                                int format)
{
    ni_session_context_t *p_session_ctx = nifilter->api_ctx;

    p_session_ctx->active_video_width = nb_width;
    p_session_ctx->active_video_height = nb_height;
    p_session_ctx->hvsplus_level = level;
    p_session_ctx->pixel_format = format;

    return ni_ai_config_hvsplus(p_session_ctx, (ni_network_data_t *)p_network);
}

int ni_quadra_ai_frame_buffer_alloc(void *p_frame,
                                    ni_quadra_network_data_t *p_network)
{
    ni_session_data_io_t *p_dst_frame = p_frame;
    return ni_ai_frame_buffer_alloc(&p_dst_frame->data.frame, (ni_network_data_t *)p_network);
}

int ni_quadra_ai_packet_buffer_alloc(void *p_pkt,
                                     ni_quadra_network_data_t *p_network)
{
    ni_session_data_io_t *p_packet = p_pkt;
    return ni_ai_packet_buffer_alloc(&p_packet->data.packet, (ni_network_data_t *)p_network);
}

uint32_t
ni_quadra_ai_network_layer_dims(ni_quadra_network_layer_params_t *p_param)
{
    return ni_ai_network_layer_dims((ni_network_layer_params_t *)p_param);
}

uint32_t ni_quadra_ai_network_layer_size(ni_quadra_network_layer_params_t *p_param)
{
    return ni_ai_network_layer_size((ni_network_layer_params_t *)p_param);
}

int ni_quadra_network_layer_convert_output(void *p_data,
                                           float *dst, uint32_t num,
                                           ni_quadra_network_data_t *p_network,
                                           uint32_t layer)
{
    ni_session_data_io_t *p_packet = p_data;
    return (int)ni_network_layer_convert_output(dst, num, &p_packet->data.packet,
                                    (ni_network_data_t *)p_network, layer);
}

int ni_quadra_frame_buffer_alloc_pixfmt(void *p_frame, int pixel_format,
                                        int video_width, int video_height,
                                        int linesize[], int alignment,
                                        int extra_len)
{
    ni_session_data_io_t *p_data = p_frame;
    return ni_frame_buffer_alloc_pixfmt(&p_data->data.frame, pixel_format,
                                        video_width, video_height, linesize,
                                        alignment, extra_len);
}

int ni_quadra_device_session_hwdl(void *p_ctx, void *p_data,
                                  niquadraFrameSurface1_t *hwdesc)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    ni_session_data_io_t *p_session_data = p_data;
    return ni_device_session_hwdl(p_session_ctx, p_session_data, (niFrameSurface1_t *)hwdesc);
}

int ni_quadra_init_internal_session(NIFramesContext *ni_ctx)
{
    ni_session_context_t *p_session_ctx = NULL;
    ni_session_data_io_t *dst_frame = NULL;
    int ret = NI_RETCODE_SUCCESS;
    p_session_ctx = malloc(sizeof(ni_session_context_t));
    if (!p_session_ctx)
    {
        ni_log(NI_LOG_ERROR,
               "ERROR %d:%s() line %d, alloc session context failed\n",
               NI_ERRNO, __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    } else
    {
        ni_ctx->api_ctx = p_session_ctx;
        p_session_ctx->session_run_state = SESSION_RUN_STATE_NORMAL;
    }
    if (ni_device_session_context_init(p_session_ctx) < 0)
    {
        ni_log(NI_LOG_ERROR, "ni init context failure\n");
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    }

    dst_frame = malloc(sizeof(ni_session_data_io_t));
    if (!dst_frame)
    {
        ni_log(NI_LOG_ERROR, "ERROR %d:%s() line %d, alloc dst_frame failed\n",
               NI_ERRNO, __FUNCTION__, __LINE__);
        ret = NI_RETCODE_ERROR_MEM_ALOC;
        goto end;
    } else
    {
        ni_ctx->src_session_io_data = dst_frame;
    }
    memset(dst_frame, 0, sizeof(ni_session_data_io_t));
    return ret;
end:
    if (ret != NI_RETCODE_SUCCESS)
    {
        free(p_session_ctx);
        ni_ctx->api_ctx = NULL;
        // free (p_param);
        // nifilter->params = NULL;
        free(dst_frame);
        ni_ctx->src_session_io_data = NULL;
    }
    return ret;
}

int ni_quadra_init_split_rsrc(NIFramesContext *ni_ctx, int w, int h)
{
    int i;
    int ret = NI_RETCODE_SUCCESS;
    ni_split_context_t *p_split_ctx = (ni_split_context_t *)&ni_ctx->split_ctx;
    memset(p_split_ctx, 0, sizeof(ni_split_context_t));
    for (i = 0; i < 3; i++)
    {
        p_split_ctx->w[i] = w;
        p_split_ctx->h[i] = h;
        p_split_ctx->f[i] = -1;
    }
    return ret;
}

void ni_quadra_filter_device_close(AVNIDeviceContext *ni_hw_ctx)
{
    int i;
    if (ni_hw_ctx->uploader_handle != NI_INVALID_DEVICE_HANDLE)
    {
        ni_device_close(ni_hw_ctx->uploader_handle);
        ni_hw_ctx->uploader_handle = NI_INVALID_DEVICE_HANDLE;
    }
    for (i = 0; i < NI_MAX_DEVICE_CNT; i++)
    {
        ni_device_handle_t fd = ni_hw_ctx->cards[i];
        if (fd != NI_INVALID_DEVICE_HANDLE)
        {
            ni_hw_ctx->cards[i] = NI_INVALID_DEVICE_HANDLE;
            ni_device_close(fd);
        } else
        {
            break;
        }
    }
}

int ni_quadra_upload_device_session_open(NIFramesContext *ni_ctx)
{
    return ni_device_session_open(ni_ctx->api_ctx, NI_DEVICE_TYPE_UPLOAD);
}

int ni_quadra_filter_device_session_init_framepool(NIFramesContext *ni_ctx,
                                                   uint32_t pool_size,
                                                   uint32_t pool)
{
    ni_session_context_t *p_session_ctx = ni_ctx->api_ctx;
    return ni_device_session_init_framepool(p_session_ctx, pool_size, pool);
}

int ni_quadra_filter_frames_close(NIFramesContext *ni_ctx, int pool_size)
{
    ni_session_context_t *p_session_ctx = ni_ctx->api_ctx;
    ni_session_data_io_t *p_data = ni_ctx->src_session_io_data;
    if (ni_ctx->uploader_device_id != -2 && pool_size >= 0)
    {
        if(p_data)
        {
            if (p_data->data.frame.buffer_size ||
            p_data->data.frame.metadata_buffer_size ||
            p_data->data.frame.start_buffer_size)
            {
                ni_log(NI_LOG_DEBUG, "%s:free upload src frame buffer\n", __func__);
                ni_frame_buffer_free(&p_data->data.frame);
            }
        }
        
        if(p_session_ctx)
        {
            ni_log(NI_LOG_DEBUG, "SessionID = %d!\n", p_session_ctx->session_id);
            if (p_session_ctx->session_id != NI_INVALID_SESSION_ID)
            {
                ni_device_session_close(p_session_ctx, 1, NI_DEVICE_TYPE_UPLOAD);
            }

            ni_device_session_context_clear(p_session_ctx);
        }
    }
    if (ni_ctx->suspended_device_handle != -1)
    {
        ni_log(NI_LOG_DEBUG, "%s: close file handle =%d\n", __func__,
               ni_ctx->suspended_device_handle);
        ni_device_close(ni_ctx->suspended_device_handle);
        ni_ctx->suspended_device_handle = -1;
    }
    if (ni_ctx->api_ctx)
    {
        free(ni_ctx->api_ctx);
        ni_ctx->api_ctx = NULL;
    }
    if (ni_ctx->src_session_io_data)
    {
        free(ni_ctx->src_session_io_data);
        ni_ctx->src_session_io_data = NULL;
    }
    if (ni_ctx->dst_frame)
    {
        free(ni_ctx->dst_frame);
        ni_ctx->dst_frame = NULL;
    }

    return 0;   
}

int ni_quadra_filter_frames_init(NIFramesContext *ni_ctx, int *width,
                                 int *height, int ni_fmat, int keep_alive_time)
{
    ni_session_context_t *p_session_ctx = ni_ctx->api_ctx;
    int linesize_aligned, height_aligned;
    linesize_aligned = *width;
    if (QUADRA)
    {
        linesize_aligned = NI_VPU_CEIL(linesize_aligned, 2);
    } else
    {
        linesize_aligned = ((*width + 31) / 32) * 32;
        if (linesize_aligned < NI_MIN_WIDTH)
        {
            linesize_aligned = NI_MIN_WIDTH;
        } else if (linesize_aligned > *width)
        {
        }
    }
    *width = linesize_aligned;

    height_aligned = *height;
    if (QUADRA)
    {
        *height = NI_VPU_CEIL(height_aligned, 2);
    } else
    {
        height_aligned = ((*height + 15) / 16) * 16;
        if (height_aligned < NI_MIN_HEIGHT)
        {
            *height = NI_MIN_HEIGHT;
            height_aligned = NI_MIN_HEIGHT;
        } else if (height_aligned > *height)
        {
            *height = height_aligned;
        }
    }
    p_session_ctx->active_video_width = *width;
    p_session_ctx->active_video_height = *height;
    switch (ni_fmat)
    {
        case NI_PIX_FMT_YUV420P:
            p_session_ctx->bit_depth_factor = 1;
            p_session_ctx->src_bit_depth = 8;
            p_session_ctx->pixel_format = NI_PIX_FMT_YUV420P;
            break;
        case NI_PIX_FMT_YUV420P10LE:
            p_session_ctx->bit_depth_factor = 2;
            p_session_ctx->src_bit_depth = 10;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_YUV420P10LE;
            break;
        case NI_PIX_FMT_NV12:
            p_session_ctx->bit_depth_factor = 1;
            p_session_ctx->src_bit_depth = 8;
            p_session_ctx->pixel_format = NI_PIX_FMT_NV12;
            break;
        case NI_PIX_FMT_P010LE:
            p_session_ctx->bit_depth_factor = 2;
            p_session_ctx->src_bit_depth = 10;
            p_session_ctx->pixel_format = NI_PIX_FMT_P010LE;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            break;
        case NI_PIX_FMT_RGBA:
            p_session_ctx->bit_depth_factor = 4;
            p_session_ctx->src_bit_depth = 32;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_RGBA;
            break;
        case NI_PIX_FMT_BGRA:
            p_session_ctx->bit_depth_factor = 4;
            p_session_ctx->src_bit_depth = 32;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_BGRA;
            break;
        case NI_PIX_FMT_ABGR:
            p_session_ctx->bit_depth_factor = 4;
            p_session_ctx->src_bit_depth = 32;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_ABGR;
            break;
        case NI_PIX_FMT_ARGB:
            p_session_ctx->bit_depth_factor = 4;
            p_session_ctx->src_bit_depth = 32;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_ARGB;
            break;
        case NI_PIX_FMT_BGR0:
            p_session_ctx->bit_depth_factor = 4;
            p_session_ctx->src_bit_depth = 32;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_BGR0;
            break;
        case NI_PIX_FMT_BGRP:
            p_session_ctx->bit_depth_factor = 1;
            p_session_ctx->src_bit_depth = 24;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_BGRP;
            break;
        case NI_PIX_FMT_YUYV422:
            p_session_ctx->bit_depth_factor = 1;
            p_session_ctx->src_bit_depth = 8;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_YUYV422;
            break;
        case NI_PIX_FMT_UYVY422:
            p_session_ctx->bit_depth_factor = 1;
            p_session_ctx->src_bit_depth = 8;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_UYVY422;
            break;
        case NI_PIX_FMT_NV16:
            p_session_ctx->bit_depth_factor = 1;
            p_session_ctx->src_bit_depth = 8;
            p_session_ctx->src_endian = NI_FRAME_LITTLE_ENDIAN;
            p_session_ctx->pixel_format = NI_PIX_FMT_NV16;
            break;
        default:
            ni_log(NI_LOG_ERROR, "Pixel format not supported by device.\n");
            return NI_RETCODE_FAILURE;
    }
    if (*width > NI_MAX_RESOLUTION_WIDTH ||
        *height > NI_MAX_RESOLUTION_HEIGHT ||
        *width * (*height) > NI_MAX_RESOLUTION_AREA)
    {
        ni_log(NI_LOG_ERROR, "Error XCoder resolution %dx%d not supported\n",
               *width, *height);
        ni_log(NI_LOG_ERROR, "Max Supported Width: %d Height %d Area %d\n",
               NI_MAX_RESOLUTION_WIDTH, NI_MAX_RESOLUTION_HEIGHT,
               NI_MAX_RESOLUTION_AREA);
        return NI_RETCODE_FAILURE;
    } else if (ni_ctx->uploader_device_id >= -1)
    {
        // leave it to ni_device_session_open to handle uploader session open
        // based on api_ctx.hw_id set to proper value
    } else
    {
        ni_log(NI_LOG_ERROR, "Error XCoder command line options");
        return NI_RETCODE_FAILURE;
    }
    p_session_ctx->hw_id = ni_ctx->uploader_device_id;
    p_session_ctx->keep_alive_timeout = keep_alive_time;
    if (0 == p_session_ctx->keep_alive_timeout)
    {
        p_session_ctx->keep_alive_timeout = NI_DEFAULT_KEEP_ALIVE_TIMEOUT;
    }
    return 0;
}

int ni_quadra_filter_device_create(AVNIDeviceContext *ni_hw_ctx,
                                   const char *device)
{
    char *blk_name;
    int i, module_id = 0, ret = 0;
    ni_device_handle_t fd;
    uint32_t max_io_size = 0;
    ni_device_t *p_ni_devices = NULL;

    p_ni_devices = calloc(1, sizeof(ni_device_t));
    if(!p_ni_devices)
    {
        ni_log(NI_LOG_ERROR, "could not allocate memory for p_ni_devices in %s", __func__);
        return NI_RETCODE_FAILURE;
    }

    ni_hw_ctx->uploader_ID =
        -2;   // -1 is auto load balance, default -2 invalid
    ni_hw_ctx->uploader_handle = NI_INVALID_DEVICE_HANDLE;

    if (device)
    {
        /* parse device string and fail if incorrect */
        ni_log(NI_LOG_DEBUG, "%s %s\n", __func__, device);
        ni_hw_ctx->uploader_ID = atoi(device);
        ni_log(NI_LOG_DEBUG, "%s: given uploader ID %d\n", __func__,
               ni_hw_ctx->uploader_ID);
        if (ni_hw_ctx->uploader_ID < -1)
        {
            ni_log(NI_LOG_ERROR, "%s: uploader ID %d must be >= -1.\n",
                   __func__, ni_hw_ctx->uploader_ID);
            ret =  NI_RETCODE_FAILURE;
            LRETURN;
        }
    }

    for (i = 0; i < NI_MAX_DEVICE_CNT; i++)
    {
        ni_hw_ctx->cards[i] = NI_INVALID_DEVICE_HANDLE;
    }

    /* Scan all cards on the host, only look at NETINT cards */
    if (ni_rsrc_list_all_devices(p_ni_devices) == NI_RETCODE_SUCCESS)
    {
        // Note: this only checks for Netint encoders
        for (i = 0; i < p_ni_devices->xcoder_cnt[NI_DEVICE_TYPE_ENCODER]; i++)
        {
            blk_name =
                &(p_ni_devices->xcoders[NI_DEVICE_TYPE_ENCODER][i].blk_name[0]);
            // one-to-one correspondence between card index and module_id
            module_id = p_ni_devices->xcoders[NI_DEVICE_TYPE_ENCODER][i].module_id;
            ni_log(NI_LOG_DEBUG, "%s blk name %s\n", __func__, blk_name);
            fd = ni_device_open(blk_name, &max_io_size);
            if (fd != NI_INVALID_DEVICE_HANDLE)
            {
                ni_hw_ctx->cards[module_id] = fd;
            }
        }
    } else
    {
        ret = NI_RETCODE_FAILURE;
    }
END:
    free(p_ni_devices);
    return ret;
}

int ni_quadra_filter_get_buffer(NIFramesContext *ni_ctx, uint8_t **dstbuf,
                                uint8_t **dstdata, uint32_t *data_len,
                                int width, int height, int is_planar)
{
    ni_session_context_t *p_session_ctx = ni_ctx->api_ctx;
    ni_session_data_io_t dst_session_io_data;
    ni_session_data_io_t *p_dst_session_data = &dst_session_io_data;
    int ret = NI_RETCODE_SUCCESS;
    // alloc dest avframe buff
    // alloc dest avframe buff
    memset(p_dst_session_data, 0, sizeof(dst_session_io_data));
    ret = ni_frame_buffer_alloc(&p_dst_session_data->data.frame, width, height,
                                0, 1,   // codec type does not matter, metadata exists
                                p_session_ctx->bit_depth_factor, 1, !is_planar);
    if (ret != 0)
    {
        return NI_RETCODE_ERROR_MEM_ALOC;
    }

    ni_quadra_get_frame_data_len(p_dst_session_data, data_len);
    *dstbuf = p_dst_session_data->data.frame.p_buffer;
    *dstdata = p_dst_session_data->data.frame.p_data[0];

    ni_log(NI_LOG_DEBUG, "ni_quadra_filter_get_buffer exit\n");
    return ret;
}

void ni_quadra_filter_frame_free(void *opaque, uint8_t *data)
{
    if (data)
    {
        niFrameSurface1_t *ni_surf = (niFrameSurface1_t *)data;
        if (ni_surf->ui16FrameIdx != 0)
        {
            ni_hwframe_buffer_recycle(ni_surf, ni_surf->device_handle);
        }
        ni_aligned_free(ni_surf);
    }
}

static int av_to_niframe_copy(const int dst_stride[4], ni_frame_t *dst,
                              int width, int height, int ni_fmat, int nb_planes,
                              const int *linesize, uint8_t **src_data)
{
    int src_height[4], hpad[4], vpad[4];
    int i, j, h;
    uint8_t *src_line, *dst_line, YUVsample, *sample, *dest;
    uint16_t lastidx;
    bool tenBit;

    switch (ni_fmat)
    {
        case NI_PIX_FMT_YUV420P:
            hpad[0] = NIMAX(dst_stride[0] - linesize[0], 0);
            hpad[1] = NIMAX(dst_stride[1] - linesize[1], 0);
            hpad[2] = NIMAX(dst_stride[2] - linesize[2], 0);
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = NIALIGN(height, 2) / 2;
            src_height[3] = 0;

            vpad[0] = NIALIGN(src_height[0], 2) - src_height[0];
            vpad[1] = NIALIGN(src_height[1], 2) - src_height[1];
            vpad[2] = NIALIGN(src_height[2], 2) - src_height[2];
            vpad[3] = 0;

            tenBit = false;
            break;

        case NI_PIX_FMT_YUV420P10LE:
            hpad[0] = NIMAX(dst_stride[0] - linesize[0], 0);
            hpad[1] = NIMAX(dst_stride[1] - linesize[1], 0);
            hpad[2] = NIMAX(dst_stride[2] - linesize[2], 0);
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = NIALIGN(height, 2) / 2;
            src_height[3] = 0;

            vpad[0] = NIALIGN(src_height[0], 2) - src_height[0];
            vpad[1] = NIALIGN(src_height[1], 2) - src_height[1];
            vpad[2] = NIALIGN(src_height[2], 2) - src_height[2];
            vpad[3] = 0;

            tenBit = true;
            break;

        case NI_PIX_FMT_NV12:
            hpad[0] = NIMAX(dst_stride[0] - linesize[0], 0);
            hpad[1] = NIMAX(dst_stride[1] - linesize[1], 0);
            hpad[2] = 0;
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = 0;
            src_height[3] = 0;

            vpad[0] = NIALIGN(src_height[0], 2) - src_height[0];
            vpad[1] = NIALIGN(src_height[1], 2) - src_height[1];
            vpad[2] = 0;
            vpad[3] = 0;

            tenBit = false;
            break;
        case NI_PIX_FMT_NV16:
            hpad[0] = 0;
            hpad[1] = 0;
            hpad[2] = 0;
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = height;
            src_height[2] = 0;
            src_height[3] = 0;

            vpad[0] = 0;
            vpad[1] = 0;
            vpad[2] = 0;
            vpad[3] = 0;

            tenBit = false;
            break;

        case NI_PIX_FMT_P010LE:
            hpad[0] = NIMAX(dst_stride[0] - linesize[0], 0);
            hpad[1] = NIMAX(dst_stride[1] - linesize[1], 0);
            hpad[2] = 0;
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = 0;
            src_height[3] = 0;

            vpad[0] = NIALIGN(src_height[0], 2) - src_height[0];
            vpad[1] = NIALIGN(src_height[1], 2) - src_height[1];
            vpad[2] = 0;
            vpad[3] = 0;

            tenBit = true;
            break;

        case NI_PIX_FMT_RGBA:
        case NI_PIX_FMT_BGRA:
        case NI_PIX_FMT_ABGR:
        case NI_PIX_FMT_ARGB:
        case NI_PIX_FMT_BGR0:
            hpad[0] = NIMAX(dst_stride[0] - linesize[0], 0);
            hpad[1] = 0;
            hpad[2] = 0;
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = 0;
            src_height[2] = 0;
            src_height[3] = 0;

            vpad[0] = 0;
            vpad[1] = 0;
            vpad[2] = 0;
            vpad[3] = 0;

            tenBit = false;
            break;

        case NI_PIX_FMT_BGRP:
            hpad[0] = NIMAX(dst_stride[0] - linesize[0], 0);
            hpad[1] = NIMAX(dst_stride[1] - linesize[1], 0);
            hpad[2] = NIMAX(dst_stride[2] - linesize[2], 0);
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = height;
            src_height[2] = height;
            src_height[3] = 0;

            vpad[0] = 0;
            vpad[1] = 0;
            vpad[2] = 0;
            vpad[3] = 0;

            tenBit = false;
            break;

        case NI_PIX_FMT_YUYV422:
        case NI_PIX_FMT_UYVY422:
            hpad[0] = NIMAX(dst_stride[0] - linesize[0], 0);
            hpad[1] = 0;
            hpad[2] = 0;
            hpad[3] = 0;

            src_height[0] = height;
            src_height[1] = 0;
            src_height[2] = 0;
            src_height[3] = 0;

            vpad[0] = 0;
            vpad[1] = 0;
            vpad[2] = 0;
            vpad[3] = 0;

            tenBit = false;
            break;

        default:
            ni_log(NI_LOG_ERROR, "Pixel format %s not supported\n", ni_fmat);
            break;
    }

    for (i = 0; i < nb_planes; i++)
    {
        dst_line = dst->p_data[i];
        src_line = src_data[i];

        for (h = 0; h < src_height[i]; h++)
        {
            memcpy(dst_line, src_line, NIMIN(linesize[i], dst_stride[i]));

            if (hpad[i])
            {
                lastidx = linesize[i];

                if (tenBit)
                {
                    sample = &src_line[lastidx - 2];
                    dest = &dst_line[lastidx];

                    /* two bytes per sample */
                    for (j = 0; j < hpad[i] / 2; j++)
                    {
                        memcpy(dest, sample, 2);
                        dest += 2;
                    }
                } else
                {
                    YUVsample = dst_line[lastidx - 1];
                    memset(&dst_line[lastidx], YUVsample, hpad[i]);
                }
            }

            src_line += linesize[i];
            dst_line += dst_stride[i];
        }

        /* Extend the height by cloning the last line */
        src_line = dst_line - dst_stride[i];
        for (h = 0; h < vpad[i]; h++)
        {
            memcpy(dst_line, src_line, dst_stride[i]);
            dst_line += dst_stride[i];
        }
    }

    return 0;
}

static int ni_to_avframe_copy(const ni_frame_t *src, int width, int height,
                              int ni_fmat, int nb_planes, const int *linesize,
                              uint8_t **dst_data)
{
    int src_linesize[4], src_height[4];
    int i, h;
    uint8_t *src_line, *dst_line;

    switch (ni_fmat)
    {
        case NI_PIX_FMT_YUV420P:
            src_linesize[0] = NIALIGN(width, 128);
            src_linesize[1] = NIALIGN(width / 2, 128);
            src_linesize[2] = src_linesize[1];
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = src_height[1];
            src_height[3] = 0;
            break;

        case NI_PIX_FMT_YUV420P10LE:
            src_linesize[0] = NIALIGN(width * 2, 128);
            src_linesize[1] = NIALIGN(width, 128);
            src_linesize[2] = src_linesize[1];
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = src_height[1];
            src_height[3] = 0;
            break;

        case NI_PIX_FMT_NV12:
            src_linesize[0] = NIALIGN(width, 128);
            src_linesize[1] = NIALIGN(width, 128);
            src_linesize[2] = 0;
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = 0;
            src_height[3] = 0;
            break;

        case NI_PIX_FMT_NV16:
            src_linesize[0] = NIALIGN(width, 64);
            src_linesize[1] = NIALIGN(width, 64);
            src_linesize[2] = 0;
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = height;
            src_height[2] = 0;
            src_height[3] = 0;
            break;

        case NI_PIX_FMT_YUYV422:
        case NI_PIX_FMT_UYVY422:
            src_linesize[0] = NIALIGN(width, 16) * 2;
            src_linesize[1] = 0;
            src_linesize[2] = 0;
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = 0;
            src_height[2] = 0;
            src_height[3] = 0;
            break;

        case NI_PIX_FMT_P010LE:
            src_linesize[0] = NIALIGN(width * 2, 128);
            src_linesize[1] = NIALIGN(width * 2, 128);
            src_linesize[2] = 0;
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = NIALIGN(height, 2) / 2;
            src_height[2] = 0;
            src_height[3] = 0;
            break;

        case NI_PIX_FMT_RGBA:
        case NI_PIX_FMT_BGRA:
        case NI_PIX_FMT_ABGR:
        case NI_PIX_FMT_ARGB:
        case NI_PIX_FMT_BGR0:
            src_linesize[0] = NIALIGN(width, 16) * 4;
            src_linesize[1] = 0;
            src_linesize[2] = 0;
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = 0;
            src_height[2] = 0;
            src_height[3] = 0;
            break;

        case NI_PIX_FMT_BGRP:
            src_linesize[0] = NIALIGN(width, 32);
            src_linesize[1] = NIALIGN(width, 32);
            src_linesize[2] = NIALIGN(width, 32);
            src_linesize[3] = 0;

            src_height[0] = height;
            src_height[1] = height;
            src_height[2] = height;
            src_height[3] = 0;
            break;

        default:
            ni_log(NI_LOG_ERROR, "Unsupported pixel format %d\n", ni_fmat);
            return NI_RETCODE_FAILURE;
    }

    for (i = 0; i < nb_planes; i++)
    {
        dst_line = dst_data[i];
        src_line = src->p_data[i];
        for (h = 0; h < src_height[i]; h++)
        {
            memcpy(dst_line, src_line, NIMIN(src_linesize[i], linesize[i]));
            dst_line += linesize[i];
            src_line += src_linesize[i];
        }
    }

    return 0;
}

int ni_quadra_filter_hwup_frame(NIFramesContext *ni_ctx,
                                niquadraFrameSurface1_t *dst_surf, int width,
                                int height, int ni_fmat, int nb_planes,
                                const int *linesize, uint8_t **src_data)
{
    ni_session_context_t *p_session_ctx = ni_ctx->api_ctx;
    ni_session_data_io_t *p_src_session_data = ni_ctx->src_session_io_data;
    ni_split_context_t *p_split_ctx = (ni_split_context_t *)&ni_ctx->split_ctx;
    int need_to_copy = 1;
    bool isSemiPlanar;
    int dst_stride[4];
    int ret;

    switch (ni_fmat)
    {
        /* 8-bit YUV420 planar */
        case NI_PIX_FMT_YUV420P:
            dst_stride[0] = NIALIGN(width, 128);
            dst_stride[1] = NIALIGN((width / 2), 128);
            dst_stride[2] = dst_stride[1];
            dst_stride[3] = 0;

            isSemiPlanar = false;
            break;

        /* 10-bit YUV420 planar, little-endian, least significant bits */
        case NI_PIX_FMT_YUV420P10LE:
            dst_stride[0] = NIALIGN(width * 2, 128);
            dst_stride[1] = NIALIGN(width, 128);
            dst_stride[2] = dst_stride[1];
            dst_stride[3] = 0;

            isSemiPlanar = false;
            break;

        /* 8-bit YUV420 semi-planar */
        case NI_PIX_FMT_NV12:
            dst_stride[0] = NIALIGN(width, 128);
            dst_stride[1] = dst_stride[0];
            dst_stride[2] = 0;
            dst_stride[3] = 0;

            isSemiPlanar = true;
            break;

        /* 8-bit yuv422 semi-planar */
        case NI_PIX_FMT_NV16:
            dst_stride[0] = NIALIGN(width, 64);
            dst_stride[1] = dst_stride[0];
            dst_stride[2] = 0;
            dst_stride[3] = 0;

            isSemiPlanar = false;
            break;

        /*8-bit yuv422 planar */
        case NI_PIX_FMT_YUYV422:
        case NI_PIX_FMT_UYVY422:
            dst_stride[0] = NIALIGN(width, 16) * 2;
            dst_stride[1] = 0;
            dst_stride[2] = 0;
            dst_stride[3] = 0;

            isSemiPlanar = false;
            break;

        /* 10-bit YUV420 semi-planar, little endian, most significant bits */
        case NI_PIX_FMT_P010LE:
            dst_stride[0] = NIALIGN(width * 2, 128);
            dst_stride[1] = dst_stride[0];
            dst_stride[2] = 0;
            dst_stride[3] = 0;

            isSemiPlanar = true;
            break;

        /* 32-bit RGBA packed */
        /* RGBA for the scaler has a 16-byte width/64-byte stride alignment */
        case NI_PIX_FMT_RGBA:
        case NI_PIX_FMT_BGRA:
        case NI_PIX_FMT_ABGR:
        case NI_PIX_FMT_ARGB:
        case NI_PIX_FMT_BGR0:
            dst_stride[0] = NIALIGN(width, 16) * 4;
            dst_stride[1] = 0;
            dst_stride[2] = 0;
            dst_stride[3] = 0;

            isSemiPlanar = false;
            break;

        /* 24-bit BGR planar not supported for hwupload */
        /*
        case AV_PIX_FMT_BGRP:
            dst_stride[0] = NIALIGN(width, 32);
            dst_stride[1] = NIALIGN(width, 32);
            dst_stride[2] = NIALIGN(width, 32);
            dst_stride[3] = 0;

            pixel_format = NI_PIX_FMT_BGRP;
            isSemiPlanar = false;
            break;
        */
        default:
            ni_log(NI_LOG_ERROR, "Pixel format %d not supported\n", ni_fmat);
            return NI_RETCODE_FAILURE;
    }

    // check input resolution zero copy compatible or not
    if (ni_uploader_frame_zerocopy_check(p_session_ctx, width, height,
                                         (const int *)linesize,
                                         ni_fmat) == NI_RETCODE_SUCCESS)
    {
        need_to_copy = 0;
        p_src_session_data->data.frame.extra_data_len =
            NI_APP_ENC_FRAME_META_DATA_SIZE;
        // alloc metadata buffer etc. (if needed)
        ret = ni_encoder_frame_zerocopy_buffer_alloc(
            &p_src_session_data->data.frame, width, height,
            (const int *)linesize, (const uint8_t **)src_data,
            (int)p_src_session_data->data.frame.extra_data_len);
        if (ret != NI_RETCODE_SUCCESS)
            return NI_RETCODE_ERROR_MEM_ALOC;
    } else
    {
        // allocate only once per upload Session when we have frame info
        p_src_session_data->data.frame.extra_data_len =
            NI_APP_ENC_FRAME_META_DATA_SIZE;

        ret = ni_frame_buffer_alloc_pixfmt(
            &p_src_session_data->data.frame, ni_fmat, width, height, dst_stride,
            1,   // force to av_codec_id_h264 for max compat
            (int)p_src_session_data->data.frame.extra_data_len);
        if (ret < 0)
        {
            ni_log(NI_LOG_ERROR, "Cannot allocate ni_frame %d\n", ret);
            return ret;
        }
    }
    if (need_to_copy)
    {
        ret = av_to_niframe_copy(dst_stride, &p_src_session_data->data.frame,
                                 width, height, ni_fmat, nb_planes, linesize,
                                 src_data);
        if (ret < 0)
        {
            ni_log(NI_LOG_ERROR, "%s can't copy frame\n", __func__);
            return NI_RETCODE_FAILURE;
        }
    }
    ret = ni_device_session_hwup(p_session_ctx, p_src_session_data, (niFrameSurface1_t *)dst_surf);
    if (ret < 0)
    {
        ni_log(NI_LOG_ERROR, "%s failed to upload frame %d\n", __func__, ret);
        return NI_RETCODE_FAILURE;
    }

    dst_surf->ui16width = p_split_ctx->w[0] = width;
    dst_surf->ui16height = p_split_ctx->h[0] = height;
    dst_surf->ui32nodeAddress = 0;   // always 0 offset for upload
    dst_surf->encoding_type = isSemiPlanar ? NI_PIXEL_PLANAR_FORMAT_SEMIPLANAR :
                                             NI_PIXEL_PLANAR_FORMAT_PLANAR;

    ni_log(NI_LOG_DEBUG, "%s trace ui16FrameIdx = [%u] hdl %d SID%d\n",
           __func__, dst_surf->ui16FrameIdx, dst_surf->device_handle,
           dst_surf->ui16session_ID);

    // Update frames context
    p_split_ctx->f[0] = (int)dst_surf->encoding_type;
    return ret;
}

int ni_quadra_filter_hwdl_frame(NIFramesContext *ni_ctx,
                                niquadraFrameSurface1_t *src_surf, int width,
                                int height, int ni_fmat, int nb_planes,
                                const int *linesize, uint8_t **dst_data)
{
    ni_session_context_t *p_session_ctx = ni_ctx->api_ctx;
    ni_session_data_io_t session_io_data;
    ni_session_data_io_t *p_session_data = &session_io_data;
    int ret = 0;
    memset(p_session_data, 0, sizeof(ni_session_data_io_t));
    ret = ni_frame_buffer_alloc_dl(&(p_session_data->data.frame), width, height,
                                   ni_fmat);
    if (ret != NI_RETCODE_SUCCESS)
    {
        ni_log(NI_LOG_ERROR, "%s Cannot allocate ni_frame\n", __func__);
        return NI_RETCODE_ERROR_MEM_ALOC;
    }

    p_session_ctx->is_auto_dl = false;
    ret = ni_device_session_hwdl(p_session_ctx, p_session_data, (niFrameSurface1_t *)src_surf);
    if (ret <= 0)
    {
        ni_log(NI_LOG_ERROR, "%s failed to retrieve frame\n", __func__);
        ni_frame_buffer_free(&p_session_data->data.frame);
        return NI_RETCODE_FAILURE;
    }

    ret = ni_to_avframe_copy(&p_session_data->data.frame, width, height,
                             ni_fmat, nb_planes, linesize, dst_data);
    if (ret < 0)
    {
        ni_log(NI_LOG_ERROR, "Can't copy frame %d\n", ret);
        ni_frame_buffer_free(&p_session_data->data.frame);
        return ret;
    }

    ni_frame_buffer_free(&p_session_data->data.frame);
    return ret;
}

void ni_quadra_split_ctx_set(NIFramesContext *ni_ctx,
                             niquadraFrameSurface1_t *ni_surf,
                             niquadraFrameSurface1_t *ni_surf_1,
                             niquadraFrameSurface1_t *ni_surf_2,
                             int num_extra_outputs)
{
    ni_split_context_t *split_ctx = (ni_split_context_t *)&ni_ctx->split_ctx;
    split_ctx->enabled = (num_extra_outputs >= 1) ? 1 : 0;
    split_ctx->w[0] = ni_surf->ui16width;
    split_ctx->h[0] = ni_surf->ui16height;
    split_ctx->f[0] = (int)ni_surf->encoding_type;
    split_ctx->f8b[0] = (int)ni_surf->bit_depth;
    split_ctx->w[1] = (num_extra_outputs >= 1) ? ni_surf_1->ui16width : 0;
    split_ctx->h[1] = (num_extra_outputs >= 1) ? ni_surf_1->ui16height : 0;
    split_ctx->f[1] = (num_extra_outputs >= 1) ? ni_surf_1->encoding_type : 0;
    split_ctx->f8b[1] = (num_extra_outputs >= 1) ? ni_surf_1->bit_depth : 0;
    split_ctx->w[2] = (num_extra_outputs == 2) ? ni_surf_2->ui16width : 0;
    split_ctx->h[2] = (num_extra_outputs == 2) ? ni_surf_2->ui16height : 0;
    split_ctx->f[2] = (num_extra_outputs == 2) ? ni_surf_2->encoding_type : 0;
    split_ctx->f8b[2] = (num_extra_outputs == 2) ? ni_surf_2->bit_depth : 0;
}

int ni_quadra_get_device_id(void *session_ctx)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    return p_session_ctx->hw_id;
}

uint32_t ni_quadra_get_session_id(void *session_ctx)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    return p_session_ctx->session_id;
}

void ni_quadra_device_close(int32_t handle)
{
    ni_device_close(handle);
}

int32_t ni_quadra_get_device_handle(void *session_ctx)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    return p_session_ctx->device_handle;
}

void ni_quadra_set_device_handle(void *session_ctx, int32_t handle)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    p_session_ctx->blk_io_handle = handle;
}

void ni_quadra_filter_session_dl_set(void *session_ctx, bool value)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    p_session_ctx->is_auto_dl = value;
}

int ni_quadra_device_session_copy(void *src_ctx, void *dst_ctx)
{
    return ni_device_session_copy(src_ctx, dst_ctx);
}

void ni_quadra_session_timestamp_copy(void *src_ctx, void *dst_ctx)
{
    ni_session_context_t *p_src_session = src_ctx;
    ni_session_context_t *p_dst_session = dst_ctx;
    p_dst_session->session_timestamp = p_src_session->session_timestamp;
}

int ni_quadra_get_session_status(void *session_ctx)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    return p_session_ctx->status;
}
int ni_quadra_get_session_runstatus(void *session_ctx)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    return p_session_ctx->session_run_state;
}
int ni_quadra_get_session_blkiohandle(void *session_ctx)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    return p_session_ctx->blk_io_handle;
}

uint16_t ni_quadra_get_frame_idx(void *niframe)
{
    ni_session_data_io_t *p_frame = niframe;
    niFrameSurface1_t *frame_surface = (niFrameSurface1_t *)p_frame->data.frame.p_data[3];
    return frame_surface->ui16FrameIdx;
}

uint8_t *ni_quadra_get_niframe(void *niframe)
{
    ni_session_data_io_t *p_frame = niframe;
    return (uint8_t *)&p_frame->data.frame;
}

uint8_t *ni_quadra_get_frame_buffer(void *niframe)
{
    ni_session_data_io_t *p_frame = niframe;
    return p_frame->data.frame.p_buffer;
}

uint32_t ni_quadra_get_frame_buffer_size(void *niframe)
{
    ni_session_data_io_t *p_frame = niframe;
    return p_frame->data.frame.buffer_size;
}

uint8_t **ni_quadra_get_frame_data(void *niframe)
{
    ni_session_data_io_t *p_frame = niframe;
    return (uint8_t **)&p_frame->data.frame.p_data;
}

uint8_t *ni_quadra_get_packet_data(void *nipacket)
{
    ni_session_data_io_t *p_pkt = nipacket;
    return p_pkt->data.packet.p_data;
}

uint32_t ni_quadra_get_packet_data_len(void *nipacket)
{
    ni_session_data_io_t *p_pkt = nipacket;
    return p_pkt->data.packet.data_len;
}

void ni_quadra_get_frame_data_len(void *niframe, uint32_t *dst_datalen)
{
    ni_session_data_io_t *p_frame = niframe;
    int i;
    for (i = 0; i < NI_QUADRA_MAX_NUM_DATA_POINTERS; i++)
    {
        dst_datalen[i] = p_frame->data.frame.data_len[i];
    }
}

int ni_quadra_filter_hwframe_recycle(void *session_ctx, void *niframe)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    ni_session_data_io_t *p_frame = niframe;

    return ni_hwframe_buffer_recycle(
        (niFrameSurface1_t *)p_frame->data.frame.p_data[3],
        p_session_ctx->device_handle);
}

void ni_quadra_set_session_runstatus(void *session_ctx, int status)
{
    ni_session_context_t *p_session_ctx = session_ctx;
    p_session_ctx->session_run_state = status;
}

int ni_quadra_filter_device_config_frame(void *p_ctx,
                                         ni_quadra_frame_config_t *p_cfg)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    return ni_device_config_frame(p_session_ctx, (ni_frame_config_t *)p_cfg);
}

int ni_quadra_hwframe_buffer_recycle(niquadraFrameSurface1_t *surface,
                                     int32_t device_handle)
{
    return ni_hwframe_buffer_recycle((niFrameSurface1_t *)surface, device_handle);
}

void ni_quadra_get_hw_yuv420p_dim(int width, int height, int factor,
                                  int is_semiplanar, int plane_stride[4],
                                  int plane_height[4])
{
    // strides are multiples of 128
    if (width < NI_MIN_WIDTH)
    {
        plane_stride[0] = ((NI_MIN_WIDTH * factor + 127) / 128) * 128;
        plane_stride[1] =
            (((NI_MIN_WIDTH / (is_semiplanar ? 1 : 2) * factor) + 127) / 128) *
            128;
        plane_stride[2] =
            ((((is_semiplanar ? 0 : NI_MIN_WIDTH) / 2 * factor) + 127) / 128) *
            128;
    } else
    {
        width = ((width + 1) / 2) * 2;   // pad odd resolution
        plane_stride[0] = ((width * factor + 127) / 128) * 128;
        plane_stride[1] =
            (((width / (is_semiplanar ? 1 : 2) * factor) + 127) / 128) * 128;
        plane_stride[2] = (is_semiplanar ? 0 : plane_stride[1]);
    }

    // height (in lines) just needs to be even number
    if (height < NI_MIN_HEIGHT)
    {
        plane_height[0] = NI_MIN_HEIGHT;
    } else
    {
        plane_height[0] = ((height + 1) / 2) * 2;
    }
    plane_height[1] = plane_height[2] = plane_height[0] / 2;
}

int ni_quadra_get_seiset_size()
{
    return sizeof(ni_custom_sei_set_t);
}

void ni_quadra_log_set_level(int level)
{
    ni_log_set_level(ff_to_ni_log_level(level));
}

void ni_quadra_log_set_callback(void (*log_callback)(int, const char *,
                                                     va_list))
{
    ni_log_set_callback(log_callback);
}

ni_quadra_log_level_t ni_quadra_ff_to_ni_log_level(int fflog_level)
{
    return (ni_quadra_log_level_t)ff_to_ni_log_level(fflog_level);
}

ni_quadra_hw_device_info_coder_param_t *ni_quadra_create_hw_device_info_coder_param(int mode)
{
    return (ni_quadra_hw_device_info_coder_param_t *) ni_create_hw_device_info_quadra_coder_param(mode);
}

void ni_quadra_destory_hw_device_info_coder_param(ni_quadra_hw_device_info_coder_param_t *p_quadra_hw_device_info_coder_param)
{
    ni_destory_hw_device_info_quadra_coder_param((ni_hw_device_info_quadra_coder_param_t *) p_quadra_hw_device_info_coder_param);
}

ni_quadra_hw_device_info_t *ni_quadra_hw_device_info_alloc(int device_type_num, int avaliable_card_num)
{
    return (ni_quadra_hw_device_info_t *) ni_hw_device_info_alloc_quadra(device_type_num, avaliable_card_num);
}

void ni_quadra_hw_device_info_free(ni_quadra_hw_device_info_t *p_hw_device_info)
{
    ni_hw_device_info_free_quadra((ni_hw_device_info_quadra_t *)p_hw_device_info);
}

int ni_quadra_check_hw_info(ni_quadra_hw_device_info_t **pointer_to_p_hw_device_info, 
                             int task_mode,
                             ni_quadra_hw_device_info_threshold_param_t *hw_info_threshold_param,
                             int preferential_device_type,
                             ni_quadra_hw_device_info_coder_param_t * coder_param,
                             int hw_mode,
                             int consider_mem)
{

    return ni_check_hw_info((ni_hw_device_info_quadra_t **)pointer_to_p_hw_device_info,
                            task_mode,
                            (ni_hw_device_info_quadra_threshold_param_t *)hw_info_threshold_param,
                            preferential_device_type,
                            (ni_hw_device_info_quadra_coder_param_t *)coder_param,
                            hw_mode,
                            consider_mem);
}

int ni_quadra_rsrc_get_device_by_block_name(const char *blk_name,
                                                    ni_quadra_device_type_t device_type)
{
    return ni_rsrc_get_device_by_block_name(blk_name, (ni_device_type_t)device_type);
}

/*!*****************************************************************************
 *  \brief  Open a new device session depending on the device_type parameter
 *          If device_type is NI_DEVICE_TYPE_DECODER opens decoding session
 *          If device_type is NI_DEVICE_TYPE_ENCODER opens encoding session
 *          If device_type is NI_DEVICE_TYPE_SCALER opens scaling session
 *
 *  \param[in] p_ctx        Pointer to a caller allocated
 *                          ni_session_context_t struct
 *  \param[in] device_type  NI_DEVICE_TYPE_DECODER, NI_DEVICE_TYPE_ENCODER,
 *                          or NI_DEVICE_TYPE_SCALER
 *  \return On success
 *                          NI_RETCODE_SUCCESS
 *          On failure
 *                          NI_RETCODE_INVALID_PARAM
 *                          NI_RETCODE_ERROR_MEM_ALOC
 *                          NI_RETCODE_ERROR_NVME_CMD_FAILED
 *                          NI_RETCODE_ERROR_INVALID_SESSION
 ******************************************************************************/
int ni_quadra_device_session_open(void *p_ctx,
                                    ni_quadra_device_type_t device_type)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    return ni_device_session_open(p_session_ctx, (ni_device_type_t)device_type);
}

/*!*****************************************************************************
 *  \brief  Sends data to the device
 *          If device_type is NI_DEVICE_TYPE_DECODER sends data packet to
 *          decoder
 *          If device_type is NI_DEVICE_TYPE_ENCODER sends data frame to encoder
 *          If device_type is NI_DEVICE_TYPE_AI sends data frame to ai engine
 *
 *  \param[in] p_ctx        Pointer to a caller allocated
 *                          ni_session_context_t struct
 *  \param[in] p_data       Pointer to a caller allocated
 *                          ni_session_data_io_t struct which contains either a
 *                          ni_frame_t data frame or ni_packet_t data packet to
 *                          send
 *  \param[in] device_type  NI_DEVICE_TYPE_DECODER or NI_DEVICE_TYPE_ENCODER or
 *                          NI_DEVICE_TYPE_AI
 *                          If NI_DEVICE_TYPE_DECODER is specified, it is
 *                          expected that the ni_packet_t struct inside the
 *                          p_data pointer contains data to send.
 *                          If NI_DEVICE_TYPE_ENCODER or NI_DEVICE_TYPE_AI is
 *                          specified, it is expected that the ni_frame_t
 *                          struct inside the p_data pointer contains data to
 *                          send.
 *  \return On success
 *                          Total number of bytes written
 *          On failure
 *                          NI_RETCODE_INVALID_PARAM
 *                          NI_RETCODE_ERROR_NVME_CMD_FAILED
 *                          NI_RETCODE_ERROR_INVALID_SESSION
 ******************************************************************************/
int ni_quadra_device_session_write(void *p_ctx, void *p_data,
                                   ni_quadra_device_type_t device_type)
{
    return ni_device_session_write((ni_session_context_t *)p_ctx,
                                   (ni_session_data_io_t *)p_data, (ni_device_type_t)device_type);
}

/*!*****************************************************************************
 *  \brief  Free packet buffer that was previously allocated with
 *          ni_packet_buffer_alloc
 *
 *  \param[in] p_data    Pointer to void *
 *                       ni_session_data_io_t struct
 *
 *  \return On success    NI_RETCODE_SUCCESS
 *          On failure    NI_RETCODE_INVALID_PARAM
 ******************************************************************************/
int ni_quadra_packet_buffer_free(void* p_data)
{
    ni_session_data_io_t *p_packet = p_data;
    return ni_packet_buffer_free(&p_packet->data.packet);
}

/*!*****************************************************************************
 *  \brief  Free frame buffer that was previously allocated with either
 *          ni_frame_buffer_alloc or ni_encoder_frame_buffer_alloc or
 *          ni_frame_buffer_alloc_nv
 *
 *  \param[in] p_data    Pointer to void *
 *                       ni_session_data_io_t struct
 *
 *  \return On success    NI_RETCODE_SUCCESS
 *          On failure    NI_RETCODE_INVALID_PARAM
 ******************************************************************************/
int ni_quadra_frame_buffer_free(void* p_data)
{
    ni_session_data_io_t *p_frame = p_data;
    return ni_frame_buffer_free(&p_frame->data.frame);
}

/*!*****************************************************************************
 *  \brief  Create and return the allocated ni_device_pool_t struct
 *
 *  \param  None
 *
 *  \return Pointer to ni_device_pool_t struct on success, or NULL on failure
 *******************************************************************************/
ni_quadra_device_pool_t* ni_quadra_rsrc_get_device_pool(void)
{
    return (ni_quadra_device_pool_t*)ni_rsrc_get_device_pool();
}

/*!*****************************************************************************
 *  \brief  Close device session that was previously opened by calling
 *          ni_device_session_open()
 *          If device_type is NI_DEVICE_TYPE_DECODER closes decoding session
 *          If device_type is NI_DEVICE_TYPE_ENCODER closes encoding session
 *          If device_type is NI_DEVICE_TYPE_SCALER closes scaling session
 *
 *  \param[in] p_ctx        Pointer to a caller allocated
 *                          ni_session_context_t struct
 *  \param[in] eos_received Flag indicating if End Of Stream indicator was
 *                          received
 *  \param[in] device_type  NI_DEVICE_TYPE_DECODER, NI_DEVICE_TYPE_ENCODER,
 *                          or NI_DEVICE_TYPE_SCALER
 *  \return On success
 *                          NI_RETCODE_SUCCESS
 *          On failure
 *                          NI_RETCODE_INVALID_PARAM
 *                          NI_RETCODE_ERROR_NVME_CMD_FAILED
 *                          NI_RETCODE_ERROR_INVALID_SESSION
 ******************************************************************************/
int ni_quadra_device_session_close(void *p_ctx,
                                    int eos_recieved,
                                    ni_quadra_device_type_t device_type)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    return ni_device_session_close(p_session_ctx, eos_recieved, (ni_device_type_t)device_type);
}

/*!*****************************************************************************
 *  \brief  Clear already allocated session context
 *
 *  \param[in]  p_ctx Pointer to an already allocated ni_session_context_t
 *
 *
 ******************************************************************************/
void ni_quadra_device_session_context_clear(void *p_ctx)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    ni_device_session_context_clear(p_session_ctx);
}

int ni_quadra_frame_buffer_alloc_hwenc(void *niframe,
                                        int video_width, int video_height,
                                        int extra_len)
{
    ni_session_data_io_t *p_frame = niframe;
    return ni_frame_buffer_alloc_hwenc(&p_frame->data.frame, video_width,
                                       video_height, extra_len);
}

void ni_quadra_set_framerate(void *p_ctx, int num, int den)
{
    ni_session_context_t *p_session_ctx = p_ctx;

    p_session_ctx->framerate.framerate_num = num;
    p_session_ctx->framerate.framerate_denom = den;
}

int ni_quadra_get_session_packet_num(void *p_ctx)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    return p_session_ctx->pkt_num;
}

int ni_quadra_get_session_frame_num(void *p_ctx)
{
    ni_session_context_t *p_session_ctx = p_ctx;
    return p_session_ctx->frame_num;
}
